// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:28 $
/// $Revision: 1.1 $
// ************************************************************************* //
#ifndef __COMMAND_HH__
#define __COMMAND_HH__

// ************************************************************************* //
#include <map>
#include "mystring.h"
#include "gfxtypes.h"

// ************************************************************************* //
class DagNode;

// ************************************************************************* //
class Parameter
{
protected:
  String	_m_name;
  bool		_m_hasDefault;
public:
  Parameter(const String &name, const bool hasDefault):
    _m_name(name), _m_hasDefault(hasDefault) {}
  virtual ~Parameter() {}
  virtual bool	browse(const String &parameters, String &errorMsg) = 0;
  virtual bool		readAsSignal() { return false; }
  virtual int		readAsInt() { return 0; }
  virtual String	readAsString() { return String(""); }
  virtual const Pnt3d	&readAsPnt3d() { static Pnt3d v(0.0); return v; }
  virtual const Vec3d	&readAsVec3d() { static Vec3d v(0.0); return v; }
  virtual String	typeStr() const = 0;
};

// ************************************************************************* //
class ParameterSignal: public Parameter
{
  bool	_m_defaultValue;
  bool	_m_value;
public:
  ParameterSignal(const bool defaultValue, const String &name):
    Parameter(name, true), _m_defaultValue(defaultValue) {}
  ParameterSignal(const String &name): Parameter(name, false) {}
  virtual bool	browse(const String &parameters, String &errorMsg);
  virtual bool	readAsSignal() { return _m_value; }
  virtual String	typeStr() const { return ""; }
};

// ************************************************************************* //
class ParameterInt: public Parameter
{
  int	_m_defaultValue;
  int	_m_value;
public:
  ParameterInt(const int defaultValue, const String &name):
    Parameter(name, true), _m_defaultValue(defaultValue) {}
  ParameterInt(const String &name): Parameter(name, false) {}
  virtual bool	browse(const String &parameters, String &errorMsg);
  virtual int	readAsInt() { return _m_value; }
  virtual String	typeStr() const { return "<int>"; }
};

// ************************************************************************* //
class ParameterPnt3d: public Parameter
{
  Pnt3d	_m_defaultValue;
  Pnt3d	_m_value;
public:
  ParameterPnt3d(const Pnt3d &defaultValue, const String &name):
    Parameter(name, true), _m_defaultValue(defaultValue) {}
  ParameterPnt3d(const String &name): Parameter(name, false) {}
  virtual bool	browse(const String &parameters, String &errorMsg);
  virtual const Pnt3d	&readAsPnt3d() { return _m_value; }
  virtual String	typeStr() const { return "<x, y, z>"; }
};

// ************************************************************************* //
class ParameterVec3d: public Parameter
{
  Vec3d	_m_defaultValue;
  Vec3d	_m_value;
public:
  ParameterVec3d(const Vec3d &defaultValue, const String &name):
    Parameter(name, true), _m_defaultValue(defaultValue) {}
  ParameterVec3d(const String &name): Parameter(name, false) {}
  virtual bool	browse(const String &parameters, String &errorMsg);
  virtual const Vec3d	&readAsVec3d() { return _m_value; }
  virtual String	typeStr() const { return "<x, y, z>"; }
};

// ************************************************************************* //
/// DESCRIPTION: If the parameter is the path of a node, build object with
///  "sel" as the selection index of the node. Like this, the path can
///  be specified by selecting the object, as well as in the command line.
///  In other cases, keep "sel" to default value 0.
// ************************************************************************* //
class ParameterString: public Parameter
{
  String	_m_defaultValue;
  String	_m_value;
  unsigned	_m_sel;
public:
  ParameterString(const String &defaultValue, const String &name,
		  const unsigned sel = 0):
    Parameter(name, true), _m_defaultValue(defaultValue), _m_sel(sel) {}
  ParameterString(const String &name, const unsigned sel = 0):
    Parameter(name, false), _m_sel(sel) {}
  virtual bool		browse(const String &parameters, String &errorMsg);
  virtual String	readAsString() { return _m_value; }
  virtual String	typeStr() const { return "<string>"; }
};

// ************************************************************************* //
/// DESCRIPTION: Command interpreted by script language. Extend with static 
///  Classes.
// ************************************************************************* //
class Command
{
  std::map<String, Parameter*>	_m_parameters;
public:
  Command() {}
  virtual ~Command() {}
  virtual int		getId() const = 0;
  virtual const String	&getName() const = 0;
  virtual DagNode	*execute(const String &params) = 0;
  static std::map<int, String>		&id2names();
  static std::map<String, int>		&names2id();
  static std::map<int, Command *(*)()>	&creators();
  static void	registerCommand(const int id, const String &name,
				Command *(*creator)());
  void	addSignal(const String &name);
  void	addInt(const int defaultValue, const String &name);
  void	addInt(const String &name);
  void	addPnt3d(const Pnt3d &defaultValue, const String &name);
  void	addPnt3d(const String &name);
  void	addVec3d(const Vec3d &defaultValue, const String &name);
  void	addVec3d(const String &name);
  void	addString(const String &defaultValue, const String &name,
		  const unsigned sel = 0);
  void	addString(const String &name, const unsigned sel = 0);
  bool		readAsSignal(const String &name);
  int		readAsInt(const String &name);
  const Pnt3d	&readAsPnt3d(const String &name);
  const Vec3d	&readAsVec3d(const String &name);
  String	readAsString(const String &name);
  bool		browse(const String &parameters, String &errorMsg);
  static DagNode	*run(const String &cmd);
  String	usage() const;
};

// ************************************************************************* //
/// DESCRIPTION: Registers the command.
// ************************************************************************* //
inline void	Command::registerCommand(const int	id,
					 const String	&name,
					 Command	*(*creator)())
{
  id2names()[id] = name;
  names2id()[name] = id;
  creators()[id] = creator;
}

// ************************************************************************* //
/// DESCRIPTION: Command id/names repository.
// ************************************************************************* //
inline std::map<int, String> &Command::id2names()
{
  static std::map<int, String>	snames;

  return snames;
}

// ************************************************************************* //
/// DESCRIPTION: Command id/names repository.
// ************************************************************************* //
inline std::map<String, int> &Command::names2id()
{
  static std::map<String, int>	snames;

  return snames;
}

// ************************************************************************* //
/// DESCRIPTION: Node constructor repositoty. 
// ************************************************************************* //
inline std::map<int, Command *(*)()> &Command::creators()
{
  static std::map<int, Command *(*)()>	screators;

  return screators;
}

// ************************************************************************* //
#endif//__COMMAND_HH__
