// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:28 $
/// $Revision: 1.1 $
// ************************************************************************* //
#ifndef __DAGNODE_HH__
#define __DAGNODE_HH__

// ************************************************************************* //
#include <map>
#include <list>
#include <vector>
#include "gfxtypes.h"
#include "mystring.h"

// ************************************************************************* //
class DataBlock;
class Plug;
class TwistLog;

// ************************************************************************* //
/// DESCRIPTION: Node that may contain sub-Nodes.
/// CONTEXT: To extend, declare a "creator" method and an "id" int, and pass
///  them to registerNode.
///  Call method initialize last in constructor.
///  typeStr	- name of node type
///  typeNbr	- number of node type
///  instStr	- id name of node instance
///  instNbr	- id num of node instance
/// Datablock = internal node
// ************************************************************************* //
class DagNode
{
protected:
  DagNode		*_m_parent;
  std::list<DagNode*>	_m_children;
  mutable DataBlock	*_m_dataBlock;
  int			_m_instNbr;
  String		_m_instStr;
private:
  virtual DagNode	*getNodeByPathRec(const String &path);
  virtual Plug		*getPlugByPathRec(const String &path);
  virtual void		selectedRec(std::list<DagNode*> &nodes);
  static DagNode	*_m_root;
public:
  DagNode(DagNode *parent = 0);
  virtual	~DagNode();
  DataBlock	&getDataBlock() const;
  virtual void 	add(DagNode &node)
  { _m_children.push_back(&node), node._m_parent = this; }
  virtual void	remove(DagNode &node) { _m_children.remove(&node); }
  virtual void	initialize() {}
  void		createInstName();
  // getNode *************************************************************** //
  static DagNode	*getNodeByPath(const String &path);
  static Plug		*getPlugByPath(const String &path);
  virtual DagNode	*getNode(const int typeNbr, const int instNbr);
  virtual DagNode	*getNodeParent() { return _m_parent; }
  virtual void		getNodesByDepth(int depth, std::list<DagNode *>&nodes);
  virtual void		getNodesAtDepth(int depth, std::list<DagNode *>&nodes);
  virtual void		getNodes(const char *, std::list<DagNode *> &nodes);
  virtual void		getNodesByType(const String &s,
				       std::list<DagNode *> &nodes);
  virtual void		getChildrenByType(const String &s,
					  std::list<DagNode *> &nodes);
  virtual DagNode	*getNodeParentByType(const String &s);
  static void		getNodesSelected(std::list<DagNode*> &nodes);
  const std::list<DagNode*>	&children() const { return _m_children; }
  // *********************************************************************** //
  virtual int	getTypeNbr() const = 0;
  virtual int	getInstNbr() const { return _m_instNbr; }
  virtual const String	&getTypeStr() const = 0;
  virtual const String	&getInstStr() const { return _m_instStr; }
  virtual String	path() const;
  static void	parent(DagNode &parent, DagNode &child);
  static void	unparent(DagNode &parent, DagNode &child);
  static void	connect(Plug &d1, Plug &d2);
  static void	connect(DagNode &node1, const String &sn1,
			DagNode &node2, const String &sn2);
  static void	disconnect(DagNode &node1, const String &sn1,
			   DagNode &node2, const String &sn2);
  static void	disconnect(DagNode &node, const String &sn);
  static void	disconnectAll(DagNode &node1);
  virtual void	compute(Plug &) {}
  static void	registerNode(const int id, const String &name,
			     DagNode *(*f)());
  static DagNode	*(*queryCreator(const int id))();
  static std::map<int, DagNode *(*)()>	&creators();
  static std::map<int, int>		&counters();
  static std::map<int, String>		&id2names();
  static std::map<String, int>		&names2id();
  virtual Plug	*getPlug(const String &name);
  Plug	*getPlug(const int id);
  void	getPlugByType(const int type, std::list<Plug *> &hnd);
  void	deselect();
  // Root ****************************************************************** //
  static DagNode	&root() { return *_m_root; }
  static void		setRoot(DagNode	&r) { _m_root = &r; }
  static int		freeTypeNbr();
  // Shortcuts ************************************************************* //
  // Gets ****************************************************************** //
  double			getAsDouble(const char *name) const;
  // Reads ***************************************************************** //
  void				*readAsCustom(const char *name) const;
  void				readAsSignal(const char *name) const;
  bool				readAsBool(const char *name) const;
  int				readAsInt(const char *name) const;
  double			readAsDouble(const char *name) const;
  const String			&readAsString(const char *name) const;
  const Pnt2d			&readAsPnt2d(const char *name) const;
  const Mat2i			&readAsMat2i(const char *name) const;
  const Pnt3d			&readAsPnt3d(const char *name) const;
  const Vec3d			&readAsVec3d(const char *name) const;
  const Rgb3d			&readAsRgb3d(const char *name) const;
  const Mat3d			&readAsMat3d(const char *name) const;
  const TwistLog		&readAsTwistLog(const char *name) const;
  const Mat4d			&readAsMat4d(const char *name) const;
  const std::vector<int>	&readAsIntArray(const char *name)const;
  const std::vector<double>	&readAsDoubleArray(const char *name)const;
  const std::vector<Pnt2d>	&readAsPnt2dArray(const char *name)const;
  const Pnt2d			&readAsPnt2dArray(const char *name,
						  const int i)const;
  const std::vector<Pnt3d>	&readAsPnt3dArray(const char *name)const;
  const Pnt3d			&readAsPnt3dArray(const char *name,
						  const int i)const;
  const std::vector<Vec3d>	&readAsVec3dArray(const char *name)const;
  const Vec3d			&readAsVec3dArray(const char *name,
						  const int i)const;
  const std::vector<TwistLog>	&readAsTwistLogArray(const char *name)const;
  const TwistLog		&readAsTwistLogArray(const char *name,
						     const int i)const;
  const std::list<Pnt2d>	&readAsPnt2dList(const char *name)const;
  // Writes **************************************************************** //
  void	writeAsCustom(const char *name, void *value);
  void	writeAsSignal(const char *name);
  void	writeAsBool(const char *name, const bool value);
  void	writeAsInt(const char *name, const int value);
  void	writeAsDouble(const char *name, const double value);
  void	writeAsString(const char *name, const String &value);
  void	writeAsPnt2d(const char *name, const Pnt2d &value);
  void	writeAsMat2i(const char *name, const Mat2i &value);
  void	writeAsPnt3d(const char *name, const Pnt3d &value);
  void	writeAsVec3d(const char *name, const Vec3d &value);
  void	writeAsRgb3d(const char *name, const Rgb3d &value);
  void	writeAsMat3d(const char *name, const Mat3d &value);
  void	writeAsTwistLog(const char *name, const TwistLog &value);
  void	writeAsMat4d(const char *name, const Mat4d &value);
  void	writeAsIntArray(const char *name, const std::vector<int> &value);
  void	writeAsDoubleArray(const char *name, const std::vector<double> &value);
  void	writeAsPnt2dArray(const char *name, const std::vector<Pnt2d> &value);
  void	writeAsPnt2dArray(const char *name, const int i, const Pnt2d &value);
  void	writeAsPnt3dArray(const char *name, const std::vector<Pnt3d> &value);
  void	writeAsPnt3dArray(const char *name, const int i, const Pnt3d &value);
  void	writeAsVec3dArray(const char *name, const std::vector<Vec3d> &value);
  void	writeAsVec3dArray(const char *name, const int i, const Vec3d &value);
  void	writeAsTwistLogArray(const char *name,
			     const std::vector<TwistLog> &value);
  void	writeAsTwistLogArray(const char *name, const int i,
			     const TwistLog &value);
  void	writeAsPnt2dList(const char *name, const std::list<Pnt2d> &value);
  // Save/Load ************************************************************* //
  void	save(std::ostream &os) const;
  void	load(std::istream &is, std::map<int, void *> &v2real,
	     std::map<int, std::list<void **> > &ref2v,
	     std::list<Plug*> &notSaved);
  void	clear();
  // Push/Pop ************************************************************** //
  void	pushName() const;
  void	popName() const;
  // Fix ******************************************************************* //
  void	fixme();
};

// ************************************************************************* //
/// DESCRIPTION: Return the path of the instance in the DAG.
// ************************************************************************* //
inline String	DagNode::path() const
{
  if (_m_parent)
    return _m_parent->path() + ":" + getInstStr();
  return "";
}

// ************************************************************************* //
/// DESCRIPTION: Register the node. For dynamic nodes.
// ************************************************************************* //
inline void	DagNode::registerNode(const int		id,
				      const String	&name,
				      DagNode		*(*creator)())
{
  if (creators()[id])
    {
      std::cerr << "Error: node already exists" << std::endl;
      return;
    }
  creators()[id] = creator;
  id2names()[id] = name;
  names2id()[name] = id;
  counters()[id] = 0;
}

// ************************************************************************* //
inline DagNode	*(*DagNode::queryCreator(const int id))()	
{
  return creators()[id];
}

// ************************************************************************* //
/// DESCRIPTION: Extracts node of instance name instStr.
// ************************************************************************* //
inline void	DagNode::getNodes(const char		*instStr,
				  std::list<DagNode *>	&nodes)
{
  std::list<DagNode*>::const_iterator	nit;

  for (nit = _m_children.begin(); nit != _m_children.end(); ++nit)
    if ((*nit)->getInstStr() ==  instStr)
      nodes.push_back(*nit);
}

// ************************************************************************* //
#endif//__DAGNODE_HH__
