// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:28 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include "datablock.h"
#include <qlayout.h>
#include <qspinbox.h>
#include <qcheckbox.h>
#include <qwidget.h>
#include "plug.h"
#include "dagnode.h"
#include "mystring.h"
#include <fstream>

// ************************************************************************* //
DataBlock::DataBlock(DagNode &node): _m_node(node), _m_idCounter(0)
{
}

// ************************************************************************* //
/// DESCRIPTION: Adds an entry for a bool.
// ************************************************************************* //
void	DataBlock::addCustom(Plug		&d,
			     const char		*shortName,
			     const bool		isVisible,
			     const bool		isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  d._m_shortNamei = shortName;
  d._m_longNamei = shortName;
  d._m_bits = (isVisible? BIT_ISVISIBLE: 0) | (isSaved? BIT_ISSAVED: 0);
  _m_quickMap[shortName] = &d;
  _m_plugs.push_back(&d);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a signal to the datablock.
// ************************************************************************* //
void	DataBlock::addSignal(const char *shortName,
			     const bool isVisible,
			     const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  Plug	*newh = new PlugSignal(*this, shortName, isVisible, isSaved);

  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a double to the datablock.
// ************************************************************************* //
void	DataBlock::addBool(const bool	defaultValue,
			   const char	*shortName,
			   const bool	isVisible,
			   const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  Plug	*newh = new PlugBool(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a double to the datablock.
// ************************************************************************* //
void	DataBlock::addInt(const int	defaultValue,
			  const int	minValue,
			  const int	maxValue,
			  const char	*shortName,
			  const bool	isVisible,
			  const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  PlugInt	*newh = new PlugInt(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  newh->setMinValue(minValue);
  newh->setMaxValue(maxValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a double to the datablock.
// ************************************************************************* //
void	DataBlock::addDouble(const double	defaultValue,
			     const char		*shortName,
			     const bool		isVisible,
			     const bool		isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  PlugDouble	*newh = new PlugDouble(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a double to the datablock.
// ************************************************************************* //
void	DataBlock::addDouble(const double	defaultValue,
			     const double	minValue,
			     const double	maxValue,
			     const char		*shortName,
			     const bool		isVisible,
			     const bool		isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  PlugDouble	*newh = new PlugDouble(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  newh->setMinValue(minValue);
  newh->setMaxValue(maxValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a double to the datablock.
// ************************************************************************* //
void	DataBlock::addString(const String	&defaultValue,
			     const char		*shortName,
			     const bool		isVisible,
			     const bool		isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  Plug	*newh = new PlugString(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Pnt2d to the datablock.
// ************************************************************************* //
void	DataBlock::addPnt2d(const Pnt2d	&defaultValue,
			    const char	*shortName,
			    const bool	isVisible,
			    const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  Plug	*newh = new PlugPnt2d(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Vec3d to the datablock.
// ************************************************************************* //
void	DataBlock::addMat2i(const Mat2i	&defaultValue,
			    const char	*shortName,
			    const bool	isVisible,
			    const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  Plug	*newh = new PlugMat2i(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Pnt3d to the datablock.
// ************************************************************************* //
void	DataBlock::addPnt3d(const Pnt3d	&defaultValue,
			    const char	*shortName,
			    const bool	isVisible,
			    const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  Plug	*newh = new PlugPnt3d(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Pnt3d to the datablock.
// ************************************************************************* //
void	DataBlock::addPnt3d(const Pnt3d		&defaultValue,
			    const double	minValue,
			    const double	maxValue,
			    const char		*shortName,
			    const bool		isVisible,
			    const bool		isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  PlugPnt3d	*newh = new PlugPnt3d(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  newh->setMinValue(minValue);
  newh->setMaxValue(maxValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Vec3d to the datablock.
// ************************************************************************* //
void	DataBlock::addVec3d(const Vec3d	&defaultValue,
			    const char	*shortName,
			    const bool	isVisible,
			    const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  Plug	*newh = new PlugVec3d(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Vec3d to the datablock.
// ************************************************************************* //
void	DataBlock::addVec3d(const Vec3d		&defaultValue,
			    const double	minValue,
			    const double	maxValue,
			    const char		*shortName,
			    const bool		isVisible,
			    const bool		isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  PlugVec3d	*newh = new PlugVec3d(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  newh->setMinValue(minValue);
  newh->setMaxValue(maxValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Rgb3d to the datablock.
// ************************************************************************* //
void	DataBlock::addRgb3d(const Rgb3d	&defaultValue,
			    const char	*shortName,
			    const bool	isVisible,
			    const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  PlugRgb3d	*newh = new PlugRgb3d(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  newh->setMinValue(0.0);
  newh->setMaxValue(1.0);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Rgb3d to the datablock.
// ************************************************************************* //
void	DataBlock::addRgb3d(const Rgb3d		&defaultValue,
			    const double	minValue,
			    const double	maxValue,
			    const char		*shortName,
			    const bool		isVisible,
			    const bool		isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  PlugRgb3d	*newh = new PlugRgb3d(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  newh->setMinValue(minValue);
  newh->setMaxValue(maxValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Vec3d to the datablock.
// ************************************************************************* //
void	DataBlock::addMat3d(const Mat3d	&defaultValue,
			    const char	*shortName,
			    const bool	isVisible,
			    const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  Plug	*newh = new PlugMat3d(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Vec3d to the datablock.
// ************************************************************************* //
void	DataBlock::addTwistLog(const TwistLog	&defaultValue,
			       const char	*shortName,
			       const bool	isVisible,
			       const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  Plug	*newh = new PlugTwistLog(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Mat4d to the datablock.
// ************************************************************************* //
void	DataBlock::addMat4d(const Mat4d	&defaultValue,
			    const char	*shortName,
			    const bool	isVisible,
			    const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  Plug	*newh = new PlugMat4d(*this, shortName, isVisible, isSaved);

  newh->copy(&defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds an array of int to the datablock.
// ************************************************************************* //
void	DataBlock::addIntArray(const int	n,
			       const int	defaultValue,
			       const char	*shortName,
			       const bool	isVisible,
			       const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  PlugIntArray *newh = new PlugIntArray(*this, shortName,
					      isVisible, isSaved);

  newh->resize(n, &defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds an array of double to the datablock.
// ************************************************************************* //
void	DataBlock::addDoubleArray(const int	n,
				  const double	defaultValue,
				  const char	*shortName,
				  const bool	isVisible,
				  const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  PlugDoubleArray *newh = new PlugDoubleArray(*this, shortName,
					      isVisible, isSaved);

  newh->resize(n, &defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Pnt3d to the datablock.
// ************************************************************************* //
void	DataBlock::addPnt2dArray(const int	n,
				 const Pnt2d	&defaultValue,
				 const char	*shortName,
				 const bool	isVisible,
				 const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  PlugPnt2dArray *newh = new PlugPnt2dArray(*this, shortName,
					    isVisible, isSaved);

  newh->resize(n, &defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Pnt3d to the datablock.
// ************************************************************************* //
void	DataBlock::addPnt3dArray(const int	n,
				 const Pnt3d	&defaultValue,
				 const char	*shortName,
				 const bool	isVisible,
				 const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  PlugPnt3dArray *newh = new PlugPnt3dArray(*this, shortName,
					    isVisible, isSaved);

  newh->resize(n, &defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Vec3d to the datablock.
// ************************************************************************* //
void	DataBlock::addVec3dArray(const int	n,
				 const Vec3d	&defaultValue,
				 const char	*shortName,
				 const bool	isVisible,
				 const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  PlugVec3dArray *newh = new PlugVec3dArray(*this, shortName,
					    isVisible, isSaved);

  newh->resize(n, &defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Vec3d to the datablock.
// ************************************************************************* //
void	DataBlock::addTwistLogArray(const int		n,
				    const TwistLog	&defaultValue,
				    const char		*shortName,
				    const bool		isVisible,
				    const bool		isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  PlugTwistLogArray *newh = new PlugTwistLogArray(*this, shortName,
						  isVisible, isSaved);

  newh->resize(n, &defaultValue);
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Adds a Pnt2d list to the datablock.
// ************************************************************************* //
void	DataBlock::addPnt2dList(const char	*shortName,
				const bool	isVisible,
				const bool	isSaved)
{
  if (_m_quickMap[shortName] != 0)
    {
      std::cerr << "Warning: " << shortName << " exists already, not added."
		<< std::endl;
      return;
    }

  PlugPnt2dList	*newh = new PlugPnt2dList(*this, shortName,
					  isVisible, isSaved);

  newh->_m_childrensize = 0;
  _m_quickMap[shortName] = newh;
  _m_plugs.push_back(newh);
}

// ************************************************************************* //
/// DESCRIPTION: Warns everybody that need to know, that deletion has occured.
// ************************************************************************* //
DataBlock::~DataBlock()
{
  DataBlock::removeAll(*this);

  std::list<QWidget *>::iterator	vit;

  for (vit = _m_validBits.begin(); vit != _m_validBits.end(); ++vit)
    (*vit)->close();
  _m_validBits.clear();
}

// ************************************************************************* //
/// DESCRIPTION: Call this to specify what input affects.
///  Sets dirty the affected node.
// ************************************************************************* //
void	DataBlock::affect(const char	*input,
			  const char	*output)
{
  if (!_m_quickMap[input])
    std::cerr << "Warning: " << input << " is not a plug." << std::endl;
  else if (!_m_quickMap[output])
    std::cerr << "Warning: " << output << " is not a plug." << std::endl;
  else
    affect(*_m_quickMap[input], *_m_quickMap[output]);
}

// ************************************************************************* //
/// DESCRIPTION: Call this to specify what input affects.
///  Sets dirty the affected node.
// ************************************************************************* //
void	DataBlock::affect(Plug	&input,
			  Plug	&output)
{
  std::list<Plug*>::iterator	ait;

  for (ait = input._m_affects.begin(); ait != input._m_affects.end(); ++ait)
    if (*ait == &output)
      {
	std::cerr << "Warning: plugs are already connected." << std::endl;
	return;
      }
  input._m_affects.push_back(&output);
  output._m_bits |= BIT_ISDIRTY;
}

// ************************************************************************* //
/// DESCRIPTION: keeps only one instance
// ************************************************************************* //
void	DataBlock::connect(Plug	&dh1,
			   Plug	&dh2)
{
  if (dh2._m_input)
    {
      std::cerr << "Warning: input of " << dh2.getNode().path() << "."
		<< dh2.shortName() << " was replaced." << std::endl;
      disconnect(*dh2._m_input, dh2);
    }
  if (!dh2._m_parent)
    {
      dh2.freeData();
      dh2.linkdata(dh1);
    }
  dh1._m_outputs.push_back(&dh2);
  dh2._m_input = &dh1;
  dh2._m_bits |= BIT_ISDIRTY;
  dh2.propagateDirtyBit();
}

// ************************************************************************* //
/// DESCRIPTION: 
///  dh1	- input
///  dh2	- output
// ************************************************************************* //
void	DataBlock::disconnect(Plug	&dh1,
			      Plug	&dh2)
{
  dh1._m_outputs.remove(&dh2);
  dh2._m_input = 0;
  *dh2.rawdata() = 0;//!
  if (!dh2._m_parent)
    dh2.newData();
}

// ************************************************************************* //
/// DESCRIPTION: Disconnects the plug from any existing input or output.
// ************************************************************************* //
void	DataBlock::disconnect(Plug	&dh)
{
  if (dh._m_input)
    disconnect(*dh._m_input, dh);
  while (!dh._m_outputs.empty())
    disconnect(dh, *dh._m_outputs.front());
}

// ************************************************************************* //
/// DESCRIPTION: Disconnects all attributes from other nodes.
///  and keeps affects.
// ************************************************************************* //
void	DataBlock::disconnectAll(DataBlock &db)
{
  std::list<Plug *>::iterator sit;

  for (sit = db._m_plugs.begin(); sit != db._m_plugs.end(); ++sit)
    (*sit)->disconnectAll();
}

// ************************************************************************* //
/// DESCRIPTION: static method.
// ************************************************************************* //
void	DataBlock::removeAll(DataBlock &db)
{
  std::list<Plug *>::iterator	dit;

  DataBlock::disconnectAll(db);
  for (dit = db._m_plugs.begin(); dit != db._m_plugs.end(); ++dit)
    delete *dit;
  db._m_plugs.clear();
  db._m_quickMap.clear();
}

// ************************************************************************* //
/// DESCRIPTION:
///  Valid syntax: alternative of .<number>[number]
///  d.p.x
///  d.x
///  d[2].x
///  d[4][5].x
// ************************************************************************* //
Plug	*DataBlock::getPlug(const String &name)
{
  //std::cout << "in getPlug " << name << std::endl;
  if (name.contains('.') || name.contains('['))
    {
      String	rest;
      Plug	*ret(0);
      char	c(0);

      for (int i(0), n(name.len()); i < n; ++i)
	if (name(i) == '.' || name(i) == '[')
	  {
	    c = name(i);
	    break;
	  }
      rest = name.chopFront(c);
      ret = static_cast<Plug*>(_m_quickMap[name.chopBack(c)]);
      if (!ret)
	return 0;
      do
	{
	  String	item = rest;
	  char		c(0);

	  for (int i(0), n(rest.len()); i < n; ++i)
	    if (rest(i) == '.' || rest(i) == '[' || rest(i) == ',')
	      {
		c = rest(i);
		break;
	      }
	  if (c)
	    item = rest.chopBack(c), rest = rest.chopFront(c);
	  if (item.contains(']'))
	    item = item.chopBack(']');

	  const unsigned	index = item2index(item);

	  if (index < ret->size())
	    {
	      // TEST IF INDEX VALID!
	      if (!static_cast<TPlug*>(ret)->_m_children[index])
		static_cast<TPlug*>(ret)->add(index);
	      ret = static_cast<TPlug*>(ret)->_m_children[index];
	    }
	  else
	    std::cerr << "Error: index over-estimates array capacity "
		      << std::endl;
	}
      while (rest.contains('.') || rest.contains('[') || rest.contains(','));
      return ret;
    }
  else
    return _m_quickMap[name];
}

// ************************************************************************* //
/// DESCRIPTION: Overload of the above.
// ************************************************************************* //
Plug	*DataBlock::getPlug(const int id)
{
  std::list<Plug *>::iterator	hndit;

  for (hndit = _m_plugs.begin(); hndit != _m_plugs.end(); ++hndit)
    if ((*hndit)->getId() == id)
      return *hndit;
  return 0;
}

// ************************************************************************* //
/// DESCRIPTION: Finds all the pieces of data of type type.
// ************************************************************************* //
void	DataBlock::getPlugByType(const int		type,
				 std::list<Plug *>	&hnd)
{
  std::list<Plug *>::iterator	hndit;
  TPlug	*tdata;

  for (hndit = _m_plugs.begin(); hndit != _m_plugs.end(); ++hndit)
    {
      if ((tdata = dynamic_cast<TPlug*>(*hndit)))
	{
	  // TODO: go through hierarchy here.
	}
      else
	if ((*hndit)->getTypeNbr() == type)
	  hnd.push_back(*hndit);
    }
}

// ************************************************************************* //
/// DESCRIPTION: Deselect all the data.
// ************************************************************************* //
void	DataBlock::deselect()
{
  std::list<Plug *>::iterator	hndit;

  for (hndit = _m_plugs.begin(); hndit != _m_plugs.end(); ++hndit)
    (*hndit)->deselect();
}

// ************************************************************************* //
void	DataBlock::addValidBit(QWidget *b)
{
  if (b != 0)
    _m_validBits.push_back(b);
}

// ************************************************************************* //
void	DataBlock::remValidBit(QWidget *b)
{
  _m_validBits.remove(b);
}

// Gets ****************************************************************** //
double	DataBlock::getAsDouble(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug " << name
		<< " does not exist in " << node().path() << std::endl;
      return 0;
    }
  return plug->getAsDouble();
}

// ************************************************************************* //
void	*DataBlock::readAsCustom(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug " << name
		<< " does not exist in " << node().path() << std::endl;
      return 0;
    }
  return plug->readAsCustom();
}

// ************************************************************************* //
void	DataBlock::readAsSignal(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug " << name
		<< " does not exist in " << node().path() << std::endl;
      return;
    }
  return plug->readAsSignal();
}

// ************************************************************************* //
bool	DataBlock::readAsBool(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug " << name
		<< " does not exist in " << node().path() << std::endl;
      return false;
    }
  return plug->readAsBool();
}

// ************************************************************************* //
int	DataBlock::readAsInt(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug " << name
		<< " does not exist in " << node().path() << std::endl;
      return 0;
    }
  return plug->readAsInt();
}

// ************************************************************************* //
double	DataBlock::readAsDouble(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug " << name
		<< " does not exist in " << node().path() << std::endl;
      return 0.0;
    }
  return plug->readAsDouble();
}

// ************************************************************************* //
const String	&DataBlock::readAsString(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static String	value("");

      std::cerr << "Error: plug " << name
		<< " does not exist in " << node().path() << std::endl;
      return value;

    }
  return plug->readAsString();
}

// ************************************************************************* //
const Pnt2d	&DataBlock::readAsPnt2d(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static Pnt2d	value(0.0);

      std::cerr << "Error: plug " << name
		<< " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsPnt2d();
}

// ************************************************************************* //
const Mat2i	&DataBlock::readAsMat2i(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static Mat2i	value(0);
      std::cerr << "Error: plug " << name
		<< " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsMat2i();
}

// ************************************************************************* //
const Pnt3d	&DataBlock::readAsPnt3d(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static Pnt3d	value(0.0);

      std::cerr << "Error: plug " << name
		<< " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsPnt3d();
}

// ************************************************************************* //
const Vec3d	&DataBlock::readAsVec3d(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static Vec3d	value(0.0);

      std::cerr << "Error: plug " << name
		<< " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsVec3d();
}

// ************************************************************************* //
const Rgb3d	&DataBlock::readAsRgb3d(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static Rgb3d	value(0.0);

      std::cerr << "Error: plug " << name
		<< " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsRgb3d();
}

// ************************************************************************* //
const Mat3d	&DataBlock::readAsMat3d(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static Mat3d	value(0.0);
      std::cerr << "Error: plug " << name
		<< " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsMat3d();
}

// ************************************************************************* //
const TwistLog	&DataBlock::readAsTwistLog(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static TwistLog	value(0.0);

      std::cerr << "Error: plug " << name
		<< " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsTwistLog();
}

// ************************************************************************* //
const Mat4d	&DataBlock::readAsMat4d(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static Mat4d	value(0.0);

      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsMat4d();
}

// ************************************************************************* //
const std::vector<int> &DataBlock::readAsIntArray(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static std::vector<int>	value;

      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsIntArray();
}

// ************************************************************************* //
const std::vector<double> &DataBlock::readAsDoubleArray(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static std::vector<double>	value;

      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsDoubleArray();
}

// ************************************************************************* //
const std::vector<Pnt2d> &DataBlock::readAsPnt2dArray(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static std::vector<Pnt2d>	value;

      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsPnt2dArray();
}

// ************************************************************************* //
const Pnt2d	&DataBlock::readAsPnt2dArray(const char	*name,
					     const int	i)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static Pnt2d	value(0.0);

      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsPnt2dArray(i);
}

// ************************************************************************* //
const std::vector<Pnt3d> &DataBlock::readAsPnt3dArray(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static std::vector<Pnt3d>	value;

      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsPnt3dArray();
}

// ************************************************************************* //
const Pnt3d	&DataBlock::readAsPnt3dArray(const char	*name,
					     const int	i)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static Pnt3d	value(0.0);

      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsPnt3dArray(i);
}

// ************************************************************************* //
const std::vector<Vec3d> &DataBlock::readAsVec3dArray(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static std::vector<Vec3d>	value;

      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsVec3dArray();
}

// ************************************************************************* //
const Vec3d	&DataBlock::readAsVec3dArray(const char	*name,
					     const int	i)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static Vec3d	value(0.0);

      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsVec3dArray(i);
}

// ************************************************************************* //
const std::vector<TwistLog> &DataBlock::readAsTwistLogArray(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static std::vector<TwistLog>	value;

      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsTwistLogArray();
}

// ************************************************************************* //
const TwistLog	&DataBlock::readAsTwistLogArray(const char	*name,
						const int	i)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static TwistLog	value(0.0);

      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsTwistLogArray(i);
}

// ************************************************************************* //
const std::list<Pnt2d> &DataBlock::readAsPnt2dList(const char *name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      static std::list<Pnt2d>	value;

      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return value;
    }
  return plug->readAsPnt2dList();
}

// *********************************************************************** //
void	DataBlock::writeAsCustom(const char	*name,
				 const void	*value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsCustom(value);
}

// ************************************************************************* //
void	DataBlock::writeAsSignal(const char	*name)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsSignal();
}

// ************************************************************************* //
void	DataBlock::writeAsBool(const char	*name,
			       const bool	value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsBool(value);
}

// ************************************************************************* //
void	DataBlock::writeAsInt(const char	*name,
			      const int		value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsInt(value);
}

// ************************************************************************* //
void	DataBlock::writeAsDouble(const char	*name,
				 const double	value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsDouble(value);
}

// ************************************************************************* //
void	DataBlock::writeAsString(const char	*name,
				 const String	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsString(value);
}

// ************************************************************************* //
void	DataBlock::writeAsPnt2d(const char	*name,
				const Pnt2d	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsPnt2d(value);
}

// ************************************************************************* //
void	DataBlock::writeAsMat2i(const char	*name,
				const Mat2i	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsMat2i(value);
}

// ************************************************************************* //
void	DataBlock::writeAsPnt3d(const char	*name,
				const Pnt3d	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsPnt3d(value);
}

// ************************************************************************* //
void	DataBlock::writeAsVec3d(const char	*name,
				const Vec3d	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsVec3d(value);
}

// ************************************************************************* //
void	DataBlock::writeAsRgb3d(const char	*name,
				const Rgb3d	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsRgb3d(value);
}

// ************************************************************************* //
void	DataBlock::writeAsMat3d(const char	*name,
				const Mat3d	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsMat3d(value);
}

// ************************************************************************* //
void	DataBlock::writeAsTwistLog(const char		*name,
				   const TwistLog	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsTwistLog(value);
}

// ************************************************************************* //
void	DataBlock::writeAsMat4d(const char	*name,
				const Mat4d	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsMat4d(value);
}

// ************************************************************************* //
void	DataBlock::writeAsIntArray(const char			*name,
				   const std::vector<int>	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsIntArray(value);
}

// ************************************************************************* //
void	DataBlock::writeAsDoubleArray(const char		*name,
				      const std::vector<double>	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsDoubleArray(value);
}

// ************************************************************************* //
void	DataBlock::writeAsPnt2dArray(const char			*name,
				     const std::vector<Pnt2d>	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsPnt2dArray(value);
}

// ************************************************************************* //
void	DataBlock::writeAsPnt2dArray(const char		*name,
				     const int		i,
				     const Pnt2d	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsPnt2dArray(i, value);
}

// ************************************************************************* //
void	DataBlock::writeAsPnt3dArray(const char			*name,
				     const std::vector<Pnt3d>	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsPnt3dArray(value);
}

// ************************************************************************* //
void	DataBlock::writeAsPnt3dArray(const char		*name,
				     const int		i,
				     const Pnt3d	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsPnt3dArray(i, value);
}

// ************************************************************************* //
void	DataBlock::writeAsVec3dArray(const char			*name,
				     const std::vector<Vec3d>	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsVec3dArray(value);
}

// ************************************************************************* //
void	DataBlock::writeAsVec3dArray(const char		*name,
				     const int		i,
				     const Vec3d	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsVec3dArray(i, value);
}

// ************************************************************************* //
void	DataBlock::writeAsTwistLogArray(const char			*name,
					const std::vector<TwistLog>	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsTwistLogArray(value);
}

// ************************************************************************* //
void	DataBlock::writeAsTwistLogArray(const char	*name,
					const int	i,
					const TwistLog	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsTwistLogArray(i, value);
}

// ************************************************************************* //
void	DataBlock::writeAsPnt2dList(const char			*name,
				    const std::list<Pnt2d>	&value)
{
  Plug *plug = getPlug(name);

  if (!plug)
    {
      std::cerr << "Error: plug "
		<< name << " does not exist in " << node().path() << std::endl;
      return;
    }
  plug->writeAsPnt2dList(value);
}

// ************************************************************************* //
void	DataBlock::save(std::ostream	&os)
{
  std::list<Plug *>::const_iterator sit;
  std::map<void *, int>	adress2virtual;

  os << "{" << std::endl;
  os << _m_idCounter << std::endl;
  for (sit = _m_plugs.begin(); sit != _m_plugs.end(); )
    {
      (*sit)->saveAll(os);
      if (++sit != _m_plugs.end())
	os << ',' << std::endl;
    }
  os << std::endl << "}" << std::endl;
}

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
void	DataBlock::load(std::istream						&is,
						std::map<int, void *>				&v2real,
						std::map<int, std::list<void **> >	&ref2v,
						std::list<Plug*>					&notSaved)
{
  String	skip;

  DataBlock::removeAll(*this);

  is >> skip;
  is >> _m_idCounter;

  while (skip != "}")
    {
      String	typeStr;
      int		adress;

      is >> typeStr;
      if (typeStr == "}")
		  break;

      Plug	*new_plug = (*Plug::creators()[typeStr])(*this);

      is >> adress;

      v2real[adress] = new_plug;
      new_plug->loadAll(is, v2real, ref2v);
      if (!new_plug->isSaved())
		  notSaved.push_back(new_plug);
      _m_quickMap[new_plug->shortName()] = new_plug;
      _m_plugs.push_back(new_plug);
      is >> skip;
    }
}

// ************************************************************************* //
/// DESCRIPTION: Purges the datablock.
// ************************************************************************* //
void	DataBlock::clear()
{
  DataBlock::removeAll(*this);

  std::list<QWidget *>::iterator	vit;

  for (vit = _m_validBits.begin(); vit != _m_validBits.end(); ++vit)
    (*vit)->close();
  _m_validBits.clear();
  _m_idCounter = 0;
}
