// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Fri Aug 26 11:11:19 NZST 2005
/// DESCRIPTION: 
///  Geometric Octree class (i.e. domain is 3D euclidean space).
///  Template, to store whatever's needed.
///  The octree leaves are not entirely allocated. Use the create/query
///  method to allocate leaves and read leaves from the structure.
// ************************************************************************* //
#ifndef __DATAOCTREE_HH__
#define __DATAOCTREE_HH__

// ************************************************************************* //
#include "gfxtypes.h"
#include <list>

// ************************************************************************* //
#define BITX 0x01
#define BITY 0x02
#define BITZ 0x04

// ************************************************************************* //
//   00000011 are the relevant bit
// (2 << depth) - 1
#define DEFAULTBITS 15
//#define BITS 31

// ************************************************************************* //
template <class T>
class OLeaf;
template <class T>
class OInternal;

// ************************************************************************* //
template <class T>
struct ONode
{
public:
  ONode() {}
  virtual ~ONode() {}
  virtual const OLeaf<T>	*queryRec(Pnt3i &p, unsigned bit) const = 0;
  virtual OLeaf<T>		*createRec(Pnt3i &p, unsigned bit) = 0;
  virtual void	glRender() const = 0;
  virtual void	glRenderLines() const = 0;
};

// ************************************************************************* //
/// DESCRIPTION:
///  p		- position in world coordinates.
///  elm	- stored item.
// ************************************************************************* //
template <class T>
struct OLeaf: public ONode<T>
{
  Pnt3d p;
  T	elm;
public:
  const OLeaf<T>	*queryRec(Pnt3i &, unsigned) const { return 0; }
  OLeaf<T>		*createRec(Pnt3i &, unsigned) { return 0; }
  virtual void	glRender() const { p.glVertex(); }
  virtual void	glRenderLines() const
  { (p - Vec3d(0.1, 0.0, 0.0)).glVertex(),
      (p + Vec3d(0.1, 0.0, 0.0)).glVertex(),
      (p - Vec3d(0.0, 0.1, 0.0)).glVertex(),
      (p + Vec3d(0.0, 0.1, 0.0)).glVertex(),
      (p - Vec3d(0.0, 0.0, 0.1)).glVertex(),
      (p + Vec3d(0.0, 0.0, 0.1)).glVertex(); }
};

// ************************************************************************* //
/// DESCRIPTION: A node considers itself in [0,1]^3
///  Nodes of octree touch the faces [0,1]^2
// ************************************************************************* //
template <class T>
struct OInternal: public ONode<T>
{
  ONode<T>	*children[8];
public:
  OInternal() { memset(children, 0, 8 * sizeof(ONode<T>*)); }
  virtual ~OInternal();
  const OLeaf<T>	*queryRec(Pnt3i &p, unsigned bit) const;
  const OLeaf<T>	*query(const Pnt3d &p, const unsigned bits) const;
  OLeaf<T>		*createRec(Pnt3i &p, unsigned bit);
  OLeaf<T>		*create(const Pnt3d &p, const unsigned bits);
  void			createSphereNeighbors(const Pnt3d &p, const double r2,
					      const Pnt3i &c, unsigned bit,
					      const Mat4d &l2w);
  void			querySphereNeighbors(const Pnt3d &p, const double r2,
					     const Pnt3i &c, unsigned bit,
					     const Mat4d &l2w,
					     std::list<OLeaf<T>*> &ret);
  void	clear();
  virtual void	glRender() const;
  virtual void	glRenderLines() const;
};

// ************************************************************************* //
/// DESCRIPTION: Geometric Octree class (i.e. domain is 3D euclidean space).
// ************************************************************************* //
template <class T>
class Octree
{
public:
  unsigned	_m_bits;
  BBox3d	_m_bbox;
  OInternal<T>	_m_root;
  Mat4d		_m_local2world;
  Mat4d		_m_world2local;
public:
  Octree(const unsigned bits = DEFAULTBITS);
  void	resize(const BBox3d &bbox, const unsigned bits = DEFAULTBITS);
  void	resizeCube(BBox3d bbox, const double minsize);
  void	glRender() const;
  OLeaf<T>	*create(const Pnt3d &p);
  void		create8Neighbors(const Pnt3d &p, OLeaf<T> *leafs[8]);
  void		createSphereNeighbors(const Pnt3d &p, const double radius);
  const OLeaf<T>	*query(const Pnt3d &p) const;
  void		query8Neighbors(const Pnt3d &p, const OLeaf<T> *leafs[8])const;
  void		querySphereNeighbors(const Pnt3d &p, const double radius,
				     std::list<OLeaf<T>*> &ret);
  friend std::ostream &operator<<(std::ostream &os, Octree<T> &) { return os; }
  friend std::istream &operator>>(std::istream &is, Octree<T> &) { return is; }
};

// ************************************************************************* //
/// DESCRIPTION:
///  bits	- relevant bits encoding the position, per dimension.
///		  eg: 2 bits == 00000011 == 0x3
///		      3 bits == 00000111 == 0x7
///		      4 bits == 00001111 == 0xF etc...
// ************************************************************************* //
template <class T>
Octree<T>::Octree(const unsigned bits): _m_bits(bits), _m_bbox(Pnt3d(0.0)),
  _m_local2world(Mat4d::identity()), _m_world2local(Mat4d::identity())
{
}

// ************************************************************************* //
/// DESCRIPTION: Resizes the world coordinates.
///  The data is considered invalid, and cleared.
// ************************************************************************* //
template <class T>
void	Octree<T>::resize(const BBox3d		&bbox,
			  const unsigned	bits)
{
  _m_bits = bits;
  _m_bbox = bbox;

  const Vec3d	diag = (bbox.pmax - bbox.pmin) * (1.0 / double(bits));

  _m_local2world = Mat4d(Vec3d(diag.x, 0.0, 0.0),
			 Vec3d(0.0, diag.y, 0.0),
			 Vec3d(0.0, 0.0, diag.z), bbox.pmin);
  if (1.0e-12 < fabs(_m_local2world.det()))
    _m_world2local = _m_local2world.inverse();
  _m_root.clear();
}

// ************************************************************************* //
/// DESCRIPTION:
/// minsize	- minimum guaranteed size of cell. They may be smaller.
// ************************************************************************* //
template <class T>
void	Octree<T>::resizeCube(BBox3d		bbox,
			      const double	minsize)
{
  bbox.eqCubify();

  const int	nbdivs = int((bbox.pmax.x - bbox.pmin.x) / minsize);
  int	pow2(1);

  while (pow2 < nbdivs)
    pow2 *= 2;
  resize(bbox, max(2, pow2 - 1));
}

// ************************************************************************* //
/// DESCRIPTION: OpenGL rendering.
// ************************************************************************* //
template <class T>
void	Octree<T>::glRender() const
{
  if (fabs(_m_local2world.det()) < 1.0e-6)
    return;

  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);

  glColor3d(0.0, 0.0, 0.0);
  _m_bbox.glRender();

  glColor3d(1.0, 1.0, 0.0);
  glBegin(GL_POINTS);
  _m_root.glRender();
  glEnd();//GL_POINTS

  glColor3d(0.0, 0.0, 1.0);
  glBegin(GL_LINES);
  _m_root.glRenderLines();
  glEnd();//GL_LINES

  glPopAttrib();//GL_LIGHTING_BIT
}

// ************************************************************************* //
template <class T>
const OLeaf<T>	*Octree<T>::query(const Pnt3d &pworld) const
{
  return _m_root.query(_m_world2local * pworld, _m_bits);
}

// ************************************************************************* //
const Vec3d	offsetslocal[8] = { Vec3d(-0.5, -0.5, -0.5),
				    Vec3d(0.5, -0.5, -0.5),
				    Vec3d(-0.5, 0.5, -0.5),
				    Vec3d(0.5, 0.5, -0.5),
				    Vec3d(-0.5, -0.5, 0.5),
				    Vec3d(0.5, -0.5, 0.5),
				    Vec3d(-0.5, 0.5, 0.5),
				    Vec3d(0.5, 0.5, 0.5) };

// ************************************************************************* //
/// DESCRIPTION: If the node didn't exist yet, returns it. Otherwise returns 0.
// ************************************************************************* //
template <class T>
OLeaf<T> *Octree<T>::create(const Pnt3d &pworld)
{
  const Pnt3d	plocal(_m_world2local * pworld);
  Pnt3d		pneighbor = plocal;
  OLeaf<T>	*leaf;

  pneighbor = Pnt3d(bound(pneighbor.x, 0.0, double(_m_bits)),
		    bound(pneighbor.y, 0.0, double(_m_bits)),
		    bound(pneighbor.z, 0.0, double(_m_bits)));
  
  leaf = _m_root.create(pneighbor, _m_bits);
  if (leaf)
    leaf->p = _m_local2world * Pnt3d(nearbyint(pneighbor.x),
				     nearbyint(pneighbor.y),
				     nearbyint(pneighbor.z));
  return leaf;
}

// ************************************************************************* //
/// DESCRIPTION: Creates 8 leaf nodes around inserted point, for 
///  possible linear interpolation in cells.
// ************************************************************************* //
template <class T>
void	Octree<T>::create8Neighbors(const Pnt3d	&pworld,
				    OLeaf<T>	*leafs[8])
{
  const Pnt3d	plocal(_m_world2local * pworld);

  for (unsigned i(0); i < 8; ++i)
    {
      Pnt3d	pneighbor = plocal + offsetslocal[i];

      pneighbor = Pnt3d(bound(pneighbor.x, 0.0, double(_m_bits)),
			bound(pneighbor.y, 0.0, double(_m_bits)),
			bound(pneighbor.z, 0.0, double(_m_bits)));

      leafs[i] = _m_root.create(pneighbor, _m_bits);
      if (leafs[i])
	leafs[i]->p = _m_local2world * Pnt3d(nearbyint(pneighbor.x),
					     nearbyint(pneighbor.y),
					     nearbyint(pneighbor.z));
    }
}

// ************************************************************************* //
/// DESCRIPTION: Creates all the cells that touch a sphere of radius "radius"
//   around point pworld.
// ************************************************************************* //
template <class T>
void	Octree<T>::createSphereNeighbors(const Pnt3d	&pworld,
					 const double	radius)
{
  const Pnt3d	plocal(_m_world2local * pworld);

  _m_root.createSphereNeighbors(plocal, sq(radius), Pnt3i(0),
				(_m_bits + 1) >> 1, _m_local2world);
}

// ************************************************************************* //
/// DESCRIPTION: Queries all the cells that touch a sphere of radius "radius"
//   around point pworld.
// ************************************************************************* //
template <class T>
void	Octree<T>::querySphereNeighbors(const Pnt3d		&pworld,
					const double		radius,
					std::list<OLeaf<T>*>	&ret)
{
  const Pnt3d	plocal(_m_world2local * pworld);

  _m_root.querySphereNeighbors(plocal, sq(radius), Pnt3i(0),
			       (_m_bits + 1) >> 1, _m_local2world, ret);
}

// ************************************************************************* //
/// DESCRIPTION: Finds 8 leaf nodes around queried point.
///  If one of the 8 neighbors does not exist, puts corresponding 
///  0 in array leafs[].
// ************************************************************************* //
template <class T>
void	Octree<T>::query8Neighbors(const Pnt3d		&pworld,
				   const OLeaf<T>	*leafs[8]) const
{
  const Pnt3d	plocal(_m_world2local * pworld);

  for (unsigned i(0); i < 8; ++i)
    {
      Pnt3d	pneighbor = plocal + offsetslocal[i];

      pneighbor = Pnt3d(bound(pneighbor.x, 0.0, double(_m_bits)),
			bound(pneighbor.y, 0.0, double(_m_bits)),
			bound(pneighbor.z, 0.0, double(_m_bits)));
      leafs[i] = _m_root.query(pneighbor, _m_bits);
    }
}

// ************************************************************************* //
template <class T>
OInternal<T>::~OInternal()
{
  clear();
}

// ************************************************************************* //
/// DESCRIPTION: Deletes all children.
// ************************************************************************* //
template <class T>
void	OInternal<T>::clear()
{
  for (unsigned i(0); i < 8; ++i)
    if (children[i])
      {
	delete children[i];
	children[i] = 0;
      }
}

// ************************************************************************* //
/// DESCRIPTION: query for p 
// ************************************************************************* //
template <class T>
OLeaf<T>	*OInternal<T>::createRec(Pnt3i		&p,
					 unsigned	bit)
{
  int	childindex(0);

  if (p.x & bit)
    childindex |= BITX;
  if (p.y & bit)
    childindex |= BITY;
  if (p.z & bit)
    childindex |= BITZ;
  bit >>= 1;
  if (bit == 0)
    if (!children[childindex])
      {
	children[childindex] = new OLeaf<T>;
	return static_cast<OLeaf<T> *>(children[childindex]);
      }
    else
      return 0;
  if (!children[childindex])
    children[childindex] = new OInternal<T>;
  return children[childindex]->createRec(p, bit);
}

// ************************************************************************* //
/// DESCRIPTION: p in [0, 1]^3
// ************************************************************************* //
template <class T>
OLeaf<T>	*OInternal<T>::create(const Pnt3d	&plocal,
				      const unsigned	bits)
{
  Pnt3i		ip((int)(nearbyint(plocal.x)),
		   (int)(nearbyint(plocal.y)),
		   (int)(nearbyint(plocal.z)));
  unsigned	bit((bits + 1) >> 1);

  return createRec(ip, bit);
}

// ************************************************************************* //
/// DESCRIPTION: uses the fact that the distance field of a cube 
///  can be scaled along the three directions of the cube's edges.
// ************************************************************************* //
template <class T>
void	OInternal<T>::createSphereNeighbors(const Pnt3d	  &plocal,
					    const double  r2,
					    const Pnt3i	  &cmin,
					    unsigned	  bit,
					    const Mat4d	  &l2w)
{
  const double	diag(bit);
  const Pnt3i	cmax(cmin.x + bit, cmin.y + bit, cmin.z + bit);

  bit >>= 1;
  for (unsigned i(0); i < 8; ++i)
    {
      const Pnt3i	csj((i & BITX)? cmax.x: cmin.x,
			    (i & BITY)? cmax.y: cmin.y,
			    (i & BITZ)? cmax.z: cmin.z);
      const Pnt3d	csd(csj);
      const BBox3d	bboxi(csd - Vec3d(0.5), csd + Vec3d(diag - 0.5));
      const double	d2 = (l2w * (plocal - bboxi.closest(plocal))).norm2();

      if (d2 < r2)
	if (bit == 0)
	  {
	    if (!children[i])
	      {
		children[i] = new OLeaf<T>;
		static_cast<OLeaf<T>*>(children[i])->p = l2w * csd;
	      }
	  }
	else
	  {
	    if (!children[i])
	      children[i] = new OInternal<T>;
	    static_cast<OInternal<T>*>(children[i])->
	      createSphereNeighbors(plocal, r2, csj, bit, l2w);
	  }
    }
}

// ************************************************************************* //
/// DESCRIPTION: finds all the leaves that exist and intersect the sphere
///  of center l2w * plocal and radius sqrt(r2).
///  To create leaves, use createSphereNeighbors.
// ************************************************************************* //
template <class T>
void	OInternal<T>::querySphereNeighbors(const Pnt3d	  &plocal,
					   const double	  r2,
					   const Pnt3i	  &cmin,
					   unsigned	  bit,
					   const Mat4d	  &l2w,
					   std::list<OLeaf<T>*> &ret)
{
  const double	diag(bit);
  const Pnt3i	cmax(cmin.x + bit, cmin.y + bit, cmin.z + bit);

  bit >>= 1;
  for (unsigned i(0); i < 8; ++i)
    if (children[i])
      {
	const Pnt3i	csj((i & BITX)? cmax.x: cmin.x,
			    (i & BITY)? cmax.y: cmin.y,
			    (i & BITZ)? cmax.z: cmin.z);
	const Pnt3d	csd(csj);
	const BBox3d	bboxi(csd - Vec3d(0.5), csd + Vec3d(diag - 0.5));
	const double	d2 = (l2w * (plocal - bboxi.closest(plocal))).norm2();

	if (d2 < r2)
	  if (bit == 0)
	    ret.push_back(static_cast<OLeaf<T>*>(children[i]));
	  else
	    static_cast<OInternal<T>*>(children[i])->
	      querySphereNeighbors(plocal, r2, csj, bit, l2w, ret);
      }
}

// ************************************************************************* //
/// DESCRIPTION: query for p 
// ************************************************************************* //
template <class T>
const OLeaf<T>	*OInternal<T>::queryRec(Pnt3i		&p,
					unsigned	bit) const
{
  int	childindex(0);

  if (p.x & bit)
    childindex |= BITX;
  if (p.y & bit)
    childindex |= BITY;
  if (p.z & bit)
    childindex |= BITZ;
  bit >>= 1;
  if (children[childindex])
    if (bit == 0)
      return static_cast<OLeaf<T>*>(children[childindex]);
    else
      return children[childindex]->queryRec(p, bit);
  return 0;
}

// ************************************************************************* //
/// DESCRIPTION: p in [0, 1]^3
// ************************************************************************* //
template <class T>
const OLeaf<T>	*OInternal<T>::query(const Pnt3d	  &plocal,
				     const unsigned	  bits) const
{
  Pnt3i		ip((int)(nearbyint(plocal.x)),
		   (int)(nearbyint(plocal.y)),
		   (int)(nearbyint(plocal.z)));
  unsigned	bit((bits + 1) >> 1);

  return queryRec(ip, bit);
}

// ************************************************************************* //
template <class T>
void	OInternal<T>::glRender() const
{
  for (unsigned i(0); i < 8; ++i)
    if (children[i])
      children[i]->glRender();
}

// ************************************************************************* //
template <class T>
void	OInternal<T>::glRenderLines() const
{
  for (unsigned i(0); i < 8; ++i)
    if (children[i])
      children[i]->glRenderLines();
}

// ************************************************************************* //
#endif//__DATAOCTREE_HH__

