// ************************************************************************* //
/// AUTHOR: Alexis Angelidis.
/// STARTED: Wed Sep 14 16:21:01 NZST 2005
// ************************************************************************* //
#ifndef __MATMND_HH__
#define __MATMND_HH__

// ************************************************************************* //
#include <vector>
#include "matnd.h"
#include "gfxtypes.h"

// ************************************************************************* //
class MatMNd;

// ************************************************************************* //
void svdcmp(MatMNd &a, VecNd &w, MatMNd &v);

// ************************************************************************* //
/// DESCRIPTION: MxN-dimensionnal matrices.
// ************************************************************************* //
class MatMNd
{
  unsigned	_m_m, _m_n;
  double	*_m_a;
public:
  MatMNd(): _m_n(0), _m_a(0) {}
  MatMNd(const unsigned m, const unsigned n):
    _m_m(m), _m_n(n), _m_a(new double[m * n]) {}
  MatMNd(const MatMNd &m):
    _m_m(m._m_m), _m_n(m._m_n), _m_a(new double[m.size()])
  { for (unsigned i(0), mxn(size()); i < mxn; ++i) _m_a[i] = m._m_a[i]; }
  virtual ~MatMNd() { if (_m_a) delete[] _m_a; }
  MatMNd	&operator=(const MatMNd &a);
  MatMNd	&operator*=(const double d);
  MatMNd	&operator+=(const MatMNd &a);
  double	operator()(const unsigned i, const unsigned j) const;
  double	&operator()(const unsigned i, const unsigned j);
  friend std::ostream	&operator<<(std::ostream &os, const MatMNd &a);
  friend std::istream	&operator>>(std::istream &is, MatMNd &a);
  void	eqRandom(const double vmax = 1.0, const double vmin = -1.0);
  void	eqNull();
  void	eqInverse(const MatMNd &m);
  void	resize(const unsigned m, const unsigned n);
  void	eqSub(const MatMNd &a, const MatMNd &b);
  void	eqMul(const MatMNd &a, const MatMNd &b);
  void	eqTranspose();
  const unsigned	rows() const { return _m_m; }
  const unsigned	cols() const { return _m_n; }
  const unsigned	size() const { return _m_m * _m_n; }
  void	saveBits(std::ostream &os);
  void	loadBits(std::istream &is);
};

// ************************************************************************* //
inline std::ostream	&operator<<(std::ostream	&os,
				    const MatMNd	&a)
{
  os << a._m_m << ' ' << a._m_n << ' ';
  for (unsigned i(0), mxn(a.size()); i < mxn; ++i)
    os << a._m_a[i] << ' ';
  return os;
}

// ************************************************************************* //
inline void	MatMNd::eqTranspose()
{
  MatMNd	tmp(_m_n, _m_m);

  for (unsigned i(0); i < _m_m; ++i)
    for (unsigned j(0); j < _m_n; ++j)
      tmp(j, i) = operator()(i, j);
  std::swap(_m_a, tmp._m_a);
  std::swap(_m_m, _m_n);
}

// ************************************************************************* //
inline std::istream	&operator>>(std::istream	&is,
				    MatMNd		&a)
{
  unsigned m, n;

  is >> m;
  is >> n;
  a.resize(m, n);
  for (unsigned i(0), mxn(a.size()); i < mxn; ++i)
    is >> a._m_a[i];
  return is;
}

// ************************************************************************* //
inline void	VecNd::eqMul(const MatMNd	&a,
			     const VecNd	&v)
{
  if (a.cols() != v._m_n)
    {
      std::cerr << "Error: incompatible size." << std::endl;
      return;
    }
  resize(a.rows());
  for (unsigned i(0); i < a.rows(); ++i)
    {
      double	&cur = operator()(i);

      cur = 0.0;
      for (unsigned j(0); j < a.cols(); ++j)
	cur += a(i, j) * v(j);
    }
}

// ************************************************************************* //
inline void	VecNd::eqMul(std::vector<Vec3d>	&w,
			     const MatMNd	&a,
			     const VecNd	&v)
{
  w = std::vector<Vec3d>(a.rows() / 3);
  for (unsigned i(0); i < a.rows(); ++i)
    {
      double	&cur = w[i / 3](i % 3);

      cur = 0.0;
      for (unsigned j(0); j < a.cols(); ++j)
	cur += a(i, j) * v(j);
    }
}

// ************************************************************************* //
inline void	MatMNd::resize(const unsigned m,
			       const unsigned n)
{
  if (size() < m * n)
    {
      if (_m_a)
	delete[] _m_a;
      _m_a = new double[m * n];
    }
  _m_m = m;
  _m_n = n;
}

// ************************************************************************* //
inline double	MatMNd::operator()(const unsigned	i,
				   const unsigned	j) const
{
  return _m_a[_m_m * j + i];
}

// ************************************************************************* //
inline double	&MatMNd::operator()(const unsigned	i,
				    const unsigned	j)
{
  return _m_a[_m_m * j + i];
}

// ************************************************************************* //
inline MatMNd &MatMNd::operator=(const MatMNd &a)
{
  resize(a._m_m, a._m_n);
  for (unsigned i(0), mxn(a.size()); i < mxn; ++i)
    _m_a[i] = a._m_a[i];
  return *this;
}

// ************************************************************************* //
inline MatMNd &MatMNd::operator*=(const double d)
{
  for (unsigned i(0), mxn(size()); i < mxn; ++i)
    _m_a[i] *= d;
  return *this;
}

// ************************************************************************* //
inline MatMNd &MatMNd::operator+=(const MatMNd &a)
{
  for (unsigned i(0), mxn(size()); i < mxn; ++i)
    _m_a[i] += a._m_a[i];
  return *this;
}

// ************************************************************************* //
inline void	MatMNd::eqRandom(const double vmax,
				 const double vmin)
{
  for (unsigned i(0), mxn(size()); i < mxn; ++i)
    _m_a[i] = vmin + (vmax - vmin) * rand() / double(RAND_MAX);
}

// ************************************************************************* //
inline void	MatMNd::eqNull()
{
  for (unsigned i(0), mxn(size()); i < mxn; ++i)
    _m_a[i] = 0.0;
}

// ************************************************************************* //
inline void	MatMNd::eqInverse(const MatMNd &m)
{
  MatMNd	ut(m);
  MatMNd	viw(m._m_n, m._m_n);
  VecNd		w(m._m_n);

  svdcmp(ut, w, viw);
  ut.eqTranspose();
  for (unsigned i(0); i < m._m_n; ++i)
    for (unsigned j(0); j < m._m_n; ++j)
      viw(i, j) /= w(j);
  eqMul(viw, ut);
}

// ************************************************************************* //
inline void	MatMNd::eqSub(const MatMNd &a,
			      const MatMNd &b)
{
  resize(a._m_m, a._m_n);
  for (unsigned i(0), mxn(_m_m * _m_n); i < mxn; ++i)
    _m_a[i] = a._m_a[i] - b._m_a[i];
}

// ************************************************************************* //
inline void	MatMNd::saveBits(std::ostream &os)
{
  os.write(reinterpret_cast<const char *>(&_m_m), sizeof(unsigned));
  os.write(reinterpret_cast<const char *>(&_m_n), sizeof(unsigned));
  for (unsigned i(0), mxn(size()); i < mxn; ++i)
    os.write(reinterpret_cast<const char *>(&_m_a[i]), sizeof(double));
}

// ************************************************************************* //
inline void	MatMNd::loadBits(std::istream &is)
{
  unsigned m, n;

  is.read(reinterpret_cast<char *>(&m), sizeof(unsigned));
  is.read(reinterpret_cast<char *>(&n), sizeof(unsigned));
  resize(m, n);
  for (unsigned i(0), mxn(size()); i < mxn; ++i)
    is.read(reinterpret_cast<char *>(&_m_a[i]), sizeof(double));
}

// ************************************************************************* //
inline void	MatMNd::eqMul(const MatMNd &a,
			      const MatMNd &b)
{
  if (a._m_n != b._m_m)
    {
      std::cerr << "Error: incompatible size." << std::endl;
      return;
    }
  resize(a._m_m, b._m_n);
  for (unsigned i(0); i < _m_m; ++i)
    for (unsigned j(0); j < _m_n; ++j)
      {
	operator()(i, j) = 0.0;
	for (unsigned k(0); k < a._m_n; ++k)
	  operator()(i, j) += a(i, k) * b(k, j);
      }
}

// ************************************************************************* //
/// DESCRIPTION: See www.numerical-recipes.com
// ************************************************************************* //
inline void svdcmp(MatMNd	&a,
		   VecNd	&w,
		   MatMNd	&v)
{
}

// ************************************************************************* //
#endif//__MATMND_HH__
