// ************************************************************************* //
/// AUTHOR: Alexis Angelidis.
/// STARTED: Wed Sep 14 16:21:01 NZST 2005
// ************************************************************************* //
#ifndef __MATND_HH__
#define __MATND_HH__

// ************************************************************************* //
#include <math.h>
#include "gfxtypes.h"

// ************************************************************************* //
class VecNd;
class MatNd;
class MatMNd;

// ************************************************************************* //
void svdcmp(MatNd &a, VecNd &w, MatNd &v);

// ************************************************************************* //
std::ostream	&operator<<(std::ostream &os, const VecNd &v);
std::istream	&operator>>(std::istream &is, VecNd &v);
std::ostream	&operator<<(std::ostream &os, const MatNd &m);
std::istream	&operator>>(std::istream &is, MatNd &m);

// ************************************************************************* //
/// DESCRIPTION: N-dimensionnal vectors.
// ************************************************************************* //
class VecNd
{
  friend class MatNd;
  unsigned	_m_n;
  double	*_m_m;
public:
  VecNd(): _m_n(0), _m_m(0) {}
  VecNd(const unsigned n): _m_n(n), _m_m(new double[n]) {}
  VecNd(const VecNd &v): _m_n(v._m_n), _m_m(new double[v._m_n])
  { for (unsigned i(0); i < _m_n; ++i) _m_m[i] = v._m_m[i]; }
  virtual ~VecNd() { if (_m_m) delete[] _m_m; }
  VecNd		&operator=(const VecNd &m);
  double	operator()(const unsigned i) const { return _m_m[i]; }
  double	&operator()(const unsigned i) { return _m_m[i]; }
  unsigned	size() const { return _m_n; }
  friend std::ostream	&operator<<(std::ostream &os, const VecNd &v);
  friend std::istream	&operator>>(std::istream &is, VecNd &v);
  void	eqNull() { for (unsigned i(0); i < _m_n; ++i) _m_m[i] = 0.0; }
  void	eqMul(const MatNd &m, const VecNd &v);
  void	eqMul(const MatMNd &m, const VecNd &v);
  static void	eqMul(std::vector<Vec3d> &w, const MatMNd &a, const VecNd &v);
  void	resize(const unsigned n);
};

// ************************************************************************* //
/// DESCRIPTION: NxN-dimensionnal matrices.
// ************************************************************************* //
class MatNd
{
  unsigned	_m_n;
  double	*_m_m;
public:
  MatNd(): _m_n(0), _m_m(0) {}
  MatNd(const unsigned n): _m_n(n), _m_m(new double[sq(n)]) {}
  MatNd(const MatNd &m): _m_n(m._m_n), _m_m(new double[sq(m._m_n)])
  { for (unsigned i(0), nxn(sq(_m_n)); i < nxn; ++i) _m_m[i] = m._m_m[i]; }
  virtual ~MatNd() { if (_m_m) delete[] _m_m; }
  MatNd &operator=(const MatNd &m);
  MatNd &operator*=(const double d);
  MatNd &operator+=(const MatNd &m);
  double	operator()(const unsigned i, const unsigned j) const;
  double	&operator()(const unsigned i, const unsigned j);
  friend std::ostream	&operator<<(std::ostream &os, const MatNd &m);
  friend std::istream	&operator>>(std::istream &is, MatNd &m);
  void	eqRandom(const double vmax = 1.0, const double vmin = -1.0);
  void	eqNull();
  void	eqIdentity();
  void	eqInverse(const MatNd &m);
  void	resize(const unsigned n);
  void	eqDiagonal(const VecNd &v);
  void	eqSub(const MatNd &a, const MatNd &b);
  void	eqMul(const MatNd &a, const MatNd &b);
  void	eqTranspose();
  const unsigned	rows() const { return _m_n; }
  const unsigned	cols() const { return _m_n; }
  const unsigned	size() const { return _m_n * _m_n; }
};

// ************************************************************************* //
inline std::ostream	&operator<<(std::ostream	&os, 
				    const VecNd		&v)
{
  os << v._m_n << ' ';
  for (unsigned i(0); i < v._m_n; ++i)
    os << v._m_m[i] << ' ';
  return os;
}

// ************************************************************************* //
inline std::istream	&operator>>(std::istream	&is,
				    VecNd		&v)
{
  unsigned	n;

  is >> n;
  v.resize(n);
  for (unsigned i(0); i < v._m_n; ++i)
    is >> v._m_m[i];
  return is;
}

// ************************************************************************* //
inline std::ostream	&operator<<(std::ostream	&os,
				    const MatNd		&m)
{
  os << m._m_n << ' ';
  for (unsigned i(0), nxn(sq(m._m_n)); i < nxn; ++i)
    os << m._m_m[i] << ' ';
  return os;
}

// ************************************************************************* //
inline std::istream	&operator>>(std::istream	&is,
				    MatNd		&m)
{
  unsigned n;

  is >> n;
  m.resize(n);
  for (unsigned i(0), nxn(sq(m._m_n)); i < nxn; ++i)
    is >> m._m_m[i];
  return is;
}

// ************************************************************************* //
inline void	VecNd::eqMul(const MatNd &m,
			     const VecNd &v)
{
  resize(m.rows());
  for (unsigned i(0); i < m.rows(); ++i)
    {
      operator()(i) = 0.0;
      for (unsigned j(0); j < m.cols(); ++j)
	operator()(i) += m(i, j) * v(j);
    }
}

// ************************************************************************* //
inline void	VecNd::resize(const unsigned n)
{
  if (_m_n < n)
    {
      if (_m_m)
	delete[] _m_m;
      _m_m = new double[n];
    }
  _m_n = n;
}

// ************************************************************************* //
inline VecNd	&VecNd::operator=(const VecNd &m)
{
  resize(m._m_n);
  for (unsigned i(0); i < _m_n; ++i)
    _m_m[i] = m._m_m[i];
  return *this;
}

// ************************************************************************* //
inline double	MatNd::operator()(const unsigned	i,
				  const unsigned	j) const
{
  return _m_m[_m_n * j + i];
}

// ************************************************************************* //
inline double	&MatNd::operator()(const unsigned	i,
				   const unsigned	j)
{
  return _m_m[_m_n * j + i];
}

// ************************************************************************* //
inline MatNd &MatNd::operator=(const MatNd &m)
{
  resize(m._m_n);
  for (unsigned i(0), nxn(sq(_m_n)); i < nxn; ++i)
    _m_m[i] = m._m_m[i];
  return *this;
}

// ************************************************************************* //
inline MatNd &MatNd::operator*=(const double d)
{
  for (unsigned i(0), nxn(sq(_m_n)); i < nxn; ++i)
    _m_m[i] *= d;
  return *this;
}

// ************************************************************************* //
inline MatNd &MatNd::operator+=(const MatNd &m)
{
  for (unsigned i(0), nxn(sq(_m_n)); i < nxn; ++i)
    _m_m[i] += m._m_m[i];
  return *this;
}

// ************************************************************************* //
inline void	MatNd::eqRandom(const double vmax,
				const double vmin)
{
  for (unsigned i(0), nxn(sq(_m_n)); i < nxn; ++i)
    _m_m[i] = vmin + (vmax - vmin) * rand() / double(RAND_MAX);
}

// ************************************************************************* //
inline void	MatNd::eqNull()
{
  for (unsigned i(0), nxn(sq(_m_n)); i < nxn; ++i)
    _m_m[i] = 0.0;
}

// ************************************************************************* //
inline void	MatNd::eqIdentity()
{
  eqNull();
  for (unsigned i(0); i < _m_n; ++i)
    operator()(i, i) = 1.0;
}

// ************************************************************************* //
inline void	MatNd::eqInverse(const MatNd &m)
{
  MatNd ut(m);
  MatNd viw(m._m_n);
  VecNd w(m._m_n);

  svdcmp(ut, w, viw);
  ut.eqTranspose();
  for (unsigned i(0); i < m._m_n; ++i)
    for (unsigned j(0); j < m._m_n; ++j)
      viw(i, j) /= w(j);
  resize(m._m_n);
  eqMul(viw, ut);
}

// ************************************************************************* //
inline void	MatNd::resize(const unsigned n)
{
  if (_m_n < n)
    {
      if (_m_m)
	delete[] _m_m;
      _m_m = new double[sq(n)];
    }
  _m_n = n;
}

// ************************************************************************* //
inline void	MatNd::eqDiagonal(const VecNd &v)
{
  resize(v._m_n);
  eqNull();
  for (unsigned i(0); i < _m_n; ++i)
    operator()(i, i) = v(i);
}

// ************************************************************************* //
inline void	MatNd::eqSub(const MatNd &a,
			     const MatNd &b)
{
  resize(a._m_n);
  for (unsigned i(0), nxn(_m_n * _m_n); i < nxn; ++i)
    _m_m[i] = a._m_m[i] - b._m_m[i];
}

// ************************************************************************* //
inline void	MatNd::eqMul(const MatNd &a,
			     const MatNd &b)
{
  resize(a._m_n);
  for (unsigned i(0); i < _m_n; ++i)
    for (unsigned j(0); j < _m_n; ++j)
      {
	operator()(i, j) = 0.0;
	for (unsigned k(0); k < _m_n; ++k)
	  operator()(i, j) += a(i, k) * b(k, j);
      }
}

// ************************************************************************* //
inline void	MatNd::eqTranspose()
{
  for (unsigned i(0); i < _m_n; ++i)
    for (unsigned j(i + 1); j < _m_n; ++j)
      std::swap(operator()(i, j), operator()(j, i));
}

// ************************************************************************* //
inline double norm(const double	a,
		   const double	b)
{
  const double absa = fabs(a);
  const double absb = fabs(b);

  if (absb < absa)
    return absa * sqrt(1.0 + sq(absb / absa));
  else
    return (absb == 0.0 ? 0.0 : absb * sqrt(1.0 + sq(absa / absb)));
}

// ************************************************************************* //
/// DESCRIPTION: See www.numerical-recipes.com
// ************************************************************************* //
inline void svdcmp(MatNd	&a,
		   VecNd	&w,
		   MatNd	&v)
{
}

// ************************************************************************* //
#endif//__MATND_HH__
