// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include "nodebezier3d.h"
#include "bezier_simplex.h"
#include "datablock.h"
#include "plug.h"
#include "gbb.h"

// ************************************************************************* //
int	NBezier3d::typeNbr(65);
String	NBezier3d::typeStr("Bezier3d");

// ************************************************************************* //
/// DESCRIPTION: Front and back can be written in to constaint the curve.
// ************************************************************************* //
void	NBezier3d::initialize()
{
  const Pnt3d	pnts[5] = { Pnt3d(0.0, 0.0, 0.0),
			    Pnt3d(0.0, 0.0, 2.5),
			    Pnt3d(0.0, 0.0, 5.0),
			    Pnt3d(0.0, 0.0, 7.5),
			    Pnt3d(0.0, 0.0, 10.0) };

  DataBlock	&db(getDataBlock());

  NShape::initialize();

  db.addMat4d(Mat4d::identity(), "matParent", false);//position
  db.addInt(5, 3, 256, "n");

  db.addPnt3d(pnts[0], "front");
  db.addPnt3d(pnts[1], "following");
  db.addPnt3d(pnts[3], "preceding");
  db.addPnt3d(pnts[4], "back");
  db.addPnt3d(pnts[4] + (pnts[4] - pnts[3]), "extrapolate");

  db.addPnt3dArray(5, Pnt3d(0.0), "ci", false);
  db.addDouble(0.0, 0.0, MAXDOUBLE, "u", false);//uParameter
  db.addDouble(0.0, 0.0, MAXDOUBLE, "l", false);//lArcLength
  db.addDouble(0.0, 0.0, MAXDOUBLE, "ul", false);//uParameterOfArcLength
  db.addPnt3d(Pnt3d(0.0), "pu", false);//paramPointLocal
  db.addPnt3d(Pnt3d(0.0), "pl", false);//arclenghtPointLocal

  db.addVec3d(Vec3d(0.0), "tu", false);//paramTangentLocal
  db.addVec3d(Vec3d(0.0), "tl", false);//arcLengthTangentLocal
  db.addDouble(0.0, 0.0, MAXDOUBLE, "ltot", false);//totArcLength
  db.addSignal("signal", false);//signal

  db.affect("n", "ci");
  db.affect("front", "ci");
  db.affect("following", "ci");
  db.affect("preceding", "ci");
  db.affect("back", "ci");

  db.affect("matParent", "front");
  db.affect("matParent", "following");
  db.affect("matParent", "preceding");
  db.affect("matParent", "back");

  db.affect("preceding", "extrapolate");
  db.affect("back", "extrapolate");

  db.affect("l", "ul");
  db.affect("ci", "ul");
  db.affect("ci", "pu");
  db.affect("u", "pu");
  db.affect("ci", "pl");
  db.affect("l", "pl");

  db.affect("ci", "tu");
  db.affect("u", "tu");

  db.affect("pl", "tl");
  db.affect("ci", "tl");
  db.affect("l", "tl");

  db.affect("ci", "ltot");

  for (int i(0); i < 5; ++i)
    db.writeAsPnt3dArray("ci", i, pnts[i]);

  buildDagNodes();
}

// ************************************************************************* //
void	NBezier3d::buildDagNodes()
{
  DagNode::connect(*_m_parent, "renderIn", *this, "renderIn");
}

// ************************************************************************* //
void	NBezier3d::glRender(Plug &plugOut)
{
  const int	mode = readAsInt("renderIn");

  plugOut.writeAsSignal();

  const int	sel(readAsInt("sel"));
  Vec3d	color;
  Plug	*hndci(getPlug("ci"));
  const std::vector<Pnt3d>	&ci(readAsPnt3dArray("ci"));

  Plug	*hndFront(getPlug("front"));
  Plug	*hndBack(getPlug("back"));
  Plug	*hndFollowing(getPlug("following"));
  Plug	*hndPreceding(getPlug("preceding"));

  if (sel == 1)
    color = Vec3d(1.0, 0.0, 0.5);
  else if (sel)
    color = Vec3d(1.0, 0.0, 0.0);
  else
    color = Vec3d(0.0, 0.0, 1.0);
  color.glColor();

  pushName();

  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);

  glPushMatrix();
  readAsMat4d("matParent").glMultMatrix();
  hndci->pushName();
  if (sel == 0 || mode != Gbb::Selecting)
    {
      glBegin(GL_LINE_STRIP);
      for (int i(0), n(16); i <= n; ++i)
	{
	  writeAsDouble("u", i / double(n));
	  readAsPnt3d("pu").glVertex();
	}
      glEnd();//GL_LINE_STRIP
    }
  for (int i(2), size(ci.size() - 2); i < size; ++i)
    {
      glPushName(i);//i
      glPushAttrib(GL_CURRENT_BIT);
      glBegin(GL_POINTS);
      if (hndci->isSelected(i))
	glColor3d(1.0, 1.0, 0.0);
      ci[i].glVertex();
      glEnd();//GL_POINTS
      glPopAttrib();//GL_CURRENT_BIT
      glPopName();//i
    }
  hndci->popName();

  glPopMatrix();

  hndFront->pushName();
  glPushAttrib(GL_CURRENT_BIT);
  glBegin(GL_POINTS);
  if (hndFront->isSelected())
    glColor3d(1.0, 1.0, 0.0);
  hndFront->readAsPnt3d().glVertex();
  glEnd();//GL_POINTS
  glPopAttrib();//GL_CURRENT_BIT
  hndFront->popName();

  hndBack->pushName();
  glPushAttrib(GL_CURRENT_BIT);
  glBegin(GL_POINTS);
  if (hndBack->isSelected())
    glColor3d(1.0, 1.0, 0.0);
  hndBack->readAsPnt3d().glVertex();
  glEnd();//GL_POINTS
  glPopAttrib();//GL_CURRENT_BIT
  hndBack->popName();

  hndFollowing->pushName();
  glPushAttrib(GL_CURRENT_BIT);
  glBegin(GL_POINTS);
  if (hndFollowing->isSelected())
    glColor3d(1.0, 1.0, 0.0);
  hndFollowing->readAsPnt3d().glVertex();
  glEnd();//GL_POINTS
  glPopAttrib();//GL_CURRENT_BIT
  hndFollowing->popName();

  hndPreceding->pushName();
  glPushAttrib(GL_CURRENT_BIT);
  glBegin(GL_POINTS);
  if (hndPreceding->isSelected())
    glColor3d(1.0, 1.0, 0.0);
  hndPreceding->readAsPnt3d().glVertex();
  glEnd();//GL_POINTS
  glPopAttrib();//GL_CURRENT_BIT
  hndPreceding->popName();

  glPopAttrib();//GL_LIGHTING_BIT
  popName();
}

// ************************************************************************* //
void	NBezier3d::compute(Plug &plugOut)
{
  if (plugOut.shortName() == "pu")
    {
      const std::vector<Pnt3d>	&ci(readAsPnt3dArray("ci"));
      const double	u(readAsDouble("u"));
      const double	ut[2] = { u, 1.0 - u };
      const int		n(readAsInt("n"));

      plugOut.writeAsPnt3d(bezierSimplex<2>(ut, ci, n - 1));
    }
  else if (plugOut.shortName() == "pl")
    {
      const std::vector<Pnt3d>	&ci(readAsPnt3dArray("ci"));
      const double	ul(readAsDouble("ul"));
      const double	ut[2] = { ul, 1.0 - ul };
      const int		n(readAsInt("n"));

      plugOut.writeAsPnt3d(bezierSimplex<2>(ut, ci, n - 1));
    }
  else if (plugOut.shortName() == "ul")
    {
      const double		l(readAsDouble("l"));
      const std::vector<Pnt3d>	&ci(readAsPnt3dArray("ci"));
      double	lcur(0.0), lbit, u0, u1, ut[2] = { 0.0, 1.0 };
      Pnt3d	p0, p1;
      const int		n(readAsInt("n"));

      p1 = bezierSimplex<2>(ut, ci, n - 1);
      u1 = 0.0;
      for (int i(0), m(16); i < m; ++i, lcur += lbit)
	{
	  u0 = u1;
	  u1 = i / (m - 1.0);
	  ut[0] = u1;
	  ut[1] = 1.0 - u1;
	  p0 = p1;
	  p1 = bezierSimplex<2>(ut, ci, n - 1);
	  lbit = (p1 - p0).norm();
	  if (l < lcur + lbit)
	    break;
	}
      plugOut.writeAsDouble(u0 + (u1 - u0) * (l - lcur) / lbit);
    }
  else if (plugOut.shortName() == "tu")
    {
      const std::vector<Pnt3d>	&ci(readAsPnt3dArray("ci"));
      const double	u(readAsDouble("u"));
      const double	ut1[2] = { u + 1.0e-6, 1.0 - (u + 1.0e-6)};
      const int		n(readAsInt("n"));
      const Pnt3d	p1 = bezierSimplex<2>(ut1, ci, n - 1);
      const Pnt3d	p0 = readAsPnt3d("pu");

      plugOut.writeAsVec3d((p1 - p0).unit());
    }
  else if (plugOut.shortName() == "tl")
    {
      const std::vector<Pnt3d>	&ci(readAsPnt3dArray("ci"));
      const double	u(readAsDouble("ul"));
      const double	ut1[2] = { (u + 1.0e-6), 1.0 - (u + 1.0e-6)};
      const int		n(readAsInt("n"));
      const Pnt3d	p1 = bezierSimplex<2>(ut1, ci, n - 1);
      const Pnt3d	p0 = readAsPnt3d("pl");

      plugOut.writeAsVec3d((p1 - p0).unit());
    }
  else if (plugOut.shortName() == "renderOut")
    glRender(plugOut);
  else if (plugOut.shortName() == "ltot")
    {
      Pnt3d	pnts[16];
      double	ltot(0.0);

      for (int i(0), n(16); i < n; ++i)
	{
	  writeAsDouble("u", i / (n - 1.0));
	  pnts[i] = readAsPnt3d("pu");
	}
      for (int i(1), n(16); i < n; ++i)
	ltot += (pnts[i] - pnts[i - 1]).norm();
      plugOut.writeAsDouble(ltot);
    }
  else if (plugOut.shortName() == "ci")
    {
      const int	n = readAsInt("n");
      const std::vector<Pnt3d>	ci(readAsPnt3dArray("ci"));
      const int	oldn = ci.size();
      const Pnt3d	&front = readAsPnt3d("front");
      const Pnt3d	&back = readAsPnt3d("back");
      const Pnt3d	&following = readAsPnt3d("following");
      const Pnt3d	&preceding = readAsPnt3d("preceding");
      const Mat4d	m = readAsMat4d("matParent");
      const Mat4d	im = m.inverse();

      if (oldn != n)
	{
	  plugOut.resize(n);
	  for (int i(1); i < n - 1; ++i)
	    {
	      const double	u(i / (n - 1.0));
	      const double	ut[2] = { u, 1.0 - u };

	      // TODO: sample control polygon, not curve.
	      plugOut.writeAsPnt3dArray(i, bezierSimplex<2>(ut, ci, oldn - 1));
	    }
	}
      plugOut.writeAsPnt3dArray(0, im * front);
      plugOut.writeAsPnt3dArray(1, im * following);
      plugOut.writeAsPnt3dArray(n - 2, im * preceding);
      plugOut.writeAsPnt3dArray(n - 1, im * back);
      plugOut.writeAsSignal();
    }
  else if (plugOut.shortName() == "front")
    plugOut.writeAsPnt3d(readAsMat4d("matParent") *
			 readAsPnt3dArray("ci").front());
  else if (plugOut.shortName() == "back")
    plugOut.writeAsPnt3d(readAsMat4d("matParent") *
			 readAsPnt3dArray("ci").back());
  else if (plugOut.shortName() == "preceding")
    {
      const std::vector<Pnt3d>	ci(readAsPnt3dArray("ci"));

      plugOut.writeAsPnt3d(readAsMat4d("matParent") * ci[ci.size() - 2]);
    }
  else if (plugOut.shortName() == "following")
    plugOut.writeAsPnt3d(readAsMat4d("matParent") * 
			 readAsPnt3dArray("ci")[1]);
  else if (plugOut.shortName() == "extrapolate")
    {
      const Pnt3d	&back = readAsPnt3d("back");
      const Pnt3d	&preceding = readAsPnt3d("preceding");

      plugOut.writeAsPnt3d(back + (back - preceding));
    }
}
