// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <GL/glu.h>
#include "nodecamera.h"
#include "datablock.h"
#include "gbb.h"

// ************************************************************************* //
int	NCamera::typeNbr(53);
String	NCamera::typeStr("Camera");

// ************************************************************************* //
int	PlugFrustum::typeNbr(67);
String	PlugFrustum::typeStr("PlugFrustum");

// ************************************************************************* //
/// DESCRIPTION: Draws a circle of radius r in the plane z = 0.
// ************************************************************************* //
void	glCircle(const GLdouble	r,
		 const GLint	pieces)
{
  glBegin(GL_LINE_LOOP);
  for (GLint i(0); i < pieces; ++i)
    {
      const GLdouble	a = 2.0 * i * M_PI / pieces;

      glVertex3d(cos(a) * r, sin(a) * r, 0.0);
    }
  glEnd();//GL_LINE_LOOP
}

// ************************************************************************* //
/// DESCRIPTION:
///  eye	- center of camera.
///  dir	- viewing direction.
///  up		- up vector, orthogonal to viewing direction.
///  model	- modelview matrix.
///  frustum	- camera frustum.
///  zNear	- closest clipping plane.
///  zFar	- farthest clipping plane.
// ************************************************************************* //
void	NCamera::initialize()
{
  NShape::initialize();

  DataBlock	&db(getDataBlock());

  db.addMat4d(Mat4d::identity(), "matParent", false);//position
  db.addMat4d(Mat4d::identity(), "imatParent", false);//position inverse

  db.addInt(0, MININT, MAXINT, "f", false);
  db.addDouble(0.0, 0.0, MAXDOUBLE, "dt", false);
  db.addSignal("update", false);
  db.addSignal("signal", false);

  db.addPnt3d(Pnt3d(0.0), "eye");
  db.addVec3d(Vec3d(0.0), "dir");
  db.addVec3d(Vec3d(0.0, 1.0, 0.0), "up");
  db.addMat4d(Mat4d::identity(), "model", false);
  db.addMat4d(Mat4d::identity(), "proj", false);
  db.addMat2i(Mat2i(0), "view", false);
  db.addCustom(*new PlugFrustum(db), "frustum", false);
  db.addDouble(0.0, 0.0, MAXDOUBLE, "zNear");
  db.addDouble(0.0, 0.0, MAXDOUBLE, "zFar");
  db.addDouble(0.5, 0.0, MAXDOUBLE, "size");
  db.addBool(false, "rendering", false);
  db.addBool(true, "isVisible", false);
  db.addInt(320, MININT, MAXINT, "width");
  db.addInt(240, MININT, MAXINT, "height");
  db.addSignal("constraintAim", false);

  db.affect("matParent", "imatParent");
  db.affect("matParent", "model");
  db.affect("model", "eye");
  //db.affect("model", "renderOut");

  buildDagNodes();
}

// ************************************************************************* //
/// DESCRIPTION: Creates necessary nodes.
// ************************************************************************* //
void	NCamera::buildDagNodes()
{
  DagNode	*oldParent(_m_parent);
  DagNode	*tr = Command::run(String("new -type Transform -path :"));
  DagNode	*clock = DagNode::getNodeByPath(":Clock0");

  if (oldParent)
    DagNode::parent(*oldParent, *tr);
  DagNode::parent(*tr, *this);
  DagNode::connect(*oldParent, "renderIn", *tr, "renderIn");
  DagNode::connect(*tr, "renderIn", *this, "renderIn");
  DagNode::connect(*tr, "matChild", *this, "matParent");
  DagNode::connect(*clock, "f", *this, "f");
  DagNode::connect(*clock, "dt", *this, "dt");
}

// ************************************************************************* //
void	NCamera::glRender(Plug &plugOut)
{
  if (getInstStr() != "Camera0")
  if (readAsInt("renderIn") != Gbb::Rendering &&
      readAsBool("isVisible"))
    {
      const double		size(readAsDouble("size"));
      static GLUquadric	*quad = gluNewQuadric();
      const int		polys(24);
      const Mat4d		&model = readAsMat4d("model");

      pushName();

      glPushMatrix();
      model.inverse().glMultMatrix();
      glScaled(size, size, size);

      const BBox3d	bbox(Pnt3d(-0.5, -1.0, -1.5),
			     Pnt3d(0.5, 1.0, 1.5));
      const int	sel(readAsInt("sel"));

      if (sel == 1)
	glColor4d(1.0, 0.0, 0.5, 0.2);
      else if (sel)
	glColor4d(1.0, 0.5, 0.0, 0.2);
      else
	glColor4d(0.0, 0.25, 1.0, 0.2);
      //glColor4d(0.0, 0.8, 0.2, 0.2);

      // Shape ************************************************************* //
      glEnable(GL_POLYGON_OFFSET_FILL);
      glEnable(GL_POLYGON_OFFSET_LINE);
      glPolygonOffset(1.0, 1.0);

      // Box *************************************************************** //
      bbox.glRenderSmooth();

      // Films ************************************************************* //
      glPushMatrix();
      glTranslated(0.0, 2.0, 0.0);
      glRotated(90.0, 0.0, 1.0, 0.0);
      glTranslated(-1.0, 0.0, 0.0);

      // Film back ********************************************************* //
      glPushMatrix();
      glTranslated(0.0, 0.0, -0.25);
      glPushMatrix();
      glRotated(180.0, 1.0, 0.0, 0.0);
      gluDisk(quad, 0.0, 1.0, polys, 1);
      glPopMatrix();
      gluCylinder(quad, 1.0, 1.0, 0.5, polys, 1);
      glTranslated(0.0, 0.0, 0.5);
      gluDisk(quad, 0.0, 1.0, polys, 1);
      glPopMatrix();

      glTranslated(2.0, 0.0, 0.0);
      // Film front ******************************************************** //
      glPushMatrix();
      glTranslated(0.0, 0.0, -0.25);
      glPushMatrix();
      glRotated(180.0, 1.0, 0.0, 0.0);
      gluDisk(quad, 0.0, 1.0, polys, 1);
      glPopMatrix();
      gluCylinder(quad, 1.0, 1.0, 0.5, polys, 1);
      glTranslated(0.0, 0.0, 0.5);
      gluDisk(quad, 0.0, 1.0, polys, 1);
      glPopMatrix();
      glPopMatrix();

      // Lens ************************************************************** //
      glPushMatrix();
      glTranslated(0.0, 0.0, -1.5);
      glTranslated(0.0, 0.0, -0.65);
      gluCylinder(quad, 0.7, 0.5, 0.65, polys, 1);
      glRotated(180.0, 1.0, 0.0, 0.0);
      glColor4d(1.0, 1.0, 1.0, 0.2);
      gluDisk(quad, 0.0, 0.7, polys, 1);
      glPopMatrix();

      glDisable(GL_POLYGON_OFFSET_LINE);
      glDisable(GL_POLYGON_OFFSET_FILL);

      // WIRE
      glColor4d(0.0, 0.0, 0.0, 1.0);

      glPushAttrib(GL_LIGHTING_BIT);
      glDisable(GL_LIGHTING);

      // Box *************************************************************** //
      bbox.glRender();

      // Films ************************************************************* //
      glPushMatrix();
      glTranslated(0.0, 2.0, 0.0);
      glRotated(90.0, 0.0, 1.0, 0.0);
      glTranslated(-1.0, 0.0, 0.0);

      // Film back ********************************************************* //
      glPushMatrix();
      glTranslated(0.0, 0.0, -0.25);
      glCircle(1.0, polys);
      glTranslated(0.0, 0.0, 0.5);
      glCircle(1.0, polys);
      glPopMatrix();

      glTranslated(2.0, 0.0, 0.0);
      // Film front ******************************************************** //
      glPushMatrix();
      glTranslated(0.0, 0.0, -0.25);
      glCircle(1.0, polys);
      glTranslated(0.0, 0.0, 0.5);
      glCircle(1.0, polys);
      glPopMatrix();
      glPopMatrix();

      // Lens ************************************************************** //
      glPushMatrix();
      glTranslated(0.0, 0.0, -1.5);
      glCircle(0.5, polys);
      glTranslated(0.0, 0.0, -0.65);
      glPushAttrib(GL_LINE_BIT);
      glLineWidth(2.0);
      glCircle(0.7, polys);
      glPopAttrib();//GL_LINE_BIT

      glPopMatrix();

      glPopAttrib();//GL_LIGHTING_BIT
      glPopMatrix();

      popName();
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
/// DESCRIPTION: GBB compute.
// ************************************************************************* //
void	NCamera::compute(Plug &plugOut)
{
  if (plugOut.shortName() == "renderOut")
    glRender(plugOut);
  else if (plugOut.shortName() == "imatParent")
    plugOut.writeAsMat4d(readAsMat4d("matParent").inverse());
  else if (plugOut.shortName() == "model")
    plugOut.writeAsMat4d(readAsMat4d("matParent").inverse());
  else if (plugOut.shortName() == "eye")
    plugOut.writeAsPnt3d(readAsMat4d("model").inverse() * Pnt3d(0.0));
}
