// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <GL/glu.h>
#include "nodenoise.h"
#include "datablock.h"
#include "plugcustom.h"
#include "command.h"
#include "gbb.h"

// ************************************************************************* //
int	NNoiseGenerator::typeNbr(62);
String	NNoiseGenerator::typeStr("Noise");

// ************************************************************************* //
int	PlugNoiseVortices::typeNbr(70);
String	PlugNoiseVortices::typeStr("NoiseVortices");

// ************************************************************************* //
int	PlugOctreeNoise::typeNbr(77);
String	PlugOctreeNoise::typeStr("PlugOctreeNoise");

// ************************************************************************* //
void	NNoiseGenerator::initialize()
{
  NShape::initialize();

  DataBlock	&db(getDataBlock());

  db.addInt(0, MININT, MAXINT, "f", false);
  db.addInt(0, MININT, MAXINT, "fPrev", false);
  db.addDouble(0.0, 0.0, MAXDOUBLE, "dt", false);
  db.addSignal("update", false);
  db.addMat4d(Mat4d::identity(), "matParent", false);//position
  db.addMat4d(Mat4d::identity(), "imatParent", false);//position inverse
  db.addPnt3d(Pnt3d(0.0), "p", false);
  db.addTwistLog(TwistLog(0.0), "rlog", false);
  db.addCustom(*new PlugNoiseVortices(db), "nv", false);
  db.addCustom(*new PlugBBox3d(db), "bbox", false);
  db.addCustom(*new PlugOctreeNoise(db), "octree", false);
  db.addDouble(1.0, 0.0001, MAXDOUBLE, "octreeResolution");

  db.addInt(12, 1, MAXINT, "mod");// FREQ OF SAMPLES
  db.addDouble(3.0, 0.0, MAXDOUBLE, "radius");
  db.addDouble(1.0, 0.0, MAXDOUBLE, "strength");
  db.addBool(true, "showNoise");
  db.addDouble(0.5, 0.0, MAXDOUBLE, "ttb");// Time to birth
  db.addDouble(7.0, 0.0, MAXDOUBLE, "ttl");// Time to live
  db.addDouble(8.0, 0.0, MAXDOUBLE, "ttd");// Time to die

  db.affect("f", "update");
  db.affect("p", "rlog");
  db.affect("showNoise", "renderOut");
  db.affect("nv", "bbox");
  db.affect("bbox", "octree");
  db.affect("matParent", "imatParent");
  buildDagNodes();
}

// ************************************************************************* //
void	NNoiseGenerator::buildDagNodes()
{
  DagNode	*oldParent(_m_parent);
  DagNode	*tr = Command::run(String("new -type Transform -path :"));
  DagNode	*clock = DagNode::getNodeByPath(":Clock0");

  if (oldParent)
    DagNode::parent(*oldParent, *tr);

  DagNode::parent(*tr, *this);
  DagNode::connect(*oldParent, "renderIn", *tr, "renderIn");
  DagNode::connect(*tr, "renderIn", *this, "renderIn");
  DagNode::connect(*tr, "matChild", *this, "matParent");
  DagNode::connect(*clock, "f", *this, "f");
  DagNode::connect(*clock, "dt", *this, "dt");
}

// ************************************************************************* //
void	NNoiseGenerator::glRender(Plug &plug)
{
  if (readAsInt("renderIn") != Gbb::Rendering)
    {
      std::list<NoiseVortex>::iterator nvit;
      std::list<NoiseVortex> *nv = 
	static_cast<std::list<NoiseVortex> *>(readAsCustom("nv"));
      const int	sel(readAsInt("sel"));
      Vec3d		col;
      const Mat4d	&imatParent(readAsMat4d("imatParent"));

      pushName();
      glPushMatrix();
      readAsMat4d("matParent").glMultMatrix();

      if (sel == 1)
	col = Vec3d(1.0, 0.0, 0.5);
      else if (sel)
	col = Vec3d(1.0, 0.5, 0.0);
      else
	col = Vec3d(0.0, 0.0, 1.0);
      col.glColor();

      glPushAttrib(GL_LIGHTING_BIT);
      glDisable(GL_LIGHTING);
      glBegin(GL_LINES);
      glVertex3d(-1.0, 0.0, 0.0);
      glVertex3d(1.0, 0.0, 0.0);
      glVertex3d(0.0, -1.0, 0.0);
      glVertex3d(0.0, 1.0, 0.0);
      glVertex3d(0.0, 0.0, -1.0);
      glVertex3d(0.0, 0.0, 1.0);
      glEnd();//GL_LINES
      glPopAttrib();//GL_LIGHTING_BIT

      if (readAsBool("showNoise"))
	{
	  const Octree<NoiseVortex*> &octree = 
	    *static_cast<const Octree<NoiseVortex*>*>(readAsCustom("octree"));

	  glPushMatrix();
	  imatParent.glMultMatrix();
	  for (nvit = nv->begin(); nvit != nv->end(); ++nvit)
	    nvit->glRender();
	  //octree.glRender();
	  glPopMatrix();
	}
      glPopMatrix();
      popName();
    }
  plug.writeAsSignal();
}

// ************************************************************************* //
/// DESCRIPTION: Places noise vortices in world coordinates
// ************************************************************************* //
void	NNoiseGenerator::update(Plug &plugOut)
{
  DataBlock	&db(getDataBlock());
  std::list<NoiseVortex> *nv = 
    static_cast<std::list<NoiseVortex> *>(readAsCustom("nv"));
  Plug		*fPrevData = db.getPlug("fPrev");
  const int	fPrev = fPrevData->readAsInt();
  const int	f = readAsInt("f");
  static unsigned	seedp(0);

  if (f == 0 || f < fPrev)
    {
      seedp = 0;
      nv->clear();
    }
  else if (!(f % readAsInt("mod")))
    {
      const double	radius(readAsDouble("radius"));
      const Mat4d	&matParent = readAsMat4d("matParent");

      nv->push_back(NoiseVortex(matParent * Pnt3d(0.0).jitter(1.0, &seedp),
					Vec3d(0.0).jitter(1.0, &seedp).unit(),
					radius, int(4.0 * (rand_r(&seedp) - 1.0) /
					RAND_R_MAX)));
    }

  std::list<NoiseVortex>::iterator nvit;
  const double	dt(readAsDouble("dt"));

  DagNode	*rs = DagNode::getNodeByPath(":VortexSet0");//
  TwistLog	tlog;
  Mat3d		jac;
  Plug		&plugp(*rs->getPlug("ptotal"));
  Plug		&plugRlog(*rs->getPlug("rlogTotal"));
  Plug		&plugJac(*rs->getPlug("jacTotal"));
  const double	ttd(readAsDouble("ttd"));

  if (fPrev < f)
    for (nvit = nv->begin(); nvit != nv->end(); )
      if (ttd < (nvit->t += dt))
	nv->erase(nvit++);
      else
	{
	  plugp.writeAsPnt3d(nvit->c);
	  tlog = plugRlog.readAsTwistLog();
	  jac = plugJac.readAsMat3d();
	  (tlog * dt).exp(nvit->c);
	  nvit->n = (jac * nvit->n).unit();
	  ++nvit;
	  // TODO: compute presence!
	}
  fPrevData->writeAsInt(f);
  writeAsSignal("nv");
  plugOut.writeAsSignal();
}

// ************************************************************************* //
/// DESCRIPTION: According to the particle's clock, returns the presence 
///  of the particle in [0, 1]. 1 is present, 0 is absent.
// ************************************************************************* //
double	NoiseVortex::presence(const double	ttb,
			      const double	ttl,
			      const double	ttd) const
{
  if (t < 0.0 || ttd < t)
    return 0.0;
  if (ttl < t) // fade out particle before disapearing.
    {
      const double	l((t - ttl) / (ttd - ttl));

      //return 1.0 + l * l * l * (l * (15.0 - 6.0 * l) - 10.0);
      return 1.0 - l * l * (3.0 - 2.0 * l);
    }
  if (t < ttb)
    {
      const double	l(t / ttb);

      //return -l * l * l * (l * (15.0 - 6.0 * l) - 10.0);
      return l * l * (3.0 - 2.0 * l);
    }
  return 1.0;
}

// ************************************************************************* //
/// DESCRIPTION: The list of Noise vortices are pre-stored in the octree
///  cells. An simple octree query provides all the potentially influent
///  vortices.
// ************************************************************************* //
void	NNoiseGenerator::computeRlog(Plug &plugOut)
{
  TwistLog	tlog(0.0);
  const Pnt3d	&p(readAsPnt3d("p"));
  const BBox3d	&bbox = *static_cast<BBox3d *>(readAsCustom("bbox"));
  const OLeaf<std::vector<NoiseVortex *> >	*leaf;
  const Octree<std::vector<NoiseVortex*> >	&octree = 
    *static_cast<Octree<std::vector<NoiseVortex*> >*>(readAsCustom("octree"));

  if (bbox.contains(p) && (leaf = octree.query(p)))
    {
      const std::vector<NoiseVortex*>	&l = leaf->elm;
      std::vector<NoiseVortex*>::const_iterator	lit;
      const double	strength(readAsDouble("strength"));
      const double	ttb(readAsDouble("ttb"));
      const double	ttl(readAsDouble("ttl"));
      const double	ttd(readAsDouble("ttd"));

      for (lit = l.begin(); lit != l.end(); ++lit)
	tlog += (*lit)->tlog(p) * (strength * (*lit)->presence(ttb, ttl, ttd));
    }
  plugOut.writeAsTwistLog(tlog);
}

// ************************************************************************* //
void	NNoiseGenerator::compute(Plug &plugOut)
{
  if (plugOut.shortName() == "rlog")
    computeRlog(plugOut);
  else if (plugOut.shortName() == "update")
    {
      update(plugOut);

      // Reset octree ****************************************************** //
      Octree<std::vector<NoiseVortex*> >	&octree = 
	*static_cast<Octree<std::vector<NoiseVortex*> >*>
	(readAsCustom("octree"));
      const BBox3d	&bbox = *static_cast<BBox3d *>(readAsCustom("bbox"));
      const double	radius(readAsDouble("radius"));
      std::list<NoiseVortex>::iterator	nvit;
      std::list<NoiseVortex>	*nv = 
	static_cast<std::list<NoiseVortex> *>(readAsCustom("nv"));

      octree.resizeCube(bbox, readAsDouble("octreeResolution"));
      for (nvit = nv->begin(); nvit != nv->end(); ++nvit)
	{
	  std::list<OLeaf<std::vector<NoiseVortex*> >*>	ret;
	  std::list<OLeaf<std::vector<NoiseVortex*> >*>::iterator	rit;

	  octree.createSphereNeighbors(nvit->c, radius);
	  octree.querySphereNeighbors(nvit->c, radius, ret);
	  for (rit = ret.begin(); rit != ret.end(); ++rit)
	    (*rit)->elm.push_back(&(*nvit));
	}
    }
  else if (plugOut.shortName() == "renderOut")
    {
      readAsSignal("update");
      glRender(plugOut);
    }
  else if (plugOut.shortName() == "bbox")
    {
      std::list<NoiseVortex>::iterator nvit;
      std::list<NoiseVortex> *nv = 
	static_cast<std::list<NoiseVortex> *>(readAsCustom("nv"));
      BBox3d	bbox(0.0);
      
      if (!nv->empty())
	{
	  bbox = BBox3d(nv->front().c);
	  for (nvit = nv->begin(), ++nvit; nvit != nv->end(); ++nvit)
	    bbox += nvit->c;
	  bbox.eqExpand(readAsDouble("radius"));
	}
      plugOut.writeAsCustom(&bbox);
    }
  else if (plugOut.shortName() == "imatParent")
    plugOut.writeAsMat4d(readAsMat4d("matParent").inverse());
}

// ************************************************************************* //
void	PlugNoiseVortices::saveData(std::ostream &os)
{
  std::list<NoiseVortex>::iterator	dit;

  os << _m_data->size() << ' ';
  if (_m_data)
    for (dit = _m_data->begin(); dit != _m_data->end(); ++dit)
      os << *dit << ' ';
}

// ************************************************************************* //
void	PlugNoiseVortices::loadData(std::istream &is)
{
  int	size;

  is >> size;
  if (_m_data)
    for (int i(0); i < size; ++i)
      {
	_m_data->push_back(NoiseVortex());
	is >> _m_data->back();
      }
}

// ************************************************************************* //
double	rbfNoise(const Pnt3d	&p,
		 const Pnt3d	&c,
		 const double	r2)
{
  const double	l2 = (c - p).norm2();

  return (r2 < l2)? 0.0: sq(4.0 - 20.0 / (l2 / r2 + 4.0));
}

// ************************************************************************* //
TwistLog	NoiseVortex::tlog(const Pnt3d &p) const
{
  const double	d(rbfNoise(p, c, sq(r)));

  if (d == 0.0)
    return TwistLog(0.0);
  return TwistLog(n, Vec3d(c) ^ n) * d;
}

// ************************************************************************* //
void	NoiseVortex::glRender() const
{
  if (type == 0)
    glColor3d(1.0, 0.0, 0.0);
  else if (type == 1)
    glColor3d(0.0, 1.0, 0.0);
  else if (type == 2)
    glColor3d(0.0, 0.0, 1.0);

  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);

  glBegin(GL_LINES);
  c.glVertex();
  (c + n * r).glVertex();
  glEnd();//GL_LINES
  glBegin(GL_POINTS);
  c.glVertex();
  glEnd();

  glPopAttrib();//GL_LIGHTING_BIT
}
