// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <GL/glu.h>
#include "noderingcheckpoint.h"
#include "datablock.h"
#include "plug.h"
#include "command.h"
#include "gbb.h"

// ************************************************************************* //
int	NRingCheckpoint::typeNbr(85);
String	NRingCheckpoint::typeStr("RingCheckpoint");

// ************************************************************************* //
/// DESCRIPTION:
///  dir	- direction vector in world coordinates.
// ************************************************************************* //
void	NRingCheckpoint::initialize()
{
  NShape::initialize();

  DataBlock	&db(getDataBlock());

  db.addInt(0, MININT, MAXINT, "f", false);
  db.addDouble(0.0, 0.0, MAXDOUBLE, "dt", false);
  db.addMat4d(Mat4d::identity(), "matParent", false);
  db.addDouble(0.75, 0.0, MAXDOUBLE, "inner");
  db.addDouble(2.0, 0.0, MAXDOUBLE, "outer");
  db.addDouble(1.0, MINDOUBLE, MAXDOUBLE, "strength");
  db.addPnt3d(Pnt3d(0.0), "center", false);
  db.addVec3d(Vec3d(0.0), "dir", false);

  db.addSignal("update", false);
  db.addSignal("passItOn", false);

  db.affect("f", "update");
  db.affect("matParent", "center");
  db.affect("matParent", "dir");

  buildDagNodes();
}

// ************************************************************************* //
void	NRingCheckpoint::buildDagNodes()
{
  DagNode	*oldParent(_m_parent);
  DagNode	*tr = Command::run(String("new -type Transform -path :"));
  DagNode	*clock = DagNode::getNodeByPath(":Clock0");

  if (oldParent)
    DagNode::parent(*oldParent, *tr);
  DagNode::parent(*tr, *this);
  DagNode::connect(*oldParent, "renderIn", *tr, "renderIn");
  DagNode::connect(*tr, "renderIn", *this, "renderIn");
  DagNode::connect(*tr, "matChild", *this, "matParent");
  DagNode::connect(*clock, "f", *this, "f");
  DagNode::connect(*clock, "dt", *this, "dt");
}

// ************************************************************************* //
/// TODO: draw two balls.. select inner only.
// ************************************************************************* //
void	NRingCheckpoint::glRender(Plug &plugOut)
{
  const int	renderIn = readAsInt("renderIn");

  if (renderIn != Gbb::Rendering)
    {
      static GLUquadric	*quad = gluNewQuadric();
      const int		sel(readAsInt("sel"));
      const double	outer(readAsDouble("outer"));
      Rgb3d	col;

      pushName();
      glPushMatrix();
      readAsMat4d("matParent").glMultMatrix();

      if (sel == 1)
	col = Rgb3d(1.0, 0.0, 5.0);
      else if (sel)
	col = Rgb3d(1.0, 0.5, 0.0);
      else
	col = Rgb3d(0.0, 0.5, 1.0);
      glColor4d(col.r, col.g, col.b, 1.0);
      gluSphere(quad, readAsDouble("inner"), 16, 12);

      glPushAttrib(GL_LIGHTING_BIT);
      glDisable(GL_LIGHTING);
      glBegin(GL_LINES);
      glVertex3d(0.0, 0.0, 0.0);
      glVertex3d(0.0, 0.0, outer);
      glEnd();//GL_LINES
      glPopAttrib();//GL_LIGHTING_BIT

      if (renderIn != Gbb::Selecting)
	{
	  glPushAttrib(GL_DEPTH_BUFFER_BIT);
	  glDepthMask(GL_FALSE);
	  glColor4d(col.r, col.g, col.b, 0.2);
	  gluSphere(quad, outer, 16, 12);
	  glPopAttrib();//GL_DEPTH_BUFFER_BIT
	}
      glPopMatrix();
      popName();
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
/// DESCRIPTION: Attracts any vortex ring withing area of ingluence
/// when ring enters checkpoint area, it carries the charge (freed)
// ************************************************************************* //
void	NRingCheckpoint::update(Plug &plugOut)
{
  const int	f = readAsInt("f");
  const double	dt = readAsDouble("dt");

  std::list<DagNode*>	nodes;
  std::list<DagNode*>::iterator	nit;
  const Mat4d	&matParent = readAsMat4d("matParent");
  const Pnt3d	c = matParent * Pnt3d(0.0);//
  const Vec3d	dir = readAsVec3d("dir");//
  const double	rescale = (matParent * Vec3d(1.0, 0.0, 0.0)).norm();
  const double	inner = readAsDouble("inner") * rescale;
  const double	outer = readAsDouble("outer") * rescale;
  const double	strgth = readAsDouble("strength");

  DagNode::root().getNodesByType("VortexRing", nodes);
  for (nit = nodes.begin(); nit != nodes.end(); ++nit)
    if ((*nit)->readAsInt("m") != getInstNbr())
      {
	const Pnt3d	&ci = (*nit)->readAsPnt3d("center");
	const double	l2 = (ci - c).norm2();

	//std::cout << "??? " << std::endl;
	if (l2 < sq(inner))
	  {
	    (*nit)->writeAsInt("m", getInstNbr());
	    // TODO: pass it on to a curve!
	    Plug	*passItOn = getPlug("passItOn");
	    DagNode	*inputNode = passItOn->getInputNode();

	    if (!inputNode)
	      {
		std::cout << "no node to pass this onto.."  << std::endl;
	      }
	    else
	      {
		// TODO: connect ring!
		
		DagNode::disconnect(*(*nit)->getNodeParent(), "renderIn",
				    *(*nit), "renderIn");
		DagNode::connect(*inputNode, "renderIn", **nit, "renderIn");

		parent(*inputNode, **nit);
				
	      }
	  }
	else if (l2 < sq(outer))
	  {
	    // TODO: use direction to modulate magnet influence.
	    const Vec3d	&diri = (*nit)->readAsVec3d("dir");
	    const bool	isAligned(0.0 < dir * diri);
	    const bool	isPos(0.0 < dir * (ci - c));
	    Mat4d	mrot = Mat4d::identity();
	    Vec3d	ex = (*nit)->readAsVec3d("ex");
	    Vec3d	ey = (*nit)->readAsVec3d("ey");
	    Vec3d	ez = (*nit)->readAsVec3d("ez");
	    
	    if (isAligned)
	      {
		if (!isPos)
		  {
		    const Vec3d		vtarget = (c - ci).unit();
		    const double	d(strgth * (dir * diri) * 
					  (dir * vtarget));

		    // Find twist;
		    TwistLog rot(d * acos(bound(diri.unit() * vtarget,
						0.0, 1.0)),
				 (diri.unit() ^ vtarget).unit(), ci);
		    mrot = rot.exp();
		  }
	      }
	    else
	      {	
		const double	d(-strgth * (dir * diri));
		const Vec3d	away = (c - ci).unit();

		if (!isPos)
		  {
		    const Vec3d	vtarget =
		      (diri - away * (diri * away)).unit();
		    TwistLog rot
		      (d * acos(bound(diri.unit() * vtarget, 0.0, 1.0)),
		       (diri.unit() ^ vtarget).unit(), ci);
		    mrot = rot.exp();
		  }
	      }
	    (*nit)->writeAsVec3d("ex", (mrot * ex).unit());
	    (*nit)->writeAsVec3d("ey", (mrot * ey).unit());
	    (*nit)->writeAsVec3d("ez", (mrot * ez).unit());
	  }
      }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
void	NRingCheckpoint::compute(Plug &plugOut)
{
  if (plugOut.name() == "renderOut")
    glRender(plugOut);
  else if (plugOut.name() == "update")
    update(plugOut);
  else if (plugOut.name() == "center")
    plugOut.writeAsPnt3d(readAsMat4d("matParent") * Pnt3d(0.0));
  else if (plugOut.name() == "dir")
    plugOut.writeAsVec3d((readAsMat4d("matParent") * 
			  Vec3d(0.0, 0.0, 1.0)).unit());
}

