// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <GL/glu.h>
#include <QGLContext>
#include "nodesmokeparticles.h"
#include "nodesmokegenerator.h"
#include "datablock.h"
#include "plugmesh.h"
#include "plugcustom.h"
#include "plugoctree.h"
#include "gbb.h"
#include "command.h"

// ************************************************************************* //
int	NSmokeGenerator::typeNbr(50);
String	NSmokeGenerator::typeStr("SmokeGenerator");

// ************************************************************************* //
/// DESCRIPTION: Returns a point and its normal on the surface of the torus.
///  Surface is parameterized in (u,v) in [0,1]x[0,1], with u for the large
///  circle and v for the little circle.
/// The cordinates loop, hence u = u % 1 and v = v % 1.
// ************************************************************************* //
void	torus(const double	u,
	      const double	v,
	      const double	r,
	      Pnt3d		&p,
	      Vec3d		&n)
{
  const double	alpha(2.0 * M_PI * u);
  const double	cosalpha(cos(alpha));
  const double	sinalpha(sin(alpha));

  const Pnt3d	c(cosalpha, sinalpha, 0.0);
  const Vec3d	x(cosalpha, sinalpha, 0.0);
  const Vec3d	y(0.0, 0.0, 1.0);

  const double	beta(2.0 * M_PI * v);
  const double	cosbeta(cos(beta));
  const double	sinbeta(sin(beta));

  n = x * cosbeta + y * sinbeta;
  p = c + x * (cosbeta * r) + y * (sinbeta * r);
}

// ************************************************************************* //
void	drawTorus(const int imax = 32,
		  const int jmax = 16)
{
  for (int i(0); i < imax; ++i)
    {
      glBegin(GL_QUAD_STRIP);
      for (int j(0); j <= jmax; ++j)
	{
	  Pnt3d	p;
	  Vec3d	n;

	  torus(i / double(imax), j / double(jmax), 0.2, p, n);
	  n.glNormal();
	  p.glVertex();
	  torus((i + 1.0) / imax, j / double(jmax), 0.2, p, n);
	  n.glNormal();
	  p.glVertex();
	}
      glEnd();//GL_QUAD_STRIP
    }
}

// ************************************************************************* //
void	drawEicosahedra()
{
  const double	k(2.0 / sqrt(5.0));
  const double	j(1.0 - sqrt(6.0 / 5.0 - k));
  const GLdouble	p[36] = 
    { 0.0, 0.0, 1.0,
      k * cos(M_PI / 5), k * sin(M_PI / 5), j,
      k * cos(3 * M_PI / 5), k * sin(3 * M_PI / 5), j,
      k * cos(5 * M_PI / 5), k * sin(5 * M_PI / 5), j,
      k * cos(7 * M_PI / 5), k * sin(7 * M_PI / 5), j,
      k * cos(9 * M_PI / 5), k * sin(9 * M_PI / 5), j,
      0.0, 0.0, -1.0,
      k * cos(0 * M_PI / 5), k * sin(0 * M_PI / 5), -j,
      k * cos(2 * M_PI / 5), k * sin(2 * M_PI / 5), -j,
      k * cos(4 * M_PI / 5), k * sin(4 * M_PI / 5), -j,
      k * cos(6 * M_PI / 5), k * sin(6 * M_PI / 5), -j,
      k * cos(8 * M_PI / 5), k * sin(8 * M_PI / 5), -j };
  const double	n[60] = { 0.1796111906, 0.5527864045, 0.7608452130,
			  -0.4702282018, 0.3416407865, 0.7608452130,
			  -0.4702282018, -0.3416407865, 0.7608452130,
			  0.1796111906, -0.5527864045, 0.7608452130,
			  0.5812340224, -0.0000000000, 0.7608452130,
			  0.4702282018, 0.3416407865, -0.7608452130,
			  -0.1796111906, 0.5527864045, -0.7608452130,
			  -0.5812340224, 0.0000000000, -0.7608452130,
			  -0.1796111906, -0.5527864045, -0.7608452130,
			  0.4702282018, -0.3416407865, -0.7608452130,
			  0.7608452130, 0.5527864045, -0.1796111906,
			  -0.2906170112, 0.8944271910, -0.1796111906,
			  -0.9404564037, 0.0000000000, -0.1796111906,
			  -0.2906170112, -0.8944271910, -0.1796111906,
			  0.7608452130, -0.5527864045, -0.1796111906,
			  0.9404564037, -0.0000000000, 0.1796111906,
			  0.2906170112, 0.8944271910, 0.1796111906,
			  -0.7608452130, 0.5527864045, 0.1796111906,
			  -0.7608452130, -0.5527864045, 0.1796111906,
			  0.2906170112, -0.8944271910, 0.1796111906 };
  const GLdouble	pt[180] = { p[6], p[7], p[8], p[0], p[1], p[2],
				    p[3], p[4], p[5], p[9], p[10], p[11],
				    p[0], p[1], p[2], p[6], p[7], p[8],
				    p[12], p[13], p[14], p[0], p[1], p[2],
				    p[9], p[10], p[11], p[15], p[16], p[17],
				    p[0], p[1], p[2], p[12], p[13], p[14],
				    p[3], p[4], p[5], p[0], p[1], p[2],
				    p[15], p[16], p[17], p[21], p[22], p[23],
				    p[18], p[19], p[20], p[24], p[25], p[26],
				    p[24], p[25], p[26], p[18], p[19], p[20],
				    p[27], p[28], p[29], p[27], p[28], p[29],
				    p[18], p[19], p[20], p[30], p[31], p[32],
				    p[30], p[31], p[32], p[18], p[19], p[20],
				    p[33], p[34], p[35], p[33], p[34], p[35],
				    p[18], p[19], p[20], p[21], p[22], p[23],
				    p[24], p[25], p[26], p[3], p[4], p[5],
				    p[21], p[22], p[23], p[27], p[28], p[29],
				    p[6], p[7], p[8], p[24], p[25], p[26],
				    p[30], p[31], p[32], p[9], p[10], p[11],
				    p[27], p[28], p[29], p[33], p[34], p[35],
				    p[12], p[13], p[14], p[30], p[31], p[32],
				    p[21], p[22], p[23], p[15], p[16], p[17],
				    p[33], p[34], p[35], p[3], p[4], p[5],
				    p[15], p[16], p[17], p[21], p[22], p[23],
				    p[6], p[7], p[8], p[3], p[4], p[5],
				    p[24], p[25], p[26], p[9], p[10], p[11],
				    p[6], p[7], p[8], p[27], p[28], p[29],
				    p[12], p[13], p[14], p[9], p[10], p[11],
				    p[30], p[31], p[32], p[15], p[16], p[17],
				    p[12], p[13], p[14], p[33], p[34], p[35] };
  double	nt[180];

  for (int i(0); i < 20; ++i)
    {
      nt[9 * i] = nt[9 * i + 3] = nt[9 * i + 6] = n[3 * i];
      nt[9 * i + 1] = nt[9 * i + 4] = nt[9 * i + 7] = n[3 * i + 1];
      nt[9 * i + 2] = nt[9 * i + 5] = nt[9 * i + 8] = n[3 * i + 2];
    }

  glEnableClientState(GL_NORMAL_ARRAY);
  glEnableClientState(GL_VERTEX_ARRAY);
  glVertexPointer(3, GL_DOUBLE, 0, pt);
  glNormalPointer(GL_DOUBLE, 0, nt);
  glDrawArrays(GL_TRIANGLES, 0, 60);
  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_NORMAL_ARRAY);
}

// ************************************************************************* //
/// DESCRIPTION:
///  r	- radius of sphere
///  freq	- frequency of appearance of particles
// ************************************************************************* //
void	NSmokeGenerator::initialize()
{
  DataBlock	&db(getDataBlock());
  
  db.addInt(0, 0, MAXINT, "sel", false);
  db.addInt(0, 0, MAXINT, "renderIn", false);
  db.addSignal("renderOut", false);
  db.addMat4d(Mat4d::identity(), "matParent", false);//position
  db.addInt(0, MININT, MAXINT, "f", false);//frame
  db.addInt(0, MININT, MAXINT, "fPrev", false);
  db.addSignal("update", false);
  db.addDouble(0.1, 0.0, MAXDOUBLE, "radius");//radius
  db.addDouble(7.0, 0.0, MAXDOUBLE, "freq");// Inverse of grid resolution in ti
  db.addInt(2, 1, MAXINT, "spaceRes");// Grid resolution in space
  db.addDouble(0.5, 0.0, MAXDOUBLE, "size", false);//sizeOfParticle
  db.addDouble(0.001, 0.0, MAXDOUBLE, "mass");//densityOfParticle
  db.addSignal("vortexSet", false);
  db.addDouble(0.0, 0.0, MAXDOUBLE, "accum", false);
  // discrete emiter time.
  db.addDouble(0.0, 0.0, MAXDOUBLE, "asynchtime", false);
  db.addDouble(2.0, 0.0, MAXDOUBLE, "speedSlow");
  db.addDouble(10.0, 0.0, MAXDOUBLE, "speedFast");
  db.addDouble(0.8, 0.0, 1.0, "diffuse");
  db.addDouble(0.25, 0.0, 1.0, "refraction");
  db.addDouble(0.8, 0.0, 1.0, "shadow");
  db.addRgb3d(Rgb3d(0.9, 0.9, 0.9), "color");
  db.addRgb3d(Rgb3d(1.0, 1.0, 1.0), "colorDiffuse");
  db.addRgb3d(Rgb3d(0.0, 0.0, 0.25), "colorShadow");
  db.addRgb3d(Rgb3d(0.9, 0.9, 0.9), "colorFast");
  db.addInt(0, 0, 1, "mode");

  db.affect("renderIn", "renderOut");
  db.affect("f", "update");
  buildDagNodes();
}

// ************************************************************************* //
/// DESCRIPTION: SmokeGenerator factory. Builds the inside of the node,
///   possibly made out of nodes.
///   Transform
///    '-SmokeGenerator
// ************************************************************************* //
void	NSmokeGenerator::buildDagNodes()
{
  DagNode	*oldParent(_m_parent);
  DagNode	*tr = Command::run(String("new -type Transform -path :"));
  DagNode	*clock = DagNode::getNodeByPath(":Clock0");
  DagNode	*smokeset = DagNode::getNodeByPath(":SmokeParticles0");
  DagNode	*vortexset = DagNode::getNodeByPath(":VortexSet0");

  if (oldParent)
    DagNode::parent(*oldParent, *tr);
  DagNode::parent(*tr, *this);
  DagNode::connect(*oldParent, "renderIn", *tr, "renderIn");
  DagNode::connect(*tr, "renderIn", *this, "renderIn");
  DagNode::connect(*tr, "matChild", *this, "matParent");
  DagNode::connect(*clock, "f", *this, "f");
  DagNode::connect(*smokeset, "size", *this, "size");
  DagNode::connect(*vortexset, "signal", *this, "vortexSet");
}

// ************************************************************************* //
void	NSmokeGenerator::glRender(Plug &plugOut)
{
  if (readAsInt("renderIn") != Gbb::Rendering)
    {
      const int		sel(readAsInt("sel"));
      const double	r(readAsDouble("radius"));

      pushName();
      glPushMatrix();
      readAsMat4d("matParent").glMultMatrix();
      glPushAttrib(GL_CURRENT_BIT);
      if (sel == 1)
	glColor3d(1.0, 0.0, 5.0);
      else if (sel)
	glColor3d(1.0, 0.5, 0.0);
      else
	glColor3d(0.0, 0.5, 1.0);
      glPushMatrix();
      glScaled(r, r, r);
      switch (readAsInt("mode"))
	{
	case 0:
	  {
	    drawEicosahedra();
	    break;
	  }
	case 1:
	  {
	    drawTorus();
	    break;
	  }
	}
      glPopMatrix();
      glPopAttrib();//GL_CURRENT_BIT
      glPopMatrix();
      popName();
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
void	NSmokeGenerator::updateReset()
{
  DataBlock	&db(getDataBlock());
  Plug		*fPrevPlug = db.getPlug("fPrev");
  const int	f = readAsInt("f");
  const int	fPrev = fPrevPlug->readAsInt();

  if (f == 0 && f < fPrev)
    {
      writeAsDouble("accum", 0.0);
      writeAsDouble("asynchtime", 0.0);
    }
  fPrevPlug->writeAsInt(f);
}

// ************************************************************************* //
/// DESCRIPTION: Generates particles in a cube.
// ************************************************************************* //
void	NSmokeGenerator::spawnInCube(const int timeRes,
				     const int spaceRes)
{
  DagNode		*smokeSet = DagNode::getNodeByPath(":SmokeParticles0");
  DagNode		*vortexSet = DagNode::getNodeByPath(":VortexSet0");
  const Mat4d	&pos(readAsMat4d("matParent"));

  DagNode		&clock(*DagNode::getNodeByPath(":Clock0"));
  const double	dt(clock.readAsDouble("dt"));
  const int		f(clock.readAsInt("f"));
  const double	worldtime(dt * f);
  double	asynchtime(readAsDouble("asynchtime"));
  const double	freq(readAsDouble("freq"));
  const double	period(1.0 / freq);//Period of apearance of particles
  const double	radius(readAsDouble("radius"));
  const double	size(readAsDouble("size"));
  const double	mass(readAsDouble("mass"));

  std::list<SmokeParticle> *sp = static_cast<std::list<SmokeParticle> *>
    (smokeSet->readAsCustom("sp"));

  for (int t(0); t < timeRes; ++t) // time
    for (int i(0); i < spaceRes; ++i) // space
      for (int j(0); j < spaceRes; ++j)
	for (int k(0); k < spaceRes; ++k)
	  {
	    Pnt3d		ps;
	    double	pt, toff;
	    TwistLog	rlog;

	    if (spaceRes == 1)
	      {
		ps = pos * Pnt3d(0.0);
		toff = 0.0;
		pt = (worldtime - asynchtime) + (t * period);
	      }
	    else
	      {
		ps = pos * 
		  ((Pnt3d(i, j, k) * (2.0 / (spaceRes - 1.0)) -
		    Vec3d(1.0)) * radius);
		ps.eqJitter(radius * 0.5 / (spaceRes - 1.0));
		toff = rand() / double(RAND_MAX) - 0.5;
		pt = (worldtime - asynchtime) + (t + toff) * period;
	      }
	    vortexSet->writeAsPnt3d("p", ps);
	    rlog = vortexSet->readAsTwistLog("rlog");
	    (pt * rlog).exp(ps);
	    sp->push_back(SmokeParticle(ps, size, mass));
	    sp->back().shadei = getInstNbr();
	    sp->back().t = toff * period;
	  }
}

// ************************************************************************* //
/// DESCRIPTION: Generates partciles along a circle.
// ************************************************************************* //
void	NSmokeGenerator::spawnInTorus(const int timeRes,
				      const int spaceRes)
{
  DagNode	*smokeSet = DagNode::getNodeByPath(":SmokeParticles0");
  DagNode	*vortexSet = DagNode::getNodeByPath(":VortexSet0");

  DagNode	&clock(*DagNode::getNodeByPath(":Clock0"));
  const double	dt(clock.readAsDouble("dt"));
  const int	f(clock.readAsInt("f"));
  const double	worldtime(dt * f);
  double	asynchtime(readAsDouble("asynchtime"));

  // Period of apearance of particles:
  const double	period(1.0 / readAsDouble("freq"));

  const Mat4d	&pos(readAsMat4d("matParent"));

  const double	radius(readAsDouble("radius"));
  const double	size(readAsDouble("size"));
  const double	mass(readAsDouble("mass"));

  std::list<SmokeParticle> *sp = static_cast<std::list<SmokeParticle> *>
    (smokeSet->readAsCustom("sp"));

  static unsigned	seedp;

  if (f == 0)
    seedp = 0;

  const double	a(rand_r(&seedp) * 2.0 * M_PI / RAND_R_MAX);

  for (int t(0); t < timeRes; ++t) // time
    for (int i(0); i < spaceRes; ++i) // space
      {
	const double	alpha(a + 2.0 * i * M_PI / spaceRes);
	Pnt3d		ps = pos * Pnt3d(cos(alpha) * radius,
					 sin(alpha) * radius, 0.0);
	double		pt, toff;
	TwistLog	rlog;

	vortexSet->writeAsPnt3d("p", ps);
	rlog = vortexSet->readAsTwistLog("rlog");
	toff = rand() / double(RAND_MAX) - 0.5;
	pt = (worldtime - asynchtime) + (t + toff) * period;
	(pt * rlog).exp(ps);
	sp->push_back(SmokeParticle(ps, size, mass));
	sp->back().shadei = getInstNbr();
	sp->back().t = toff * period;
      }
}

// ************************************************************************* //
/// DESCRIPTION: Particles appear at an asynchronous frequency.
// ************************************************************************* //
void	NSmokeGenerator::update(Plug &plugOut)
{
  updateReset();
  DagNode	&clock(*DagNode::getNodeByPath(":Clock0"));
  const double	dt(clock.readAsDouble("dt"));
  const int	f(clock.readAsInt("f"));
  const double	freq(readAsDouble("freq"));
  const int	spaceRes(readAsInt("spaceRes"));
  const double	period(1.0 / freq);//Period of apearance of particles
  double	accum(readAsDouble("accum")); // dt in which particles appear
  double	asynchtime(readAsDouble("asynchtime"));
  const double	worldtime(dt * f);

  accum += dt;// dt has passed... pt has passed.

  const int	timeRes = int(accum * freq);

  switch (readAsInt("mode"))
    {
    case 0:
      {
	spawnInCube(timeRes, spaceRes);
	break;
      }
    case 1:
      {
	spawnInTorus(timeRes, spaceRes);
	break;
      }
    }
  if (0.0 <= freq)
    {
      writeAsDouble("asynchtime", asynchtime + timeRes * period);
      writeAsDouble("accum", accum - timeRes * period);
    }
  else
    {
      writeAsDouble("asynchtime", worldtime);
      writeAsDouble("accum", 0.0);
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
void	NSmokeGenerator::compute(Plug &plugOut)//plugOut
{
  if (plugOut.shortName() == "update")
    if (readAsInt("mode") == 0)
      update(plugOut);
    else
      update(plugOut);
  else if (plugOut.shortName() == "renderOut")
    glRender(plugOut);
}
