// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <GL/glu.h>
#include <QGLContext>
#include "nodesmokegpuparticles.h"
#include "nodesmokegpugenerator.h"
#include "datablock.h"
#include "plugmesh.h"
#include "plugcustom.h"
#include "plugoctree.h"
#include "gbb.h"
#include "command.h"

// ************************************************************************* //
int	NSmokeGPUGenerator::typeNbr(83);
String	NSmokeGPUGenerator::typeStr("SmokeGPUGenerator");

// ************************************************************************* //
void	drawEicosahedra();

// ************************************************************************* //
/// DESCRIPTION:
///  r	- radius of sphere
///  freq	- frequency of appearance of particles
// ************************************************************************* //
void	NSmokeGPUGenerator::initialize()
{
  DataBlock	&db(getDataBlock());
  
  db.addInt(0, 0, MAXINT, "sel", false);
  db.addInt(0, 0, MAXINT, "renderIn", false);
  db.addSignal("renderOut", false);
  db.addMat4d(Mat4d::identity(), "matParent", false);//position
  db.addInt(0, MININT, MAXINT, "f", false);//frame
  db.addInt(0, MININT, MAXINT, "fPrev", false);
  db.addSignal("update", false);
  db.addDouble(0.1, 0.0, MAXDOUBLE, "radius");//radius
  db.addDouble(7.0, 0.0, MAXDOUBLE, "freq");// Inverse of grid resolution in ti
  db.addInt(2, 1, MAXINT, "spaceRes");// Grid resolution in space
  db.addDouble(0.5, 0.0, MAXDOUBLE, "size", false);//sizeOfParticle
  db.addDouble(0.001, 0.0, MAXDOUBLE, "mass");//densityOfParticle
  db.addSignal("vortexSet", false);
  db.addDouble(0.0, 0.0, MAXDOUBLE, "accum", false);
  // discrete emiter time.
  db.addDouble(0.0, 0.0, MAXDOUBLE, "asynchtime", false);
  db.addDouble(2.0, 0.0, MAXDOUBLE, "speedSlow");
  db.addDouble(10.0, 0.0, MAXDOUBLE, "speedFast");
  db.addDouble(0.8, 0.0, 1.0, "diffuse");
  db.addDouble(0.25, 0.0, 1.0, "refraction");
  db.addDouble(0.8, 0.0, 1.0, "shadow");
  db.addRgb3d(Rgb3d(1.0, 1.0, 0.8), "color");
  db.addRgb3d(Rgb3d(1.0, 1.0, 0.8), "colorDiffuse");
  db.addRgb3d(Rgb3d(0.0, 0.0, 0.0), "colorShadow");
  db.addRgb3d(Rgb3d(1.0, 1.0, 0.8), "colorFast");

  db.affect("renderIn", "renderOut");
  db.affect("f", "update");
  buildDagNodes();
}

// ************************************************************************* //
/// DESCRIPTION: SmokeGenerator factory. Builds the inside of the node,
///   possibly made out of nodes.
///   Transform
///    '-SmokeGenerator
// ************************************************************************* //
void	NSmokeGPUGenerator::buildDagNodes()
{
  DagNode	*oldParent(_m_parent);
  DagNode	*tr = Command::run(String("new -type Transform -path :"));
  DagNode	*clock = DagNode::getNodeByPath(":Clock0");
  DagNode	*smokeset = DagNode::getNodeByPath(":SmokeGPUParticles0");
  DagNode	*vortexset = DagNode::getNodeByPath(":VortexSet0");

  if (oldParent)
    DagNode::parent(*oldParent, *tr);
  DagNode::parent(*tr, *this);
  DagNode::connect(*oldParent, "renderIn", *tr, "renderIn");
  DagNode::connect(*tr, "renderIn", *this, "renderIn");
  DagNode::connect(*tr, "matChild", *this, "matParent");
  DagNode::connect(*clock, "f", *this, "f");
  DagNode::connect(*smokeset, "size", *this, "size");
  DagNode::connect(*vortexset, "signal", *this, "vortexSet");
}

// ************************************************************************* //
void	NSmokeGPUGenerator::glRender(Plug &plugOut)//plugOut
{
  const int mode = readAsInt("renderIn");

  if (mode != Gbb::Rendering)
    {
      const int	sel(readAsInt("sel"));
      const double	r(readAsDouble("radius"));

      pushName();
      glPushMatrix();
      readAsMat4d("matParent").glMultMatrix();

      glPushAttrib(GL_CURRENT_BIT);
      if (sel == 1)
	glColor3d(1.0, 0.0, 5.0);
      else if (sel)
	glColor3d(1.0, 0.5, 0.0);
      else
	glColor3d(0.0, 0.5, 1.0);

      glPushMatrix();
      glScaled(r, r, r);
      drawEicosahedra();
      glPopMatrix();

      glPopAttrib();//GL_CURRENT_BIT

      glPopMatrix();
      popName();
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
void	NSmokeGPUGenerator::updateReset()
{
  DataBlock	&db(getDataBlock());
  Plug		*fPrevPlug = db.getPlug("fPrev");
  const int	f = readAsInt("f");
  const int	fPrev = fPrevPlug->readAsInt();

  if (f == 0 && f < fPrev)
    {
      writeAsDouble("accum", 0.0);
      writeAsDouble("asynchtime", 0.0);
    }
  fPrevPlug->writeAsInt(f);
}

// ************************************************************************* //
/// DESCRIPTION: Particles appear at an asynchronous frequency.
// ************************************************************************* //
void	NSmokeGPUGenerator::update(Plug &plugOut)
{
  updateReset();

  DagNode	&clock(*DagNode::getNodeByPath(":Clock0"));
  const double	dt(clock.readAsDouble("dt"));
  const int	f(clock.readAsInt("f"));
  const Mat4d	&pos(readAsMat4d("matParent"));
  const double	radiuss(readAsDouble("radius"));
  const double	freq(readAsDouble("freq"));
  const int	spaceRes(readAsInt("spaceRes"));
  const double	period(1.0 / freq);//Period of apearance of particles
  double	accum(readAsDouble("accum")); // dt in which particles appear
  double	asynchtime(readAsDouble("asynchtime"));
  const double	worldtime(dt * f);
  const double	size(readAsDouble("size"));
  const double	mass(readAsDouble("mass"));

  std::list<DagNode *>	smokeSets;

  DagNode::root().getNodesByType("SmokeGPUParticles", smokeSets);

  Plug		*vortexSetData;
  DagNode	*vortexSetNode; // For evaluating log(p).

  accum += dt;// dt has passed... pt has passed.

  const int	timeRes = int(accum * freq);

  if (0.0 <= freq)
    {
      writeAsDouble("asynchtime", asynchtime + timeRes * period);
      writeAsDouble("accum", accum - timeRes * period);
    }
  else
    {
      writeAsDouble("asynchtime", worldtime);
      writeAsDouble("accum", 0.0);
    }

  if (!smokeSets.empty() && 0.0 < freq)
    {
      std::list<SmokeGPUParticle> *sp = 
	static_cast<std::list<SmokeGPUParticle> *>
	(smokeSets.front()->readAsCustom("sp"));
      TwistLog	rlog;

      if ((vortexSetData = getDataBlock().getPlug("vortexSet")) &&
	  (vortexSetNode = vortexSetData->getInputNode()))
	if (timeRes && spaceRes)
	  for (int t(0); t < timeRes; ++t) // time
	    for (int i(0); i < spaceRes; ++i) // space
	      for (int j(0); j < spaceRes; ++j)
		for (int k(0); k < spaceRes; ++k)
		  {
		    Pnt3d	ps;
		    double	pt, toff;

		    if (spaceRes == 1)
		      {
			ps = pos * Pnt3d(0.0);
			toff = 0.0;
			pt = (worldtime - asynchtime) + (t * period);
		      }
		    else
		      {
			ps = pos * 
			  ((Pnt3d(i, j, k) * (2.0 / (spaceRes - 1.0)) -
			    Vec3d(1.0)) * radiuss );
			ps.eqJitter(radiuss * 0.5 / (spaceRes - 1.0));
			toff = rand() / double(RAND_MAX) - 0.5;
			pt = (worldtime - asynchtime) + (t + toff) * period;
		      }
		    vortexSetNode->writeAsPnt3d("p", ps);
		    rlog = vortexSetNode->readAsTwistLog("rlog");
		    (pt * rlog).exp(ps);
		    sp->push_back(SmokeGPUParticle(ps, size, mass));
		    sp->back().shadei = 0;//!!!
		    sp->back().t = toff * period;
		  }
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
void	NSmokeGPUGenerator::compute(Plug &plugOut)//plugOut
{
  if (plugOut.shortName() == "update")
    update(plugOut);
  else if (plugOut.shortName() == "renderOut")
    glRender(plugOut);
}
