// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:30 $
/// $Revision: 1.1 $
// ************************************************************************* //
#define GLX_GLXEXT_PROTOTYPES
#include <GL/glu.h>
#include <QGLContext>
#include "nodesmokeparticles.h"
#include "nodesmokegenerator.h"
#include "datablock.h"
#include "plugmesh.h"
#include "plugcustom.h"
#include "plugoctree.h"
#include "parcel1.dat"
#include "parcel2.dat"
#include "parcel3.dat"
#include "parcel4.dat"
#include "parcel5.dat"
#include "parcel6.dat"
#include "gbb.h"
#include "command.h"
#include "dataclock.h"
#ifndef _WIN32
#include <GL/glx.h>
#else //_WIN32
#include <GL/glext.h>
#endif//_WIN32

// Declares function pointer *********************************************** //
PFNGLBLENDFUNCSEPARATEPROC	glBlendFuncSeparate;

// ************************************************************************* //
int	PlugSmokeParticles::typeNbr(64);
String	PlugSmokeParticles::typeStr("SmokeParticles");
int	NSmokeParticles::typeNbr(49);
String	NSmokeParticles::typeStr("SmokeParticles");

// ************************************************************************* //
void	SmokeParticle::glRenderEllipsoid(const double ttb,
					 const double ttl,
					 const double ttd) const
{
  const double	opacity = presence(ttb, ttl, ttd);

  if (opacity == 0.0)
    return;

  static GLUquadric	*quad = gluNewQuadric();
  Vec3d		axes[3], evec[3];
  double	eval[3] = { 0.0, 0.0, 0.0 };
  Mat4d		m;

  glColor4d(0.8 * i, 0.7 * i, 0.5 * i, 0.5 + 0.5 * mass);
  if (cov.eigenDecomposition(eval, evec) == 3)
    {
      glPushMatrix();
      for (int i(0); i < 3; ++i)
	axes[i] = evec[i].unit() * sqrt(3.0 * fabs(eval[i]));
      if (0.0 <= (evec[0] ^ evec[1]) * evec[2])
	m = Mat4d(axes[0], axes[1], axes[2], Pnt3d(0.0));
      else
	m = Mat4d(axes[1], axes[0], axes[2], Pnt3d(0.0));
      p.glTranslate();
      m.glMultMatrix();
      gluSphere(quad, 1.0, 16, 12);
      glPopMatrix();
    }
  else
    std::cerr << "Error: singular ellipsoid." << std::endl;
}

// ************************************************************************* //
/// DESCRIPTION: Returns volume of ellipsoid.
// ************************************************************************* //
double	SmokeParticle::volume() const
{
  return sqrt(fabs(cov.det())) * M_PI / 6.0;
}

// ************************************************************************* //
/// DESCRIPTION: Returns mass per unit volume.
// ************************************************************************* //
double	SmokeParticle::density() const
{
  const double v = volume();

  if (1.0e-8 < v)
    return mass / v;
  return 0.0;
}

// ************************************************************************* //
void	SmokeParticle::glRenderSphere(const double size) const
{
  static GLUquadric	*quad = gluNewQuadric();

  glColor4d(0.8, 0.7, 0.5, 0.5 + 0.5 * mass);
  glPushMatrix();
  p.glTranslate();
  gluSphere(quad, size, 16, 12);
  glPopMatrix();
}

// ************************************************************************* //
/// DESCRIPTION: According to the particle's clock, returns the presence 
///  of the particle in [0, 1]. 1 is present, 0 is absent.
// ************************************************************************* //
double	SmokeParticle::presence(const double ttb,
				const double ttl,
				const double ttd) const
{
  if (t <= 0.0 || ttd <= t)
    return 0.0;
  if (ttl < t) // fade out particle before disapearing.
    {
      const double	l((t - ttl) / (ttd - ttl));

      //return 1.0 + l * l * l * (l * (15.0 - 6.0 * l) - 10.0);
      return 1.0 - l * l * (3.0 - 2.0 * l);
    }
  if (t < ttb)
    {
      const double	l(t / ttb);

      //return -l * l * l * (l * (15.0 - 6.0 * l) - 10.0);
      return l * l * (3.0 - 2.0 * l);
    }
  return 1.0;
}

// ************************************************************************* //
/// DESCRIPTION:
///  ^
///  |--_
///  |   `._
///  |______`--__
/// 
// ************************************************************************* //
double	rbfSmokeShade(const double	l)
{
  if (l <= 0.0)
    return 1.0;
  if (1.0 <= l)
    return 0.0;
  return sq(4.0 - 20.0 / (l * l + 4.0));
}

// ************************************************************************* //
/// DESCRIPTION:
///  rescaleTex	- size of texture displayed on top of ellipsoid.
// ************************************************************************* //
void	SmokeParticle::glRenderEllipse(const Pnt3d	&v,
				       double		*const parcelQuads,
				       double		*const parcelColor,
				       const double	ttb,
				       const double	ttl,
				       const double	ttd,
				       const double	rescaleTex,
				       const SmokeShader	&shader) const
{
  double	opacity = presence(ttb, ttl, ttd);

  if (opacity == 0.0)
    {
      parcelColor[3] = parcelColor[7] =
	parcelColor[11] = parcelColor[15] = 0.0;
      parcelQuads[0] = parcelQuads[3] = parcelQuads[6] = parcelQuads[9] = p.x;
      parcelQuads[1] = parcelQuads[4] = parcelQuads[7] = parcelQuads[10] = p.y;
      parcelQuads[2] = parcelQuads[5] = parcelQuads[8] = parcelQuads[11] = p.z;
      return;
    }
  const Vec3d	z((v - p).unit());
  Vec3d		x, y;

  if (fabs(z.x) < fabs(z.y) && fabs(z.x) < fabs(z.z))
    y = Vec3d(1.0, 0.0, 0.0) - z * z.x;
  else if (fabs(z.y) < fabs(z.z))
    y = Vec3d(0.0, 1.0, 0.0) - z * z.y;
  else
    y = Vec3d(0.0, 0.0, 1.0) - z * z.z;
  y.eqUnit();
  x = y ^ z;

  Vec3d	x1, y1;

  if (cov.equals(Mat3d::identity(), 1.0e-6))
    {
      std::cout << "!!!" << std::endl;
      x1 = x, y1 = y;
      opacity *= density();
    }
  else
    {
      Mat2d cov2d(cov.project(x, y));

      double	eval[2] = { 0.0, 0.0 };
      Vec2d		evec[2];
      const int	nbEvals(cov2d.eigenvalues(eval, 1.0e-8));

      //std::cout << density() << " " << std::flush;
      //opacity *= 2.0 * density();
      opacity *= 2.0 * sqrt(fabs(cov.project(z))) * density();
      //opacity *= 2.0 * sqrt(fabs(cov.project(z)));

      if (nbEvals == 2)
	{
	  Vec2d		axis0, axis1;
	  double	l0, l1;

	  cov2d.eigenvectors(eval, evec);
	  l0 = bound(sqrt(2.0 * fabs(eval[0])), 1.0e-6, 2.0);
	  l1 = bound(sqrt(2.0 * fabs(eval[1])), 1.0e-6, 2.0);
	  if (1.0e-6 < evec[0].norm2())
	    axis0 = evec[0].unit() * l0;
	  else
	    axis0 = Vec2d(0.0);
	  if (1.0e-6 < evec[1].norm2())
	    axis1 = evec[1].unit() * l1;
	  else
	    axis1 = Vec2d(0.0);
	  x1 = x * axis0.x + y * axis0.y;
	  y1 = x * axis1.x + y * axis1.y;
	  if (axis0.x * axis1.y <= axis0.y * axis1.x)
	    x1 = -x1;
	}
      else
	std::cout << "???" << std::endl;
    }

  // TODO:
  // bound(1.0 - p.z / 10.0, 0.0, 1.0)

  // Shade ***************************************************************** //
  {
    // TODO: give  to generator an index, and get color from generator array
    for (int k(0); k < 16; k += 4)
      {
	Rgb3d		col(0.0);
	const double	vl = this->v.norm();
	const double	u = 1.0 - 
	  rbfSmokeShade((vl - shader.speedSlow) / shader.speedFast);

	// Add velocity color
	col = (shader.col + (shader.colFast - shader.col) * u);

	const Vec3d	n = (cov.inverse() * z).unit();
	const double	ln = shader.l * n;
	const double	isd = ((0.0 < ln)? (0.0 + ln) / (0.0 + 1.0):
			       (0.0 - ln * shader.refraction) / (0.0 + 1.0));

	// Attenuate with non-diffuse reflection
	col = col * (1.0 - shader.diffuse * isd) +
	  shader.colDiffuse * (isd * shader.diffuse);

	// Add shadow ... use ambient to fill shadow?
	const double	ip = (1.0 - shader.shadow) + shader.shadow * i;

	col = col * ip + shader.colShadow * (1.0 - ip);

	parcelColor[k + 0] = col.r;
	parcelColor[k + 1] = col.g;
	parcelColor[k + 2] = col.b;
	parcelColor[k + 3] = opacity;
      }
  }

  // Geometry ************************************************************** //
  {
    const Pnt3d p0 = (p + (x1 + y1) * rescaleTex);
    const Pnt3d p1 = (p - (x1 - y1) * rescaleTex);
    const Pnt3d p2 = (p - (x1 + y1) * rescaleTex);
    const Pnt3d p3 = (p + (x1 - y1) * rescaleTex);

    parcelQuads[0] = p0.x;
    parcelQuads[1] = p0.y;
    parcelQuads[2] = p0.z;
    parcelQuads[3] = p1.x;
    parcelQuads[4] = p1.y;
    parcelQuads[5] = p1.z;
    parcelQuads[6] = p2.x;
    parcelQuads[7] = p2.y;
    parcelQuads[8] = p2.z;
    parcelQuads[9] = p3.x;
    parcelQuads[10] = p3.y;
    parcelQuads[11] = p3.z;
  }
}

// ************************************************************************* //
void	PlugSmokeParticles::saveData(std::ostream &os)
{
  std::list<SmokeParticle>::iterator	dit;

  os << _m_data->size() << ' ';
  if (_m_data)
    for (dit = _m_data->begin(); dit != _m_data->end(); ++dit)
      os << *dit << ' ';
}

// ************************************************************************* //
void	PlugSmokeParticles::loadData(std::istream &is)
{
  int	size;

  is >> size;
  if (_m_data)
    for (int i(0); i < size; ++i)
      {
	_m_data->push_back(SmokeParticle());
	is >> _m_data->back();
      }
}

// ************************************************************************* //
void	NSmokeParticles::initialize()
{
  DataBlock	&db(getDataBlock());

  db.addInt(0, 0, MAXINT, "renderIn", false);
  db.addSignal("renderOut", false);
  db.addInt(0, MININT, MAXINT, "f", false);
  db.addInt(0, MININT, MAXINT, "fPrev", false);
  db.addSignal("update", false);
  db.addSignal("vortexSet", false);
  db.addPnt3d(Pnt3d(0.0), "eye", false);
  db.addVec3d(Vec3d(0.0, 0.0, -1.0), "light");// Position (in h coords)

  db.addDouble(0.5, 0.0, MAXDOUBLE, "ttb");// Time to birth
  db.addDouble(4.0, 0.0, MAXDOUBLE, "ttl");// Time to live
  db.addDouble(5.0, 0.0, MAXDOUBLE, "ttd");// Time to die
  db.addDouble(2.0, 0.0, MAXDOUBLE, "maxStretch");
  db.addDouble(0.05, 0.0, MAXDOUBLE, "minDensity");
  db.addDouble(0.0001, 0.0, MAXDOUBLE, "minMass");
  db.addBool(true, "renderEllipse");
  db.addBool(false, "split");
  db.addBool(false, "shaded");

  db.addDouble(0.1, 0.0, MAXDOUBLE, "size");
  db.addDouble(0.5, 0.0, MAXDOUBLE, "transparency");

  db.addCustom(*new PlugOctree(db), "octree", false);

  //Rescale the texture to cover the ellipsoid.
  db.addDouble(2.5, 0.0, MAXDOUBLE, "rescaleTex");
  db.addInt(0, 0, 5, "dat");

  db.addCustom(*new PlugSmokeParticles(db), "sp", false);
  db.addCustom(*new PlugBBox3d(db), "bbox", false);

  db.addInt(0, 0, MAXINT, "nbParticles");

  db.affect("renderIn", "renderOut");
  db.affect("renderEllipse", "renderOut");
  db.affect("f", "update");
  db.affect("update", "nbParticles");
  db.affect("f", "bbox");
  buildDagNodes();
}

// ************************************************************************* //
void	NSmokeParticles::buildDagNodes()
{
  DagNode	*clock = DagNode::getNodeByPath(":Clock0");
  DagNode	*vortexset = DagNode::getNodeByPath(":VortexSet0");

  DagNode::connect(*clock, "f", *this, "f");
  DagNode::connect(*_m_parent, "renderIn", *this, "renderIn");
  DagNode::connect(*vortexset, "signal", *this, "vortexSet");
  DagNode::connect(*vortexset, "octree", *this, "octree");
}

// ************************************************************************* //
/// DESCRIPTION: Resets the smoke parcels when restarting animation.
// ************************************************************************* //
void	NSmokeParticles::updateReset()
{
  DataBlock	&db(getDataBlock());
  std::list<SmokeParticle> *sp = 
    static_cast<std::list<SmokeParticle> *>(readAsCustom("sp"));
  Plug		*fPrevData = db.getPlug("fPrev");
  const int	fPrev = fPrevData->readAsInt();
  const int	f = readAsInt("f");

  if (f == 0 && f < fPrev)
    {
      srand(1);
      sp->clear();
    }
  fPrevData->writeAsInt(f);
}

// ************************************************************************* //
/// DESCRIPTION: Moves the smoke parcels.
// ************************************************************************* //
void	NSmokeParticles::updateMove()
{
  DataBlock	&db(getDataBlock());
  Plug		*vortexSetData;
  DagNode	*rs = 0;
  DagNode	&clock = *DagNode::getNodeByPath(":Clock0");
  const double	dt = clock.readAsDouble("dt");
  std::list<SmokeParticle>::iterator pit;
  std::list<SmokeParticle> *sp = 
    static_cast<std::list<SmokeParticle> *>(readAsCustom("sp"));
  const bool	split = readAsBool("split");
  const bool	isShaded = readAsBool("shaded");

  if ((vortexSetData = db.getPlug("vortexSet")) &&
      (rs = vortexSetData->getInputNode()))
    {
      Plug	&plugp = *rs->getPlug("pnoisy");
      Plug	&plugRlog = *rs->getPlug("rlogNoisy");
      Plug	&plugJac = *rs->getPlug("jacNoisy");

      //std::cout << "Nb particles = " << sp->size() << std::endl;
      for (pit = sp->begin(); pit != sp->end(); ++pit)
	{
	  TwistLog	rlog;
	  Mat3d		jac;

	  plugp.writeAsPnt3d(pit->p);
	  rlog = plugRlog.readAsTwistLog();
	  if (split)
	    {
	      jac = plugJac.readAsMat3d();
	      pit->cov.eqMultTransMult(jac);
	    }
	  if (isShaded)
	    pit->v = rlog * pit->p;
	  (rlog * dt).exp(pit->p);
	}
    }
}

// ************************************************************************* //
/// DESCRIPTION: 
///  Deletes:	- the dead smoke parcels (whose time is over).
///		- particle inside obstacles
///  Asigns a short lifespan to light particles or transparent particles.
// ************************************************************************* //
void	NSmokeParticles::updateDelete()
{
  std::list<SmokeParticle> *sp = 
    static_cast<std::list<SmokeParticle> *>(readAsCustom("sp"));
  std::list<SmokeParticle>::iterator pit;
  const double	ttl(readAsDouble("ttl"));
  const double	ttd(readAsDouble("ttd"));
  DagNode	&clock(*DagNode::getNodeByPath(":Clock0"));
  const double	dt(clock.readAsDouble("dt"));
  const double	minDensity(readAsDouble("minDensity"));
  const double	minMass(readAsDouble("minMass"));

  std::list<DagNode *>	nodes;
  DagNode::root().getNodesByType("Obstacle", nodes);
  DagNode	*obs = nodes.empty()? 0: nodes.front();
  Mesh	*mesh = obs? static_cast<Mesh *>(obs->readAsCustom("mesh")): 0;
  const Mat4d	imatParent(obs? obs->readAsMat4d("imatParent"): Mat4d(0.0));

  for (pit = sp->begin(); pit != sp->end(); )
    if ((ttd < (pit->t += dt)) || 
	(mesh && mesh->contains(imatParent * pit->p)))
      sp->erase(pit++);
    else
      {
	// schedule particle death
	if (pit->t < ttl && (pit->density() <= minDensity ||
			     pit->mass <= minMass))
	  pit->t = ttl;
	++pit;
      }
}

// ************************************************************************* //
/// DESCRIPTION: Splits the smoke parcels that are stretched too much.
// ************************************************************************* //
void	NSmokeParticles::updateSplit()
{
  std::list<SmokeParticle> *sps = 
    static_cast<std::list<SmokeParticle> *>(readAsCustom("sp"));
  std::list<SmokeParticle>::iterator pit;

  double	size(readAsDouble("size"));
  const double	maxEigen(sq(readAsDouble("maxStretch") * size) / 3.0);
  const double	minDensity(readAsDouble("minDensity"));

  for (pit = sps->begin(); pit != sps->end(); )
    {
      SmokeParticle	&sp(*pit);
      bool		isIncremented(false);
    
      if (!sp.cov.equals(Mat3d::identity()) && minDensity < sp.density())
	{
	  double	evals[3];

	  sp.cov.sortEigenvalues(evals);//TODO: get max eigen only!
	  // check?
	  /*
	  std::cout << "eij = " << sp.cov(0, 0) << " "
		    << sp.cov(1, 1) << " "
		    << sp.cov(2, 2) << std::endl;
	  std::cout << "hij = " << evals[0] << " "
		    << evals[1] << " "
		    << evals[2] << std::endl;
	  */
	  if (maxEigen < evals[0])
	    {
	      double	eval[3];
	      Vec3d	evec[3];

	      // PROBLEM... too thick...
	      if (sp.cov.eigenDecomposition(eval, evec) == 3)
		{
		  const Mat3d	mp(evec[0], evec[1], evec[2]);
		  const Mat3d	md(fabs(eval[0] * 0.25),
				   fabs(eval[1]), fabs(eval[2]));
		  const double	offset(0.4 * sqrt(3.0 * fabs(evals[0])));
		  const Vec3d	splitOffset(evec[0].unit() * offset);

		  sp.cov = mp * md * mp.inverse();

		  sp.cov.eqSymmetric();
		  sp.mass *= 0.5;

		  sps->push_back(SmokeParticle(sp.p + splitOffset, sp));
		  sps->push_back(SmokeParticle(sp.p - splitOffset, sp));
		  sps->erase(pit++);
		  isIncremented = true;
		}
	    }
	}
      if (!isIncremented)
	++pit;
    }
}

// ************************************************************************* //
/// DESCRIPTION: Sorts the particles. Result is a list with first particle 
///  being the farthest from the eye, etc... i.e. the farthest is rendered
///  first.
// ************************************************************************* //
void	NSmokeParticles::sortInViewDepth()
{
  std::list<SmokeParticle> *sp = 
    static_cast<std::list<SmokeParticle> *>(readAsCustom("sp"));
  std::list<SmokeParticle>::iterator pit;
  const Pnt3d	&eye(readAsPnt3d("eye"));

  for (pit = sp->begin(); pit != sp->end(); ++pit)
    pit->z = (pit->p - eye).norm2();
  sp->sort();
  sp->reverse();
}

// ************************************************************************* //
/// DESCRIPTION:
///  ^
///  |--_
///  |   `._
///  |______`--__
/// 
// ************************************************************************* //
double	rbfSmokeParticle(const double	l2,
			 const double	r2)
{
  return (r2 < l2)? 0.0: sq(4.0 - 20.0 / (l2 / r2 + 4.0));
}

// ************************************************************************* //
/// DESCRIPTION:
///  Uses two grids to traverse space and accumulate opacities.
///  Assumes particles are undeformable spheres of constant size.
///  Radius of particle is that of space between two slices...
// ************************************************************************* //
void	NSmokeParticles::shade()
{
  Vec3d	x, y, z;
  const double	ttb(readAsDouble("ttb"));
  const double	ttl(readAsDouble("ttl"));
  const double	ttd(readAsDouble("ttd"));
  const double	opac(readAsDouble("transparency"));
  const Vec3d	light(readAsVec3d("light"));
  std::list<DagNode *>	smokeSets;

  DagNode::root().getNodesByType("SmokeParticles", smokeSets);
  std::list<SmokeParticle> *sp = static_cast<std::list<SmokeParticle> *>
    (smokeSets.front()->readAsCustom("sp"));

  // Creates new coordinates frame ***************************************** //
  if (fabs(light.x) < fabs(light.y) && fabs(light.x) < fabs(light.z))
    x = Vec3d(1.0, 0.0, 0.0);
  else if (fabs(light.y) < fabs(light.z))
    x = Vec3d(0.0, 1.0, 0.0);
  else
    x = Vec3d(0.0, 0.0, 1.0);
  x = (x - light * (x * light)).unit();
  y = light ^ x;
  z = light;

  // Reset lighting and sort atoms in lightdepth *************************** //
  std::list<SmokeParticle>::iterator	ait;

  for (ait = sp->begin(); ait != sp->end(); ++ait)
    ait->z = Vec3d(ait->p) * light;
  sp->sort();

  // Finds a cube around data, where z orthogonal to map ******************* //
  BBox3d	box(sp->front().p);

  for (ait = sp->begin(); ait != sp->end(); ++ait)
    {
      const Vec3d	p(ait->p);

      ait->i = 0.0;
      box.fill(Pnt3d(p * x, p * y, p * z));
    }
  box.eqCubify();

  const double	edgeLen(box.pmax.x - box.pmin.x);// Length of cube edge

  // *********************************************************************** //
  std::list<SmokeParticle>::iterator	pitMarch;
  std::list<SmokeParticle>::iterator	pitWriteLight;

  const unsigned	res(128);
  const int		ressigned(128);
  double		accum0[res * res];
  double		accum1[res * res];

  for (unsigned i(0), n(res * res); i < n; ++i)
    accum1[i] = 1.0;

  const double	halfedge(0.5 * edgeLen / (res - 1.0));
  //const double	s(readAsDouble("size"));
  const double	s(edgeLen / (res - 1.0));
  const double	span(halfedge + s); // Space depth/2 considered per slice

  pitMarch = pitWriteLight = sp->begin();
  // Accumulates the opacity *********************************************** //
  for (unsigned k = 0; k < res; ++k)
    {
      // Depth of plane:
      const double	zaccum(box.pmin.z + k * edgeLen / (res - 1.0));

      for (unsigned i(0), n(res * res); i < n; ++i) // Swap grids
	accum0[i] = accum1[i];

      // Pass particles that don't need to rater
      while (pitMarch != sp->end() && Vec3d(pitMarch->p) * z < zaccum - span)
	++pitMarch;

      // Raster opacity in grid ******************************************** //
      std::list<SmokeParticle>::iterator	pitRaster;

      for (pitRaster = pitMarch; pitRaster != sp->end() &&
	     Vec3d(pitRaster->p) * z < zaccum + span; ++pitRaster)
	{
	  const SmokeParticle	&part(*pitRaster);
	  const Vec3d		p(part.p);
	  const Pnt3d		r(p * x, p * y, p * z); // In map coords
	  const Pnt2d		r2d(r.x - box.pmin.x, r.y - box.pmin.y);
	  const double		world2index((res - 1.0) / edgeLen);
	  const double		index2world(edgeLen / (res - 1.0));

	  // Find voxels affected by particle ****************************** //
	  int	imin = max(int(floor(world2index * (r2d.x - s))), 0);
	  int	imax = min(int(ceil(world2index * (r2d.x + s))), ressigned);
	  int	jmin = max(int(floor(world2index * (r2d.y - s))), 0);
	  int	jmax = min(int(ceil(world2index * (r2d.y + s))), ressigned);

	  for (int i(imin); i < imax; ++i)
	    for (int j(jmin); j < jmax; ++j)
	      {
		const Pnt3d	q(i * index2world + box.pmin.x,
				  j * index2world + box.pmin.y, zaccum);
		const Vec3d	l(fabs(q.x - r.x),
				  fabs(q.y - r.y),
				  fabs(q.z - r.z));

		if (l.x < span && l.y < span && l.z < span)
		  {
		    double	dx(l.x / span);
		    double	dy(l.y / span);
		    double	dz(l.z / span);
		    //double	volume(8.0 * s * s * s);
		    double	volume(0.01);
		    double	density(part.mass / volume);

		    dx = 1.0 + dx * dx * dx * (dx * (15.0 - 6.0 * dx) - 10.0);
		    dy = 1.0 + dy * dy * dy * (dy * (15.0 - 6.0 * dy) - 10.0);
		    dz = 1.0 + dz * dz * dz * (dz * (15.0 - 6.0 * dz) - 10.0);
		    accum1[j * res + i] *=
		      (1.0 - density * part.presence(ttb, ttl, ttd) *
		       opac * dx * dy * dz);
		  }
	      }
	}

      // Read radiance in buffer ******************************************* //
      for (; pitWriteLight != sp->end() &&
	     Vec3d(pitWriteLight->p) * z < zaccum; ++pitWriteLight)
	{
	  const Vec3d	p(pitWriteLight->p);
	  const Pnt3d	r(p * x, p * y, p * z);
	  double	u, v, w;
	  int	il, jl, ih, jh, kl;

	  u = (res - 1.0) * (r.x - box.pmin.x) / edgeLen;
	  v = (res - 1.0) * (r.y - box.pmin.y) / edgeLen;
	  w = (res - 1.0) * (r.z - box.pmin.z) / edgeLen;
	  il = max(int(floor(u)), 0);
	  ih = min(int(ceil(u)), ressigned);
	  jl = max(int(floor(v)), 0);
	  jh = min(int(ceil(v)), ressigned);
	  kl = max(int(floor(w)), 0);
	  u -= il;
	  v -= jl;
	  w -= kl;

	  const double	vi00(accum0[res * jl + il] * (1.0 - u) + 
			     accum0[res * jl + ih] * u);
	  const double	vi10(accum0[res * jh + il] * (1.0 - u) + 
			     accum0[res * jh + ih] * u);
	  const double	vi11(accum1[res * jh + il] * (1.0 - u) + 
			     accum1[res * jh + ih] * u);
	  const double	vi01(accum1[res * jl + il] * (1.0 - u) + 
			     accum1[res * jl + ih] * u);
	  const double t = (vi00 * (1.0 - v) + vi10 * v) * (1.0 - w) +
	    (vi01 * (1.0 - v) + vi11 * v) * w;

	  pitWriteLight->i = bound(t, 0.0, 1.0);
	}
    }
}

// ************************************************************************* //
void	NSmokeParticles::update(Plug &plugOut)
{
  readAsInt("f");
  updateDelete();
  updateReset();
  updateMove();
  if (readAsBool("split"))
    updateSplit();
  if (readAsBool("shaded"))
    {
      if (readAsInt("renderIn") == Gbb::Rendering)
	{
	  Clock	clock(Clock::CPUTime);

	  clock.tic();
	  shade();
	  DagNode::root().writeAsDouble("shadingTime", clock.tic());
	}
      else
	shade();
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
static void	*getProcAddress(const char *name)
{
#ifdef _WIN32
  return (void *)wglGetProcAddress(name);
#else
  return (void *)glXGetProcAddressARB((const GLubyte *)name);
#endif /* _WIN32 */
}

// ************************************************************************* //
void	NSmokeParticles::glRender(Plug &plugOut)
{
  /// Manages texure Id per context. *************************************** //
  static std::map<int, std::vector<GLuint> >	texNameMap;
  const int		curContextId = int(QGLContext::currentContext());
  std::vector<GLuint>   *texName(&texNameMap[curContextId]);

  //glEnable(GL_TEXTURE_2D);
  {
    if (texName->empty())
      {
	*texName = std::vector<GLuint>(6);
	GLuint	texNameForGen[6];
  
	glGenTextures(6, texNameForGen);
	ADDTEXTURE(1);
	ADDTEXTURE(2);
	ADDTEXTURE(3);
	ADDTEXTURE(4);
	ADDTEXTURE(5);
	ADDTEXTURE(6);
      }
  }
  //glDisable(GL_TEXTURE_2D);

  if (readAsBool("shaded"))
    sortInViewDepth();

  readAsSignal("renderIn");
  if (readAsInt("renderIn") == Gbb::Selecting)
    {
      plugOut.writeAsSignal();
      return;
    }
  plugOut.writeAsSignal();

  const std::list<SmokeParticle> *sp = 
    static_cast<std::list<SmokeParticle> *>(readAsCustom("sp"));
  std::list<SmokeParticle>::const_iterator pit;

  const double	ttb(readAsDouble("ttb"));
  const double	ttl(readAsDouble("ttl"));
  const double	ttd(readAsDouble("ttd"));

  if (!readAsBool("renderEllipse"))
    {
      for (pit = sp->begin(); pit != sp->end(); ++pit) //DEBUG: REM
	pit->glRenderEllipsoid(ttb, ttl, ttd); //DEBUG: REM
      return;
    }

  const Pnt3d	&eye(readAsPnt3d("eye"));

  GLdouble		*parcelQuads, *iteQuads;
  GLdouble		*parcelTex, *iteTex;
  GLdouble		*parcelColor, *iteColor;
  const double	rescaleTex(readAsDouble("rescaleTex"));

  parcelQuads = new GLdouble[12 * sp->size()];
  parcelColor = new GLdouble[16 * sp->size()];
  parcelTex = new GLdouble[8 * sp->size()];

  //SmokeShader *shaders TODO: build shaders
  // One shader per generator.

  std::list<DagNode*>			smokeGenerator;
  std::list<DagNode*>::iterator	sit;
  std::map<int, SmokeShader>	shaders;

  DagNode::root().getNodesByType("SmokeGenerator", smokeGenerator);

  for (sit = smokeGenerator.begin(); sit != smokeGenerator.end(); ++sit)
    {
      const int	shadei = (*sit)->getInstNbr();

      shaders[shadei].col = (*sit)->readAsRgb3d("color");
      shaders[shadei].colDiffuse = (*sit)->readAsRgb3d("colorDiffuse");
      shaders[shadei].colShadow = (*sit)->readAsRgb3d("colorShadow");
      shaders[shadei].colFast = (*sit)->readAsRgb3d("colorFast");
      shaders[shadei].speedSlow = (*sit)->readAsDouble("speedSlow");
      shaders[shadei].speedFast = (*sit)->readAsDouble("speedFast");
      shaders[shadei].diffuse = (*sit)->readAsDouble("diffuse");
      shaders[shadei].refraction = (*sit)->readAsDouble("refraction");
      shaders[shadei].shadow = (*sit)->readAsDouble("shadow");
      shaders[shadei].l = Vec3d(0.0, 0.0, 1.0);
    }

  for (pit = sp->begin(), iteQuads = parcelQuads, iteTex = parcelTex,
	 iteColor = parcelColor;
       pit != sp->end(); ++pit, iteQuads += 12, iteTex += 8, iteColor += 16)
    {
      pit->glRenderEllipse(eye, iteQuads, iteColor, ttb, ttl, ttd, rescaleTex,
			   shaders[pit->shadei]);
      iteTex[0] = iteTex[2] = iteTex[3] = iteTex[5] = 1.0;
      iteTex[1] = iteTex[4] = iteTex[6] = iteTex[7] = 0.0;
    }

  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);
  glDepthMask(GL_FALSE);
  glEnable(GL_TEXTURE_2D);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glBindTexture(GL_TEXTURE_2D, (*texName)[readAsInt("dat")]);
  glEnable(GL_BLEND);

  glBlendFuncSeparate = (PFNGLBLENDFUNCSEPARATEPROC)
    getProcAddress("glBlendFuncSeparateEXT");
  if (glBlendFuncSeparate)
    glBlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_ONE, GL_ONE);
  else
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glEnableClientState(GL_COLOR_ARRAY);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  glEnableClientState(GL_VERTEX_ARRAY);

  glTexCoordPointer(2, GL_DOUBLE, 0, parcelTex);
  glColorPointer(4, GL_DOUBLE, 0, parcelColor);
  glVertexPointer(3, GL_DOUBLE, 0, parcelQuads);
  glDrawArrays(GL_QUADS, 0, sp->size() * 4);

  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  glDisableClientState(GL_COLOR_ARRAY);

  glDisable(GL_TEXTURE_2D);
  glDepthMask(GL_TRUE);

  glPopAttrib();//GL_LIGHTING_BIT

  delete[] parcelQuads;
  delete[] parcelColor;
  delete[] parcelTex;
}

// ************************************************************************* //
void	NSmokeParticles::glRenderSpheres(Plug &plugOut)
{
  readAsSignal("renderIn");
  plugOut.writeAsSignal();

  const double	size(readAsDouble("size"));
  const std::list<SmokeParticle> *sp = 
    static_cast<std::list<SmokeParticle> *>(readAsCustom("sp"));
  std::list<SmokeParticle>::const_iterator pit;

  for (pit = sp->begin(); pit != sp->end(); ++pit)
    pit->glRenderSphere(size);
}

// ************************************************************************* //
void	NSmokeParticles::compute(Plug &plugOut)
{
  if (plugOut.shortName() == "renderOut")
    glRender(plugOut);
  else if (plugOut.shortName() == "update")
    update(plugOut);
  else if (plugOut.shortName() == "bbox")
    {
      std::list<SmokeParticle>::iterator spit;
      std::list<SmokeParticle> *sp =
	static_cast<std::list<SmokeParticle> *>(readAsCustom("sp"));
      BBox3d	bbox;

      readAsInt("f");
      if (sp->empty())
	bbox = BBox3d(Pnt3d(0.0));
      else
	{
	  bbox = BBox3d(sp->front().p);
	  for (spit = sp->begin(); spit != sp->end(); ++spit)
	    bbox.fill(spit->p);
	}
      plugOut.writeAsCustom(&bbox);
    }
  else if (plugOut.shortName() == "nbParticles")
    {
      const std::list<SmokeParticle> *sp = 
	static_cast<const std::list<SmokeParticle> *>(readAsCustom("sp"));

      plugOut.writeAsInt(sp->size());
    }
}

// ************************************************************************* //
/// DESCRIPTION: compute the log of twist at p using the octree.
/// NOTE: Points outside of the octree may be queried because of forward finite
///   difference evaluation. Those are forced inside.
// ************************************************************************* //
void	NSmokeParticles::computeRlogOctree(const Pnt3d	&p,
					   TwistLog	&rlog)
{
  Octree<TwistLog> &octree = 
    *static_cast<Octree<TwistLog>*>(readAsCustom("octree"));
  const OLeaf<TwistLog>	*rleaf[8];

  octree.query8Neighbors(p, rleaf);

  for (unsigned i(0); i < 8; ++i)
    if (!rleaf[i])
      {
	rlog = TwistLog(0.0);
	//std::cerr << "Error: no octree node at " << p << "." << std::endl;
	return;
      }

  double x(0.0), y(0.0), z(0.0); // in [0,1]

  if (rleaf[1]->p.x != rleaf[0]->p.x)
    x = (p.x - rleaf[0]->p.x) / (rleaf[1]->p.x - rleaf[0]->p.x);
  if (rleaf[2]->p.y != rleaf[0]->p.y)
    y = (p.y - rleaf[0]->p.y) / (rleaf[2]->p.y - rleaf[0]->p.y);
  if (rleaf[4]->p.z != rleaf[0]->p.z)
    z = (p.z - rleaf[0]->p.z) / (rleaf[4]->p.z - rleaf[0]->p.z);
  
  TwistLog	v00x = rleaf[0]->elm * (1 - x) + rleaf[1]->elm * x;
  TwistLog	v01x = rleaf[2]->elm * (1 - x) + rleaf[3]->elm * x;
  TwistLog	v10x = rleaf[4]->elm * (1 - x) + rleaf[5]->elm * x;
  TwistLog	v11x = rleaf[6]->elm * (1 - x) + rleaf[7]->elm * x;

  TwistLog	v0yx = v00x * (1.0 - y) + v01x * y;
  TwistLog	v1yx = v10x * (1.0 - y) + v11x * y;

  TwistLog	vzyx = v0yx * (1.0 - z) + v1yx * z;

  rlog = vzyx;
}

// ************************************************************************* //
/// DESCRIPTION: Computes the jacobian matrix of plugRlog(plugp).
// ************************************************************************* //
void	NSmokeParticles::computeJac(const Pnt3d	&q,
				    Mat3d	&jac)
{
  DagNode	&clock(*DagNode::getNodeByPath(":Clock0"));
  const double	dt(clock.readAsDouble("dt"));
  //const double	dt(readAsDouble("dt"));
  Pnt3d		p(q), px(q), py(q), pz(q);
  TwistLog	rlog, rlogx, rlogy, rlogz;

  px.x += 1.0e-8;
  py.y += 1.0e-8;
  pz.z += 1.0e-8;
  computeRlogOctree(p, rlog);
  computeRlogOctree(px, rlogx);
  computeRlogOctree(py, rlogy);
  computeRlogOctree(pz, rlogz);
  (rlog * dt).exp(p);
  (rlogx * dt).exp(px);
  (rlogy * dt).exp(py);
  (rlogz * dt).exp(pz);
  jac = Mat3d(px - p, py - p, pz - p) * 1.0e8;
}
