// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:30 $
/// $Revision: 1.1 $
// ************************************************************************* //
#ifndef __NODESMOKEPARTICLES_HH__
#define __NODESMOKEPARTICLES_HH__

// ************************************************************************* //
#include "gfxtypes.h"
#include "plug.h"
#include "dagnode.h"

// ************************************************************************* //
#define ADDTEXTURE(i) \
(*texName)[i - 1] = texNameForGen[i - 1];\
glBindTexture(GL_TEXTURE_2D, (*texName)[i - 1]);\
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);\
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);\
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);\
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);\
glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 64, 64, 0, GL_RGBA,\
	     GL_UNSIGNED_BYTE, parcel##i##_dat);

// ************************************************************************* //
struct SmokeShader
{
  Rgb3d		col;// Color of particles
  Rgb3d		colDiffuse;// Color of diffuse reflection
  Rgb3d		colShadow;// Color of shadow
  Rgb3d		colFast;// Color for fast moving particles
  double	speedSlow;// Max Velocity of slow particles
  double	speedFast;// Min Velocity of fast particles
  double	diffuse;//in [0, 1] amount of diffuse reflection
  double	shadow;// in [0, 1]. Intensity of shadow (1 is intense)
  double	refraction;// Light vector, pointing towards light
  Vec3d		l;// Light vector, pointing towards light
public:
  SmokeShader() {}
};

// ************************************************************************* //
/// DESCRIPTION: Parcel / particle of smoke.
// ************************************************************************* //
class SmokeParticle
{
public:
  Pnt3d		p;// Position
  Vec3d		v;// Velocity
  Mat3d		cov;// Matrix of covariance for split.
  double	mass;// Quantity of smoke within ellipsoid.
  double	t;// Time of life (to check against ttb, ttl and ttd).
  double	z;// Depth for sorting
  double	i;// Light intensity
  int		shadei;// Index of shader, held in generator.
public:
  SmokeParticle() {}
  SmokeParticle(const Pnt3d &p_, const double size, const double mass_):
    p(p_), v(0.0), cov(Mat3d(sq(size), 0.0, 0.0,
			     0.0, sq(size), 0.0,
			     0.0, 0.0, sq(size))),
    mass(mass_), t(0.0), z(0.0), i(1.0) {}
  SmokeParticle(const Pnt3d &p_, const SmokeParticle &sp):
    p(p_), v(sp.v), cov(sp.cov), mass(sp.mass), t(sp.t), z(0.0), i(1.0),
    shadei(sp.shadei) {}
  friend std::ostream	&operator<<(std::ostream &os, const SmokeParticle &sp)
  { return os << sp.p << ' ' << sp.cov << ' ' << sp.t; }
  friend std::istream	&operator>>(std::istream &is, SmokeParticle &sp)
  { return is >> sp.p >> sp.cov >> sp.t; }
  void		glRenderEllipse(const Pnt3d &v, double *const parcelQuads,
				double *const parcelColor, const double ttb,
				const double ttl, const double ttd,
				const double rescale,
				const SmokeShader &shader) const;
  void		glRenderEllipsoid(const double ttb, const double ttl,
				  const double ttd) const;
  void		glRenderSphere(const double size) const;
  double	volume() const;
  double	density() const;
  double	presence(const double ttb, const double ttl,
			 const double ttd) const;
  bool	operator<(const SmokeParticle &sp) const { return z < sp.z; }
};

// ************************************************************************* //
/// DESCRIPTION: Extends a data node. This is stored in the scene graph.
// ************************************************************************* //
class PlugSmokeParticles: public Plug
{
protected:
  std::list<SmokeParticle>	*_m_data;
public:
  static int	typeNbr;
  static String	typeStr;
public:
  PlugSmokeParticles(DataBlock &db): Plug(db), _m_data(0) { newData(); }
  virtual ~PlugSmokeParticles() { freeData(); }
  static Plug	*creator(DataBlock &db) { return new PlugSmokeParticles(db); }
  virtual int	getTypeNbr() const { return typeNbr; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual void	freeData() { if (_m_data) delete _m_data; _m_data = 0; }
  virtual void	newData()
  { if (!_m_data) _m_data = new std::list<SmokeParticle>; }
  virtual void	saveData(std::ostream &os);
  virtual void	loadData(std::istream &is);
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const std::list<SmokeParticle>*>(v); }
  virtual void	**rawdata() { return (void**)(&_m_data); }
};

// ************************************************************************* //
/// DESCRIPTION: Particles of smoke.
// ************************************************************************* //
class NSmokeParticles: public DagNode
{
public:
  static int	typeNbr;
  static String	typeStr;
public:
  NSmokeParticles(DagNode *parent = 0): DagNode(parent) {}
  virtual void		initialize();
  virtual int		getTypeNbr() const { return typeNbr; }
  virtual const String	&getTypeStr() const { return typeStr; }
  static DagNode	*creator() { return new NSmokeParticles; }
  void			buildDagNodes();
  virtual void		compute(Plug &hndOut);
  void			glRender(Plug &hndOut);
  void			glRenderSpheres(Plug &hndOut);
  void			update(Plug &hndOut);
  void			updateReset();
  void			updateMove();
  void			updateDelete();
  void			updateSplit();
  void			sortInViewDepth();
  void			shade();
  // *** //
  void			computeRlogOctree(const Pnt3d &p, TwistLog &rlog);
  void			computeJac(const Pnt3d &q, Mat3d &jac);
};

// ************************************************************************* //
#endif//__NODESMOKEPARTICLES_HH__
