// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:30 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include "gbb.h"
#include "datablock.h"
#include "plug.h"
#include "command.h"
#include "nodeturbulence.h"

// ************************************************************************* //
int	NTurbulence::typeNbr(84);
String	NTurbulence::typeStr("Turbulence");
int	PlugTurbulence::typeNbr(79);
String	PlugTurbulence::typeStr("PlugTurbulence");

// ************************************************************************* //
void	NTurbulence::initialize()
{
  NShape::initialize();

  DataBlock	&db(getDataBlock());

  db.addInt(0, MININT, MAXINT, "f", false);
  db.addDouble(0.0, 0.0, MAXDOUBLE, "dt", false);
  db.addDouble(0.5, 0.0, 1.0, "radius");
  db.addDouble(2.0, 0.0, MAXDOUBLE, "strength");
  db.addInt(32, 1, MAXINT, "size");
  db.addInt(2, 1, MAXINT, "repeat");
  db.addSignal("update", false);
  db.addMat4d(Mat4d::identity(), "matParent", false);//position
  db.addMat4d(Mat4d::identity(), "imatParent", false);//position inverse
  db.addCustom(*new PlugTurbulence(db), "dtl", false);
  db.addPnt3d(0.0, MINDOUBLE, MAXDOUBLE, "p", false);
  db.addTwistLog(0.0, "rlog", false);

  db.affect("p", "rlog");
  db.affect("matParent", "imatParent");
  db.affect("f", "update");

  buildDagNodes();
}

// ************************************************************************* //
void	NTurbulence::buildDagNodes()
{
  DagNode	*oldParent(_m_parent);
  DagNode	*tr = Command::run(String("new -type Transform -path :"));
  DagNode	*clock = DagNode::getNodeByPath(":Clock0");

  if (oldParent)
    DagNode::parent(*oldParent, *tr);
  DagNode::parent(*tr, *this);
  DagNode::connect(*oldParent, "renderIn", *tr, "renderIn");
  DagNode::connect(*tr, "renderIn", *this, "renderIn");
  DagNode::connect(*tr, "matChild", *this, "matParent");
  DagNode::connect(*clock, "f", *this, "f");
  DagNode::connect(*clock, "dt", *this, "dt");
}

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
void	NTurbulence::glRender(Plug &plugOut)
{
  if (readAsInt("renderIn") != Gbb::Rendering)
    {
      pushName();
      glPushMatrix();
      glPushAttrib(GL_LIGHTING_BIT | GL_DEPTH_BUFFER_BIT | GL_ENABLE_BIT);
      glDisable(GL_LIGHTING);
      readAsMat4d("matParent").glMultMatrix();

      DataTurbulence *dtl = static_cast<DataTurbulence*>(readAsCustom("dtl"));
      const double   repeat = readAsInt("repeat");
      BBox3d	box(0.0, 1.0);
      const int	sel(readAsInt("sel"));
      Rgb3d	col;

      if (sel == 1)
	col = Rgb3d(1.0, 0.0, 5.0);
      else if (sel)
	col = Rgb3d(1.0, 0.5, 0.0);
      else
	col = Rgb3d(0.0, 0.5, 1.0);

      // Shape ************************************************************* //
      glEnable(GL_POLYGON_OFFSET_FILL);
      glEnable(GL_POLYGON_OFFSET_LINE);
      glPolygonOffset(1.0, 1.0);
      glColor4d(col.r, col.g, col.b, 0.2);
      glDepthMask(GL_FALSE);
      box.glRenderSmooth();
      glDepthMask(GL_TRUE);
      glDisable(GL_POLYGON_OFFSET_LINE);
      glDisable(GL_POLYGON_OFFSET_FILL);

      glColor3d(col.r, col.g, col.b);
      box.glRender();

      glScaled(1.0 / repeat, 1.0 / repeat, 1.0 / repeat);
      dtl->glRender();

      glPopAttrib();//GL_LIGHTING_BIT | GL_DEPTH_BUFFER_BIT | GL_ENABLE_BIT
      glPopMatrix();
      popName();
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
/// DESCRIPTION: Computes the force at p in [0,1]x[0,1]x[0,1].
// ************************************************************************* //
TwistLog	DataTurbulence::rlogTorus(const Pnt3d &p) const
{
  std::list<TVortex>::const_iterator	dit;
  TwistLog	rlog(0.0);
  Pnt3d		c;
  double	d2;

  for (dit = _m_data.begin(); dit != _m_data.end(); ++dit)
    {

      // Looks for pi within toroidal topology ***************************** //
      for (double x(-1.0); x <= 1.0; x += 1.0)
	for (double y(-1.0); y <= 1.0; y += 1.0)
	  for (double z(-1.0); z <= 1.0; z += 1.0)
	    {
	      c = dit->p + Vec3d(x, y, z);
	      d2 = (c - p).norm2();
	      if (d2 < _m_r2)
		{
		  rlog += TwistLog(dit->w, Vec3d(c) ^ dit->w) * 
		    sq(4.0 - 20.0 / (d2 / _m_r2 + 4.0));
		  goto doneIt;
		}
	    }
    doneIt: ;
    }
  return rlog * _m_strength;
}

// ************************************************************************* //
double	rbfT(const double	l2)
{
  return sq(4.0 - 20.0 / (l2 + 4.0));
}

// ************************************************************************* //
/// DESCRIPTION: computes a field for p. The flow contained 
///  in [0,1]^3 is tiled into [0,repeat]^3
///  p is assumed in [0,repeat]^3
///  If the real center of the vortex creal is near a boundary, the 
///   vortex's strength is decreased.
// ************************************************************************* //
TwistLog	DataTurbulence::rlog(const Pnt3d	&p,
				     const int		repeat) const
{
  const Pnt3d	cmin(floor(p.x), floor(p.y), floor(p.z));

  std::list<TVortex>::const_iterator	dit;
  TwistLog	rlog(0.0);
  Pnt3d		creal;
  double	d2;

  for (dit = _m_data.begin(); dit != _m_data.end(); ++dit)
    {
      for (double x(-1.0); x <= 1.0; x += 1.0)
	for (double y(-1.0); y <= 1.0; y += 1.0)
	  for (double z(-1.0); z <= 1.0; z += 1.0)
	    {
	      creal = cmin + dit->p + Vec3d(x, y, z);
	      d2 = (creal - p).norm2();
	      if (d2 < _m_r2)
		{
		  double	modulate(1.0);

		  if (sq(creal.x) < _m_r2)
		    modulate *= 1.0 - rbfT(sq(creal.x) / _m_r2);
		  else if (sq(creal.x - repeat) < _m_r2)
		    modulate *= 1.0 - rbfT(sq(creal.x - repeat) / _m_r2);
		  if (sq(creal.y) < _m_r2)
		    modulate *= 1.0 - rbfT(sq(creal.y) / _m_r2);
		  else if (sq(creal.y - repeat) < _m_r2)
		    modulate *= 1.0 - rbfT(sq(creal.y - repeat) / _m_r2);
		  if (sq(creal.z) < _m_r2)
		    modulate *= 1.0 - rbfT(sq(creal.z) / _m_r2);
		  else if (sq(creal.z - repeat) < _m_r2)
		    modulate *= 1.0 - rbfT(sq(creal.z - repeat) / _m_r2);
		  rlog += TwistLog(dit->w, Vec3d(creal) ^ dit->w) * 
		    sq(4.0 - 20.0 / (d2 / _m_r2 + 4.0)) * modulate;
		  goto doneIt;
		}
	    }
    doneIt: ;
    }
  return rlog * _m_strength;
}

// ************************************************************************* //
void	DataTurbulence::glRender() const
{
  std::list<TVortex>::const_iterator	dit;

  BBox3d	b(0.0, 1.0);
  const double	r(sqrt(_m_r2));

  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);
  glBegin(GL_POINTS);
  for (dit = _m_data.begin(); dit != _m_data.end(); ++dit)
    dit->p.glVertex();
  glEnd();//GL_POINTS
  glColor3d(0.0, 0.0, 0.0);
  b.glRender();
  /*
  glBegin(GL_LINES);
  for (dit = _m_data.begin(); dit != _m_data.end(); ++dit)
    {
      dit->p.glVertex();
      (dit->p + dit->w.unit() * r).glVertex();
    }
  glEnd();//GL_LINES
  */
  glPopAttrib();//GL_LIGHTING_BIT
}

// ************************************************************************* //
void	DataTurbulence::init(const int n)
{
  unsigned	seedp(0);

  _m_data.clear();
  for (int i(0); i < n; ++i)
    {
      const Pnt3d	p(rand_r(&seedp) / double(RAND_R_MAX),
					  rand_r(&seedp) / double(RAND_R_MAX),
					  rand_r(&seedp) / double(RAND_R_MAX));
      Vec3d		vn(0.0);

      vn.eqJitter(1.0);
      vn.eqUnit();      
      _m_data.push_back(TVortex(p, vn));
    }
}

// ************************************************************************* //
void	DataTurbulence::update(const double dt)
{
  std::list<TVortex>::iterator	dit;

  for (dit = _m_data.begin(); dit != _m_data.end(); ++dit)
    {
      Pnt3d	p(dit->p), px(p), py(p), pz(p);

      px.x += 1.0e-3;
      py.y += 1.0e-3;
      pz.z += 1.0e-3;

      (rlogTorus(px) * dt).exp(px);
      (rlogTorus(py) * dt).exp(py);
      (rlogTorus(pz) * dt).exp(pz);
      (rlogTorus(p) * dt).exp(p);

      const Mat3d	jac((px - p) * 1.0e3,
			    (py - p) * 1.0e3,
			    (pz - p) * 1.0e3);

      dit->w = (jac * dit->w).unit();
      if (p.x < 0.0)
	p.x += 1.0;
      else if (1.0 < p.x)
	p.x -= 1.0;
      if (p.y < 0.0)
	p.y += 1.0;
      else if (1.0 < p.y)
	p.y -= 1.0;
      if (p.z < 0.0)
	p.z += 1.0;
      else if (1.0 < p.z)
	p.z -= 1.0;
      dit->p = p;
    }

}

// ************************************************************************* //
void	NTurbulence::update(Plug &plugOut)
{
  const int	f = readAsInt("f");
  DataTurbulence *dtl = static_cast<DataTurbulence*>(readAsCustom("dtl"));

  if (f <= 1)
    dtl->init(readAsInt("size"));

  dtl->setStrength(readAsDouble("strength"));
  dtl->setRadius(readAsDouble("radius"));
  dtl->update(readAsDouble("dt"));

  plugOut.writeAsSignal();
}

// ************************************************************************* //
void	NTurbulence::compute(Plug &plugOut)
{
  if (plugOut.name() == "rlog")
    {
      const Pnt3d    plocal(readAsMat4d("imatParent") * readAsPnt3d("p"));
      
      if (plocal.x <= 0.0 || plocal.y <= 0.0 || plocal.z <= 0.0 || 
	  1.0 <= plocal.x || 1.0 <= plocal.y || 1.0 <= plocal.z)
	plugOut.writeAsTwistLog(0.0);
      else
	{
	  const int      repeati = readAsInt("repeat");
	  const double   repeatd = double(repeati);
	  DataTurbulence *dtl = 
	    static_cast<DataTurbulence*>(readAsCustom("dtl"));

	  plugOut.writeAsTwistLog
	    ((dtl->rlog(plocal * repeatd, repeati)).transform
	     (readAsMat4d("matParent") *
	      Scale(1.0 / repeatd, 0.0).asMatrix()));
	}
    }
  else if (plugOut.name() == "renderOut")
    glRender(plugOut);
  else if (plugOut.name() == "update")
    update(plugOut);
  else if (plugOut.name() == "imatParent")
    plugOut.writeAsMat4d(readAsMat4d("matParent").inverse());
}
