// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:30 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <GL/glu.h>
#include "nodevortexring.h"
#include "datablock.h"
#include "command.h"
#include "plugcustom.h"
#include "gbb.h"

// ************************************************************************* //
int	NVortexRing::typeNbr(60);
String	NVortexRing::typeStr("VortexRing");

// ************************************************************************* //
/// DESCRIPTION: Chebyshev Polynomials of the First Kind, from 0 to 12
///  coefficients. For linear blending of multiple frequencies.
// ************************************************************************* //
const int	chebyshev[13][13] = {
 {  1,   0,   0,    0,    0,     0,     0,    0,     0,     0,     0,    0, 0},
 {  0,   1,   0,    0,    0,     0,     0,    0,     0,     0,     0,    0, 0},
 { -1,   0,   2,    0,    0,     0,     0,    0,     0,     0,     0,    0, 0},
 {  0,  -3,   0,    4,    0,     0,     0,    0,     0,     0,     0,    0, 0},
 {  1,   0,  -8,    0,    8,     0,     0,    0,     0,     0,     0,    0, 0},
 {  0,   5,   0,  -20,    0,    16,     0,    0,     0,     0,     0,    0, 0},
 { -1,   0,  18,    0,  -48,     0,    32,    0,     0,     0,     0,    0, 0},
 {  0,  -7,   0,   56,    0,  -112,     0,   64,     0,     0,     0,    0, 0},
 {  1,   0  -32,    0,  160,     0,  -256,    0,   128,     0,     0,    0, 0},
 {  0,   9,   0, -120,    0,   432,     0, -576,     0,   256,     0,    0, 0},
 { -1,   0,  50,    0, -400,     0,  1120,    0, -1280,     0,   512,    0, 0},
 {  0, -11,   0,  220,    0, -1232,     0, 2816,     0, -2816,     0, 1024, 0},
 {  1,   0, -72,    0,  840,     0, -3584,    0,  6912,     0, -6144,  0,2048}
};

// ************************************************************************* //
/// DESCRIPTION:
///  nbHarmonics- nb of Harmonics.
///  center	- center of circle.
///  ex, ey	- plane containing the principal ellipse.
///  ez		- normal to the plane.
///  cx, cy, cz	- frequencies.
///  l		- arclenght parameter, used if ring is controlled by path
///  dir	- direction of ring's motion. Magnitude represents uncertainty.
///  paddle	- perpendicular to dir vector, with same magnitude as "dir".
///               It is updated automaticaly, or can be set manually after
///               the update signal has been pulled.
///  m		- magnet id, if ring has been attracted by a magnet.
// ************************************************************************* //
void	NVortexRing::initialize()
{
  NShape::initialize();

  DataBlock	&db(getDataBlock());

  db.addInt(0, MININT, MAXINT, "f", false);//input
  db.addDouble(0.0, "dt", false);//input
  db.addSignal("update", false);
  db.addPnt3d(Pnt3d(0.0), "p", false);
  db.addTwistLog(TwistLog(0.0), "rlog", false);

  db.addBool(true, "displayTexture", false);
  db.addBool(true, "split", false);
  db.addDouble(1.0, 0.0, MAXDOUBLE, "minStrangle", false);

  db.addInt(2, 1, MAXINT, "nbMotionHarmonics");
  db.addInt(12, 3, MAXINT, "nbMotionSamples");
  db.addInt(24, 3, MAXINT, "nbTextureSamples");
  db.addPnt3dArray(12, Pnt3d(0.0), "motionSamples", false);
  db.addPnt3dArray(12, Pnt3d(0.0), "textureSamples", false);
  //db.addPnt3dArray(12, Pnt3d(0.0), "motionSamples", false);

  //db.addVec3dArray(12, Vec3d(0.0), "spin", false);
  db.addPnt3d(0.0, "center", false);
  db.addPnt3d(0.0, "centerPrev", false);
  db.addVec3d(0.0, "ex", false);
  db.addVec3d(0.0, "ey", false);
  db.addVec3d(0.0, "ez", false);
  db.addPnt2dArray(2, Pnt2d(0.0), "cx", false);
  db.addPnt2dArray(2, Pnt2d(0.0), "cy", false);
  db.addPnt2dArray(2, Pnt2d(0.0), "cz", false);

  db.addCustom(*new PlugBBox3d(db), "bbox", false);

  db.addDouble(0.0, "spin", false);
  db.addVec3d(0.0, "dir", false);
  db.addVec3d(0.0, "g", false);//input

  db.addDouble(1.0, 0.0, MAXDOUBLE, "strength");
  db.addDouble(0.0, MINDOUBLE, MAXDOUBLE, "twist");
  db.addDouble(0.0, 0.0, MAXDOUBLE, "temperature");
  db.addDouble(5.0, 0.0, MAXDOUBLE, "thickness");
  db.addDouble(0.2, 0.0, MAXDOUBLE, "ttb");// Time to birth
  db.addDouble(7.0, 0.0, MAXDOUBLE, "ttl");// Time to live
  db.addDouble(8.0, 0.0, MAXDOUBLE, "ttd");// Time to die
  db.addDouble(0.0, 0.0, MAXDOUBLE, "t");// Time passed
  db.addDouble(0.0, 0.0, 1.0, "presence");// Presence, in [0, 1]
  db.addDouble(0.0, 0.0, MAXDOUBLE, "l", false);//Control
  db.addInt(-1, -1, MAXINT, "m", false);//Control
  db.addDouble(0.0, 0.0, MAXDOUBLE, "oldLength", false);

  // Fake frequencies:
  db.addInt(0, 0, MAXINT, "fakeHarmonic");
  db.addDouble(0.0, 0.0, MAXDOUBLE, "fakeStrength");

  // For split
  db.addIntArray(2, 0, "strangling", false);

  // Paddle
  db.addVec3d(0.0, "paddle", false);
  db.addDouble(0.0, "paddleAmount");

  // Affect **************************************************************** //
  db.affect("f", "update");

  db.affect("nbMotionSamples", "motionSamples");
  db.affect("center", "motionSamples");
  db.affect("ex", "motionSamples");
  db.affect("ey", "motionSamples");
  db.affect("ez", "motionSamples");
  db.affect("cx", "motionSamples");
  db.affect("cy", "motionSamples");
  db.affect("cz", "motionSamples");

  db.affect("motionSamples", "strangling");

  /*
  db.affect("center", "paddleAngle");
  db.affect("ex", "paddleAngle");
  db.affect("ey", "paddleAngle");
  db.affect("ez", "paddleAngle");
  db.affect("cx", "paddleAngle");
  db.affect("cy", "paddleAngle");
  db.affect("cz", "paddleAngle");
  */

  db.affect("nbTextureSamples", "textureSamples");
  db.affect("center", "textureSamples");
  db.affect("ex", "textureSamples");
  db.affect("ey", "textureSamples");
  db.affect("ez", "textureSamples");
  db.affect("cx", "textureSamples");
  db.affect("cy", "textureSamples");
  db.affect("cz", "textureSamples");

  db.affect("motionSamples", "bbox");

  db.affect("p", "rlog");

  db.affect("t", "presence");

  buildDagNodes();
}

// ************************************************************************* //
void	NVortexRing::buildDagNodes()
{
  DagNode	*clock = DagNode::getNodeByPath(":Clock0");
  DagNode	*rs = DagNode::getNodeByPath(":VortexSet0");

  DagNode::connect(*clock, "f", *this, "f");
  DagNode::connect(*clock, "dt", *this, "dt");
  DagNode::connect(*_m_parent, "renderIn", *this, "renderIn");
  DagNode::connect(*rs, "g", *this, "g");
  //DagNode::connect(*rs, "split", *this, "split");
}

// ************************************************************************* //
void	NVortexRing::glRenderTexture()
{
  const std::vector<Pnt3d> &samples = readAsPnt3dArray("textureSamples");

  const int	sel(readAsInt("sel"));
  const double	pres(readAsDouble("presence"));

  if (sel == 1)
    glColor4d(1.0, 0.0, 0.5, pres);
  else if (sel)
    glColor4d(1.0, 0.5, 0.0, pres);
  else
    glColor4d(0.0, 0.25, 1.0, pres);
  glPushAttrib(GL_LIGHTING_BIT | GL_LINE_BIT);
  glLineWidth(2.0);
  glBegin(GL_LINE_LOOP);

  const Pnt3d &center = readAsPnt3d("center");

  for (unsigned i(0), n(samples.size()); i < n; ++i)
    {
      (samples[i] - center).glNormal();
      samples[i].glVertex();
    }
  glEnd();//GL_LINE_LOOP
  glPopAttrib();//GL_LINE_BIT | GL_CURRENT_BIT
}

// ************************************************************************* //
void	NVortexRing::glRenderMotion()
{
  const std::vector<Pnt3d> &samples = readAsPnt3dArray("motionSamples");

  const int	sel(readAsInt("sel"));
  const double	pres(readAsDouble("presence"));

  if (sel == 1)
    glColor4d(1.0, 0.0, 0.5, pres);
  else if (sel)
    glColor4d(1.0, 0.5, 0.0, pres);
  else
    glColor4d(0.0, 0.25, 1.0, pres);
  glPushAttrib(GL_LIGHTING_BIT | GL_LINE_BIT);

  glLineWidth(1.0);
  glBegin(GL_LINE_LOOP);

  const Pnt3d &center = readAsPnt3d("center");

  for (unsigned i(0), n(samples.size()); i < n; ++i)
    {
      (samples[i] - center).glNormal();
      samples[i].glVertex();
    }
  glEnd();//GL_LINE_LOOP

  glDisable(GL_LIGHTING);
  glBegin(GL_POINTS);
  for (unsigned i(0), n(samples.size()); i < n; ++i)
    samples[i].glVertex();
  glEnd();//GL_POINTS

  glPopAttrib();//GL_LINE_BIT | GL_CURRENT_BIT
}

// ************************************************************************* //
void	NVortexRing::glRender(Plug &plugOut)
{
  if (readAsInt("renderIn") != Gbb::Rendering)
    {
      pushName();
      if (readAsBool("displayTexture"))
	glRenderTexture();
      else
	glRenderMotion();
      if (readAsBool("split"))
	{
	  glPushAttrib(GL_LIGHTING_BIT);
	  glDisable(GL_LIGHTING);
	  glColor4d(1.0, 0.0, 0.0, readAsDouble("presence"));
	  glBegin(GL_LINES);
	  //motionPoint(alpha[0]).glVertex();
	  //motionPoint(alpha[1]).glVertex();
	  const std::vector<int> &strangling = 
	    readAsIntArray("strangling");
	  const std::vector<Pnt3d>	&samples = 
	    readAsPnt3dArray("motionSamples");

	  //motionPoint(strangling[0]).glVertex();
	  samples[strangling[0]].glVertex();
	  samples[strangling[1]].glVertex();
	  //samples[samples.size() / 4].glVertex();
	  //samples[(3 * samples.size()) / 4].glVertex();
	  glEnd();//GL_LINES

	  if (1.0e-6 < readAsDouble("paddleAmount"))
	    {
	      glColor3d(0.0, 0.0, 0.0);
	      glBegin(GL_LINES);
	      readAsPnt3d("center").glVertex();
	      (readAsPnt3d("center") + readAsVec3d("paddle")).glVertex();
	      glEnd();//GL_LINES
	    }
	  glPopAttrib();//GL_LIGHTING_BIT
	}
      popName();
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
/// rho dw/dt = D rho ^ (g - a)
// ************************************************************************* //
void	NVortexRing::updateBuoyancy()
{
  const double	temperature = readAsDouble("temperature");

  if (temperature == 0.0)
    return;

  const Vec3d	&g(readAsVec3d("g"));
  const Pnt3d	&center = readAsPnt3d("center");
  const std::vector<Pnt3d>	&samples(readAsPnt3dArray("motionSamples"));
  const double	strength = readAsDouble("strength");
  double	sumdw = 0.0;
  double	terminalVelocity(20.0);

  // Floating ************************************************************** //
  for (int i(0), n(samples.size()); i < n; ++i)
    {
      const Vec3d	gradRho = (samples[i] - center).unit();
      const Vec3d	w = (samples[(i + 1) % n] - samples[i]).unit();

      // cannot turn w ... do a dot product.
      sumdw += ((gradRho ^ g) * w) * temperature;
    }
  writeAsDouble("strength", min(strength + sumdw / samples.size(),
				terminalVelocity));
  // Cool down ***
  writeAsDouble("temperature", temperature * 0.9);
}

// ************************************************************************* //
/// DESCRIPTION: 
///  Advects the samples forward in time. Returns previous length.
///  
// ************************************************************************* //
void	NVortexRing::updatePosition()
{
  std::vector<Pnt3d>	samples = readAsPnt3dArray("motionSamples");
  DagNode	*rs = DagNode::getNodeByPath(":VortexSet0");
  const double	dt = readAsDouble("dt");
  Plug		&plugp(*rs->getPlug("ptotal"));
  Plug		&plugRlog(*rs->getPlug("rlogTotal"));
  Mat3d		jac(0.0);
  double	spin(0.0);

  writeAsDouble("oldLength", length());
  for (unsigned i(0), n(samples.size()); i < n; ++i)
    {
      Pnt3d	&pi(samples[i]);
      Vec3d	xi, yi, zi = (samples[(i + 1) % n] - pi).unit();
      Pnt3d	px(samples[i]), py(samples[i]), pz(samples[i]);
	
      if (fabs(zi.x) < fabs(zi.y) && fabs(zi.x) < fabs(zi.z))
	yi = Vec3d(1.0, 0.0, 0.0) - zi * zi.x;
      else if (fabs(zi.y) < fabs(zi.z))
	yi = Vec3d(0.0, 1.0, 0.0) - zi * zi.y;
      else
	yi = Vec3d(0.0, 0.0, 1.0) - zi * zi.z;
      yi.eqUnit();
      xi = yi ^ zi;
      px.x += 1.0e-8;
      py.y += 1.0e-8;
      pz.z += 1.0e-8;
      plugp.writeAsPnt3d(pi);
      (plugRlog.readAsTwistLog() * dt).exp(pi);
      plugp.writeAsPnt3d(px);
      (plugRlog.readAsTwistLog() * dt).exp(px);
      plugp.writeAsPnt3d(py);
      (plugRlog.readAsTwistLog() * dt).exp(py);
      plugp.writeAsPnt3d(pz);
      (plugRlog.readAsTwistLog() * dt).exp(pz);
      jac = Mat3d(px - pi, py - pi, pz - pi) * 1.0e8;
      spin += asin(((xi ^ (jac * xi).unit()) +
		    (yi ^ (jac * yi).unit())).norm() * 0.5);
    }
  writeAsDouble("spin", readAsDouble("spin") + spin / samples.size());
  writeAsPnt3dArray("motionSamples", samples);
}

// ************************************************************************* //
/// DESCRIPTION: Deforms samples and computes basis and harmonic coeffs.
///   Analysis (projection).
///   Since the parameterization changes, a new paddle angle consistent
///   with previous one is computed.
// ************************************************************************* //
void	NVortexRing::updateHarmonics()
{
  const Pnt3d		centerPrev = readAsPnt3d("center");
  std::vector<Pnt3d>	samples(readAsPnt3dArray("motionSamples"));
  const int		nbMotionSamples = readAsInt("nbMotionSamples");
  const unsigned	nbMotionHarmonics = readAsInt("nbMotionHarmonics");

  if (nbMotionHarmonics != readAsPnt2dArray("cx").size())
    {
      const Pnt2d	defaultValue(0.0);

      getPlug("cx")->resize(nbMotionHarmonics, &defaultValue);
      getPlug("cy")->resize(nbMotionHarmonics, &defaultValue);
      getPlug("cz")->resize(nbMotionHarmonics, &defaultValue);
    }

  Pnt3d	ave(0.0);
  Mat3d	cov(0.0);

  std::vector<Pnt2d>	cx(nbMotionHarmonics);
  std::vector<Pnt2d>	cy(nbMotionHarmonics);
  std::vector<Pnt2d>	cz(nbMotionHarmonics);
  std::vector<double>	lenaccum(nbMotionSamples);
  std::vector<double>	newLengths(nbMotionSamples);
  double		newtot(0.0);

  for (unsigned i(0), n(samples.size()); i < n; ++i)
    {
      lenaccum[i] = newtot;
      newLengths[i] = (samples[(i + 1) % n] - samples[i]).norm();
      newtot += newLengths[i];
    }
  for (unsigned i(0), n(samples.size()); i < n; ++i)
    {
      const double w = (newLengths[i] + newLengths[(i - 1 + n) % n]) * 0.5;

      ave += samples[i] * w;
    }
  ave /= newtot;
  for (unsigned i(0), n(samples.size()); i < n; ++i)
    {
      const double w = (newLengths[i] + newLengths[(i - 1 + n) % n]) * 0.5;

      cov += Mat3d(samples[i] - ave) * w;
    }
  cov /= newtot;

  double	eval[3];
  Vec3d		evec[3], ex, ey, ez;

  // Find basis ************************************************************ //
  cov.eigenDecomposition(eval, evec);
  ex = evec[0].unit();
  ey = evec[1].unit();
  ez = evec[2].unit();

  // Find dir vector for twist ********************************************* //
  Vec3d	dir(0.0);
  
  for (unsigned i(0), n(samples.size()); i < n; ++i)
    dir += ((samples[i % n] - ave) ^ (samples[(i + 1) % n] - ave)).unit() * 
      newLengths[i];
  dir /= newtot;
  writeAsVec3d("dir", dir);

  // Find harmonics ******************************************************** //
  for (unsigned k(0); k < nbMotionHarmonics; ++k)
    cx[k] = cy[k] = cz[k] = Pnt2d(0.0);
  for (unsigned i(0), n(samples.size()); i < n; ++i)
    {
      const Vec3d  tmp(samples[i] - ave);
      const Vec3d  vec(tmp * ex, tmp * ey, tmp * ez);
      const double theta = (lenaccum[i] / newtot) * M_2PI;
      const double w = (newLengths[i] + newLengths[(i - 1 + n) % n]) * 0.5;

      for (unsigned k(0); k < nbMotionHarmonics; ++k)
	{
	  const Vec2d veck(cos((k + 1) * theta), sin((k + 1) * theta));

	  cx[k] += veck * vec.x * w;
	  cy[k] += veck * vec.y * w;
	  cz[k] += veck * vec.z * w;
	}
    }
  for (unsigned k(0); k < nbMotionHarmonics; ++k)
    {
      cx[k] /= newtot;
      cy[k] /= newtot;
      cz[k] /= newtot;
    }

  writeAsPnt3d("centerPrev", centerPrev);
  writeAsPnt3d("center", ave);
  writeAsVec3d("ex", ex);
  writeAsVec3d("ey", ey);
  writeAsVec3d("ez", ez);

  // Converts 2d harmonic coefficients to polar coordinates **************** //
  for (unsigned k(0); k < nbMotionHarmonics; ++k)
    {
      const double ax = cx[k].angle();
      const double nx = cx[k].norm();
      const double ay = cy[k].angle();
      const double ny = cy[k].norm();
      const double az = cz[k].angle();
      const double nz = cz[k].norm();

      cx[k].x = ax;
      cx[k].y = nx;
      cy[k].x = ay;
      cy[k].y = ny;
      cz[k].x = az;
      cz[k].y = nz;
    }
  writeAsPnt2dArray("cx", cx);
  writeAsPnt2dArray("cy", cy);
  writeAsPnt2dArray("cz", cz);
  //writeAsDouble("strength", readAsDouble("strength") * 
  //readAsDouble("oldLength") / newtot);
}

// ************************************************************************* //
double	NVortexRing::length() const
{
  const std::vector<Pnt3d>	&samples(readAsPnt3dArray("motionSamples"));
  double	oldtot = 0.0;

  for (unsigned i(0), n(samples.size()); i < n; ++i)
    oldtot += (samples[(i + 1) % n] - samples[i]).norm();
  return oldtot;
}

// ************************************************************************* //
/// DESCRIPTION:
///  A filament is split if it is long enough, and if its strangling 
///  is small enough. The scale of the filament, thickness, is used.
// ************************************************************************* //
DagNode	*NVortexRing::updateSplit()
{
  const std::vector<Pnt3d>	&motionSamples
    (readAsPnt3dArray("motionSamples"));
  const std::vector<int>	&strangling = readAsIntArray("strangling");
  const double	theta0 = strangling[0] * M_2PI / motionSamples.size();
  const double	theta1 = strangling[1] * M_2PI / motionSamples.size();
  const double	l2 = (motionPoint(theta1) - motionPoint(theta0)).norm2();
  const double	llong = length();
  const double	thickness(readAsDouble("thickness"));

  if (l2 < sq(2.0 * readAsDouble("minStrangle") / thickness) &&
      thickness < llong)
    {
      DagNode	*n0 = this;
      DagNode	*n1 = Command::run(String("new -type VortexRing -path ") +
				  getNodeParent()->path());
      std::vector<Pnt3d>	motionSamples0(motionSamples.size());
      std::vector<Pnt3d>	motionSamples1(motionSamples.size());

      const std::vector<Pnt3d>	&textureSamples
	(readAsPnt3dArray("textureSamples"));
      std::vector<Pnt3d>	textureSamples0(textureSamples.size());
      std::vector<Pnt3d>	textureSamples1(textureSamples.size());

      for (int i(0), n(motionSamples.size()); i < n; ++i)
	{
	  motionSamples0[i] = motionPoint(theta0 + i * (theta1 - theta0) / n);
	  motionSamples1[i] = motionPoint(theta1 + i * (theta0 + M_2PI -
							theta1) / n);
	}

      for (int i(0), n(textureSamples.size()); i < n; ++i)
	{
	  textureSamples0[i] = texturePoint(theta0 + i * (theta1 -
							  theta0) / n);
	  textureSamples1[i] = texturePoint(theta1 + i * (theta0 + M_2PI -
							  theta1) / n);
	}

      n0->writeAsPnt3dArray("motionSamples", motionSamples0);
      n0->writeAsPnt3dArray("textureSamples", textureSamples0);

      n1->writeAsPnt3dArray("motionSamples", motionSamples1);
      n1->writeAsPnt3dArray("textureSamples", textureSamples1);
      n1->writeAsBool("displayTexture", readAsBool("displayTexture"));
      n1->writeAsInt("nbMotionHarmonics", readAsInt("nbMotionHarmonics"));
      n1->writeAsInt("nbMotionSamples", readAsInt("nbMotionSamples"));
      n1->writeAsInt("nbTextureSamples", readAsInt("nbTextureSamples"));

      n1->getPlug("motionSamples")->clean();
      n1->getPlug("textureSamples")->clean();

      n1->writeAsDouble("spin", readAsDouble("spin"));
      n1->writeAsVec3d("dir", readAsVec3d("dir"));
      n1->writeAsDouble("strength", readAsDouble("strength"));

      n1->writeAsBool("split", readAsBool("split"));
      DagNode::connect(*getPlug("twist")->getInputNode(), "twist",
		       *n1, "twist");
      DagNode::connect(*getPlug("displayTexture")->getInputNode(),
		       "displayTexture", *n1, "displayTexture");

      n1->writeAsDouble("temperature", readAsDouble("temperature"));
      n1->writeAsDouble("thickness", thickness);
      n1->writeAsDouble("ttb", readAsDouble("ttb"));
      n1->writeAsDouble("ttl", readAsDouble("ttl"));
      n1->writeAsDouble("ttd", readAsDouble("ttd"));
      n1->writeAsDouble("l", readAsDouble("l"));
      n1->writeAsInt("m", readAsInt("m"));
      n1->writeAsInt("fakeHarmonic", readAsInt("fakeHarmonic"));
      n1->writeAsDouble("fakeStrength", readAsDouble("fakeStrength"));
      n1->writeAsDouble("oldLength", static_cast<NVortexRing*>(n1)->length());

      return n1;
    }
  return 0;
}

// ************************************************************************* //
/// DESCRIPTION: C1 piecewise function, numerically very close to C2.
///  Also numerically almost symmetric.
/// Old kernel: 1.0 + lr * lr * lr * (lr * (15.0 - 6.0 * lr) - 10.0);
// ************************************************************************* //
double	rbf(const Pnt3d		&p,
	    const Pnt3d		&c,
	    const double	r2)
{
  const double	l2 = (c - p).norm2();
  
  if (r2 < l2)
    return 0.0;
  return sq(4.0 - 20.0 / (l2 / r2 + 4.0));
}

// ************************************************************************* //
/// DESCRIPTION: C1 piecewise function, numerically very close to C2.
///  Also numerically almost symmetric.
/// Old kernel: 1.0 + lr * lr * lr * (lr * (15.0 - 6.0 * lr) - 10.0);
// ************************************************************************* //
double	rbf(const double l2)
{
  if (1.0 < l2)
    return 0.0;
  return sq(4.0 - 20.0 / (l2 + 4.0));
}


// ************************************************************************* //
/// DESCRIPTION: 
///   Synthesis of motion samples.
// ************************************************************************* //
void	NVortexRing::computeMotionSamples(Plug &plugOut)
{
  const unsigned	nbMotionSamples = readAsInt("nbMotionSamples");

  if (nbMotionSamples != plugOut.size())
    plugOut.resize(nbMotionSamples);

  std::vector<Pnt3d>	samples(plugOut.size());

  const Pnt3d	&center = readAsPnt3d("center");
  const Vec3d	&ex(readAsVec3d("ex"));
  const Vec3d	&ey(readAsVec3d("ey"));
  const Vec3d	&ez(readAsVec3d("ez"));
  const std::vector<Pnt2d>	&cx = readAsPnt2dArray("cx");
  const std::vector<Pnt2d>	&cy = readAsPnt2dArray("cy");
  const std::vector<Pnt2d>	&cz = readAsPnt2dArray("cz");
  const unsigned	nbMotionHarmonics(readAsInt("nbMotionHarmonics"));

  for (unsigned i(0); i < nbMotionSamples; ++i)
    {
      const double	alpha(i * M_2PI / nbMotionSamples);

      samples[i] = center;
      for (unsigned k(0); k < nbMotionHarmonics; ++k)
	{
	  const double	a = (k + 1) * alpha;

	  samples[i] += 2.0 * (ex * (cos(a - cx[k].x) * cx[k].y) +
			       ey * (cos(a - cy[k].x) * cy[k].y) +
			       ez * (cos(a - cz[k].x) * cz[k].y));
	}
    }
  plugOut.writeAsPnt3dArray(samples);
}

// ************************************************************************* //
/// DESCRIPTION: 
///   Synthesis of texture samples.
// ************************************************************************* //
void	NVortexRing::computeTextureSamples(Plug &plugOut)
{
  const unsigned	nbTextureSamples = readAsInt("nbTextureSamples");

  if (nbTextureSamples != plugOut.size())
    plugOut.resize(nbTextureSamples);

  std::vector<Pnt3d>	samples(plugOut.size());

  const Pnt3d	&center = readAsPnt3d("center");
  const Vec3d	&ex(readAsVec3d("ex"));
  const Vec3d	&ey(readAsVec3d("ey"));
  const Vec3d	&ez(readAsVec3d("ez"));
  const std::vector<Pnt2d>	&cx = readAsPnt2dArray("cx");
  const std::vector<Pnt2d>	&cy = readAsPnt2dArray("cy");
  const std::vector<Pnt2d>	&cz = readAsPnt2dArray("cz");
  const unsigned	nbMotionHarmonics(readAsInt("nbMotionHarmonics"));
  const unsigned	fakeHarmonic(readAsInt("fakeHarmonic"));
  const double		fakeStrength(readAsDouble("fakeStrength"));
  const Vec3d		&dir(readAsVec3d("dir"));
  const double		dir1 = dir.norm();
  const Vec3d		dirUnit = (dir1 == 0.0)? 0.0: (dir / dir1);

  for (unsigned i(0); i < nbTextureSamples; ++i)
    {
      const double	alpha(i * M_2PI / nbTextureSamples);

      samples[i] = center;
      for (unsigned k(0); k < nbMotionHarmonics; ++k)
	{
	  const double	a = (k + 1) * alpha;

	  samples[i] += 2.0 * (ex * (cos(a - cx[k].x) * cx[k].y) +
			       ey * (cos(a - cy[k].x) * cy[k].y) +
			       ez * (cos(a - cz[k].x) * cz[k].y));
	}

      unsigned	k = fakeHarmonic;

      if (0 < k && 1.0e-6 < dir1)
	{
	  const double	a = -readAsDouble("spin");
	  const double	cosa = cos(a), sina = sin(a);
	  const Vec3d	x0 = (samples[i] - center).unit();
	  const Vec3d	&z0 = dirUnit;
	  const Vec3d	x1 = x0 * cosa + z0 * sina;
	  //const Vec3d	z1 = x0 * (-sina) + z0 * cosa;
	  //samples[i] += 2.0 * (x1 * cos((k + 1) * alpha) +
	  //		       z1 * sin((k + 1) * alpha)) * 

	  samples[i] += 2.0 * (x1 * cos((k + 1) * alpha)) * 
	    fakeStrength * dir1;
	}
    }
  plugOut.writeAsPnt3dArray(samples);
}

// ************************************************************************* //
/// DESCRIPTION: According to the particle's clock, returns the presence 
///  of the particle in [0, 1]. 1 is present, 0 is absent.
// ************************************************************************* //
void	NVortexRing::computePresence(Plug &plugOut)
{
  const double ttb(readAsDouble("ttb"));
  const double ttl(readAsDouble("ttl"));
  const double ttd(readAsDouble("ttd"));
  const double t(readAsDouble("t"));

  if (t <= 0.0 || ttd <= t)
    {
      plugOut.writeAsDouble(0.0);
      return;
    }
  if (ttl < t) // fade out particle before disapearing.
    {
      const double	l((t - ttl) / (ttd - ttl));

      //return 1.0 + l * l * l * (l * (15.0 - 6.0 * l) - 10.0);
      plugOut.writeAsDouble(1.0 - l * l * (3.0 - 2.0 * l));
      return;
    }
  if (t < ttb)
    {
      const double	l(t / ttb);

      //return -l * l * l * (l * (15.0 - 6.0 * l) - 10.0);
      plugOut.writeAsDouble(l * l * (3.0 - 2.0 * l));
      return;
    }
  plugOut.writeAsDouble(1.0);
}

// ************************************************************************* //
void	NVortexRing::outputSamples()
{
  const std::vector<Pnt3d>	&samples(readAsPnt3dArray("motionSamples"))
;
  std::cout << "createNode nurbsCurve -n "  << "curve" << getInstNbr()
	    << "_" << readAsInt("f") << ";" << std::endl;
  std::cout << "setAttr -k off \".v\";" << std::endl;
  std::cout << "setAttr \".cc\" -type \"nurbsCurve\"" << std::endl;
  std::cout << "1 " << samples.size() << " 0 no 3" << std::endl;
  std::cout << samples.size() + 1;

  for (int i(0), n(samples.size() + 1); i < n; ++i)
    std::cout << " " << i;

  std::cout << std::endl;
  std::cout << samples.size() + 1 << std::endl;
  for (int i(0), n(samples.size() + 1); i < n; ++i)
    std::cout << samples[i % samples.size()] << std::endl;
  std::cout << ";" << std::endl;
}

// ************************************************************************* //
void	NVortexRing::computeBBox(Plug &plugOut)
{
  const std::vector<Pnt3d>	&samples(readAsPnt3dArray("motionSamples"));
  BBox3d	bbox(0.0);

  if (!samples.empty())
    {
      const double	r = readAsDouble("thickness");

      bbox = BBox3d(samples.front());
      for (int i(1), n(samples.size()); i < n; ++i)
	bbox.fill(samples[i]);
      bbox.eqExpand(r);
    }
  plugOut.writeAsCustom(&bbox);
}

// ************************************************************************* //
/// DESCRIPTION: Integral over segment approximated with a 
///   Riemann sum with n intervals.
// ************************************************************************* //
void	NVortexRing::computeRLog(Plug &plugOut)
{
  const Pnt3d  &p = readAsPnt3d("p");
  const BBox3d &bbox(*static_cast<const BBox3d*>(readAsCustom("bbox")));

  if (bbox.contains(p))
    {
      const std::vector<Pnt3d>	&samples(readAsPnt3dArray("textureSamples"));
      const double	strength = readAsDouble("strength");
      const double	twist = readAsDouble("twist");
      const double	paddleAmount = readAsDouble("paddleAmount");
      const Vec3d	paddle = readAsVec3d("paddle");
      const double	r2 = sq(readAsDouble("thickness"));
      const Vec3d	&dir = readAsVec3d("dir");
      TwistLog		rlog(0.0);

      for (unsigned i(0), n(samples.size()); i < n; ++i)// For each segment. //
	{
	  const Pnt3d	p0(samples[i]), p1(samples[(i + 1) % n]);

	  if (p.dist2edge(p0, p1) < r2)
	    {
	      const Vec3d	v(p1 - p0);
	      const unsigned	m = 2 * max(unsigned(1),
					    unsigned(sqrt(v.norm2() / r2)));
	      double	c(0.0);

	      if (twist == 0.0 && paddleAmount == 0.0)
		{
		  for (unsigned k(0); k < m; ++k)
		    c += rbf(p, p0 + v * (k / (m - 1.0)), r2);
		  rlog += 
		    TwistLog(v, Vec3d(p0) ^ Vec3d(p1)) * (strength * c / m);
		}
	      else if (paddleAmount == 0.0) // Twist only
		{
		  for (unsigned k(0); k < m; ++k)
		    {
		      const Pnt3d	ck = p0 + v * (k / (m - 1.0));
		      const double	frbf = rbf(p, ck, r2);

		      rlog += TwistLog(dir, Vec3d(ck) ^ dir) *
			(twist * frbf / m);
		      c += frbf;
		    }
		  rlog += 
		    TwistLog(v, Vec3d(p0) ^ Vec3d(p1)) * (strength * c / m);
		}
	      else if (twist == 0.0)
		{
		  for (unsigned k(0); k < m; ++k)
		    c += rbf(p, p0 + v * (k / (m - 1.0)), r2) * 
		      (1.0 + (paddle * v) * paddleAmount);
		  rlog += 
		    TwistLog(v, Vec3d(p0) ^ Vec3d(p1)) * (strength * c / m);
		}

	    }
	}
      plugOut.writeAsTwistLog(rlog * readAsDouble("presence"));
    }
  else
    plugOut.writeAsTwistLog(0.0);
}

// ************************************************************************* //
void	NVortexRing::update(Plug &plugOut)
{
  const double	newt = readAsDouble("t") + readAsDouble("dt");
  DagNode	*child;

  readAsInt("f");

  // TODO: To split, compare to RBF size.
  if (readAsBool("split") && (child = updateSplit()))
    {
      static_cast<NVortexRing*>(child)->updateHarmonics();
      static_cast<NVortexRing*>(child)->updateBuoyancy();
      child->writeAsDouble("t", newt);
      child->writeAsSignal("update");
    }

  //writeAsPnt3d("paddlePoint", motionPoint(readAsDouble("paddleAngle")));
  const Vec3d	dirPrevUnit = readAsVec3d("dir").unit();

  updatePosition();
  updateHarmonics();
  updateBuoyancy();
  writeAsDouble("t", newt);

  // Update paddle ********************************************************* //
  const Vec3d	dirNext = readAsVec3d("dir");
  const Vec3d	dirNextUnit = dirNext.unit();
  const Pnt3d	&center(readAsPnt3d("center"));
  const Vec3d	paddle = readAsVec3d("paddle");
  TwistLog	rot(acos(dirPrevUnit * dirNextUnit),
		    (dirPrevUnit ^ dirNextUnit).unit(), center);
  Mat4d	mrot = rot.exp();

  writeAsVec3d("paddle", mrot * paddle * dirNext.norm());
  writeAsDouble("paddleAmount", readAsDouble("paddleAmount") * 0.95);

  plugOut.writeAsSignal();
}

// ************************************************************************* //
/// DESCRIPTION: Returns a point on the motion curve.
// ************************************************************************* //
Pnt3d	NVortexRing::motionPoint(const double alpha) const
{
  Pnt3d		sample = readAsPnt3d("center");
  const Vec3d	&ex(readAsVec3d("ex"));
  const Vec3d	&ey(readAsVec3d("ey"));
  const Vec3d	&ez(readAsVec3d("ez"));
  const std::vector<Pnt2d>	&cx = readAsPnt2dArray("cx");
  const std::vector<Pnt2d>	&cy = readAsPnt2dArray("cy");
  const std::vector<Pnt2d>	&cz = readAsPnt2dArray("cz");
  const unsigned	nbMotionHarmonics(readAsInt("nbMotionHarmonics"));

  for (unsigned k(0); k < nbMotionHarmonics; ++k)
    {
      const double	a = (k + 1) * alpha;

      sample += 2.0 * (ex * (cos(a - cx[k].x) * cx[k].y) +
		       ey * (cos(a - cy[k].x) * cy[k].y) +
		       ez * (cos(a - cz[k].x) * cz[k].y));
    }
  return sample;
}

// ************************************************************************* //
/// DESCRIPTION: Returns a point on the synthetic curve.
// ************************************************************************* //
Pnt3d	NVortexRing::texturePoint(const double alpha) const
{
  Pnt3d		sample = motionPoint(alpha);
  const Pnt3d	&center(readAsPnt3d("center"));
  const unsigned	fakeHarmonic(readAsInt("fakeHarmonic"));
  const double	fakeStrength(readAsDouble("fakeStrength"));
  const Vec3d	&dir(readAsVec3d("dir"));
  const double	dir1 = dir.norm();
  const Vec3d	dirUnit = (dir1 == 0.0)? 0.0: (dir / dir1);
  unsigned	k = fakeHarmonic;

  if (0 < k && 1.0e-6 < dir1)
    {
      const double	a = -readAsDouble("spin");
      const double	cosa = cos(a), sina = sin(a);
      const Vec3d	x0 = (sample - center).unit();
      const Vec3d	&z0 = dirUnit;
      const Vec3d	x1 = x0 * cosa + z0 * sina;

      sample += 2.0 * (x1 * cos((k + 1) * alpha)) * 
	fakeStrength * dir1;
    }
  return sample;
}

// ************************************************************************* //
/// DESCRIPTION: Finds two parameters whose corresponding points are the 
///  closest.
// ************************************************************************* //
void	NVortexRing::computeStrangling(Plug &plugOut)
{
  const int	offsets[7][2] = { {0 ,0}, {-1, 0}, {0, -1}, {-1, -1},
				  {1, 0}, {0, 1}, {1, 1} };
  double	lenmax2(MAXDOUBLE);
  const std::vector<Pnt3d>	&samples(readAsPnt3dArray("motionSamples"));
  const int		n = samples.size();
  int alphaStart[2] = { n / 4, (3 * n) / 4 };
  int alphaMax[2] = { 0, 0 };

  for (int k = 0; k < n / 4; ++k)
    {
      for (int i = 0; i < 7; ++i)
	{
	  const int alphaTmp[2] = { alphaStart[0] + offsets[i][0],
				    alphaStart[1] + offsets[i][1] };
	  const double lentmp2 = (samples[alphaTmp[0]] -
				  samples[alphaTmp[1]]).norm2();

	  if (lentmp2 < lenmax2)
	    {
	      alphaMax[0] = alphaTmp[0];
	      alphaMax[1] = alphaTmp[1];
	      lenmax2 = lentmp2;
	    }
	}
      alphaStart[0] = alphaMax[0];
      alphaStart[1] = alphaMax[1];
    }
  std::vector<int>	strangling(2);
      
  strangling[0] = (alphaMax[0] + n) % n;
  strangling[1] = (alphaMax[1] + n) % n;
  plugOut.writeAsIntArray(strangling);
}

// ************************************************************************* //
void	NVortexRing::compute(Plug &plugOut)
{
  if (plugOut.name() == "rlog")
    computeRLog(plugOut);
  else if (plugOut.name() == "update")
    update(plugOut);
  else if (plugOut.name() == "renderOut")
    {
      /*
      // Renders the eigenvectors basis
      glBegin(GL_LINES);
      glColor3d(1.0, 0.0, 0.0);
      readAsPnt3d("center").glVertex();
      (readAsPnt3d("center") + readAsVec3d("ex")).glVertex();
      glColor3d(0.0, 1.0, 0.0);
      readAsPnt3d("center").glVertex();
      (readAsPnt3d("center") + readAsVec3d("ey")).glVertex();
      glColor3d(0.0, 0.0, 1.0);
      readAsPnt3d("center").glVertex();
      (readAsPnt3d("center") + readAsVec3d("ez")).glVertex();
      glEnd();//GL_LINES
      glBegin(GL_POINTS);
      motionPoint(readAsDouble("paddleAngle")).glVertex();
      glEnd();//GL_POINTS
      */

      glRender(plugOut);
    }
  else if (plugOut.name() == "motionSamples")
    computeMotionSamples(plugOut);
  else if (plugOut.name() == "textureSamples")
    computeTextureSamples(plugOut);
  else if (plugOut.name() == "bbox")
    computeBBox(plugOut);
  else if (plugOut.name() == "presence")
    computePresence(plugOut);
  else if (plugOut.name() == "strangling")
    computeStrangling(plugOut);
  /*
  else if (plugOut.name() == "paddleAngle")
    {
      {
	const Pnt3d	&paddlePoint = readAsPnt3d("paddlePoint");
	Pnt3d		center = readAsPnt3d("center");
	const Vec3d	&ex(readAsVec3d("ex"));
	const Vec3d	&ey(readAsVec3d("ey"));
	const Vec3d	&ez(readAsVec3d("ez"));
	const std::vector<Pnt2d>	&cx = readAsPnt2dArray("cx");
	const std::vector<Pnt2d>	&cy = readAsPnt2dArray("cy");
	const std::vector<Pnt2d>	&cz = readAsPnt2dArray("cz");


	const Vec3d ec = (ex * (cx[0].y * cos(cx[0].x) * 2.0) +
			  ey * (cy[0].y * cos(cy[0].x) * 2.0) + 
			  ez * (cz[0].y * cos(cz[0].x) * 2.0));
	const Vec3d es = (ex * (cx[0].y * sin(cx[0].x) * 2.0) +
			  ey * (cy[0].y * sin(cy[0].x) * 2.0) + 
			  ez * (cz[0].y * sin(cz[0].x) * 2.0));
	double	a = 0.2;
	double	fa = MAXDOUBLE;

	//paddlePoint = center + ec * cos(a) + sin(a) * es;
	//paddlePoint = center + ec * cos(a) + sin(a) * es;
	while (1.0e-12 < fa)
	  {
	    const double cosa = cos(a);
	    const double sina = sin(a);

	    fa = (center + ec * cosa + es * sina - paddlePoint) *
	      (es * cosa - ec * sina);

	    const double fpa = 
	      (es * cosa - ec * sina) * (es * cosa - ec * sina) - 
	      (center - paddlePoint + ec * cosa + es * sina) *
	      (es * sina + ec * cosa);

	    a -= fa / fpa;
	    //
	  }
	std::cout << a << std::endl;
	plugOut.writeAsDouble(a);
	// Find a that Minimizes:
	// center - paddlePoint + cosa * ec + sina * es;

	//TODO: find point!!!
      }

      //!!!

    }
  */
}
