// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:30 $
/// $Revision: 1.1 $
// ************************************************************************* //
#ifndef	__PLUG_HH__
#define	__PLUG_HH__

// ************************************************************************* //
/// Class Diagram of Plugs:
///   -------
///  [ Plug  ]
///   -------
///     |
///     +---------+--------------+------....
///     |         |              |
///   -------   ------------   ----------
///  [ TPlug ] [ PlugSignal ] [ PlugBool ]  ....
///   -------   ------------   ----------
///     |
///     +--------------+---------...
///     |              |
///   -----------   -----------
///  [ PlugPnt2d ] [ PlugPnt3d ] ....
///   -----------   -----------
// ************************************************************************* //
#include <iostream>
#include <list>
#include <vector>
#include <map>
#include "gfxtypes.h"
#include "twistlog.h"
#include "mystring.h"

// ************************************************************************* //
class Plug;
class TPlug;
class DataBlock;
class DagNode;
class PlugVec3d;
class PlugVec3dArray;

// ************************************************************************* //
#define ID_SIGNAL	1
#define ID_BOOL		2
#define ID_INT		3
#define ID_DOUBLE	4
#define ID_STRING	5
#define ID_PNT2D	6/*Recursive*/
#define ID_MAT2I	7/*Recursive*/
#define ID_PNT3D	8/*Recursive*/
#define ID_VEC3D	9/*Recursive*/
#define ID_RGB3D	10/*Recursive*/
#define ID_MAT3D	11/*Recursive*/
#define ID_TWISTLOG	12/*Recursive*/
#define ID_MAT4D	13/*Recursive*/
#define ID_INTARRAY	14/*Recursive*/
#define ID_DOUBLEARRAY	15/*Recursive*/
#define ID_PNT2DARRAY	16/*Recursive*/
#define ID_PNT3DARRAY	17/*Recursive*/
#define ID_VEC3DARRAY	18/*Recursive*/
#define ID_TWISTLOGARRAY	19/*Recursive*/
#define ID_PNT2DLIST	20/*Recursive*/

// *********************************************************************** //
#define BIT_ISCOMPUTING	0x01
#define BIT_ISDIRTY	0x02
#define BIT_ISVISIBLE	0x04
#define BIT_ISSELECTED	0x08
#define BIT_ISSAVED	0x10

// ************************************************************************* //
/// DESCRIPTION: Extend to create new data type.
///  Parent class to factorise code as well as general handle for data of
///  any type.
///  ISCOMPUTING	- for reading data while data is being updated
///  ISDIRTY		- per data, for update
///  ISVISIBLE		- per data
/// TODO: Single type and Array type
// ************************************************************************* //
class Plug
{
  friend class TPlug;
  friend class DataBlock;
//protected:
public:
  DataBlock	&_m_dataBlock; //
  TPlug		*_m_parent;    // parent in Plug hierarchy
  String	_m_shortNamei; // instance name
  String	_m_longNamei;  // instance name
  int		_m_bits;//
  int		_m_id;  // per data, for selection (array adds one extra index)
  Plug			*_m_input;  //other block.
  std::list<Plug*>	_m_affects; //same block.
  std::list<Plug*>	_m_outputs; //other block.
public:
  Plug(DataBlock &db);
  Plug(DataBlock &db, const String &shortName, const String &longName,
       const bool isVisible, const bool isSaved, TPlug *parent = 0);
  virtual	~Plug();
  virtual void	propagateDirtyBit();
  virtual void	disconnectAll();
  void		clean();

  // Bits ****************************************************************** //
  virtual void	select(const bool on = true);
  virtual void	deselect() { _m_bits &= ~BIT_ISSELECTED; }
  virtual bool	isSelected() const { return (_m_bits & BIT_ISSELECTED); }
  virtual bool	isSelected(const int) { return isSelected(); }
  virtual void	setVisible(const bool on = true);
  bool		isVisible() const { return _m_bits & BIT_ISVISIBLE; }
  bool		isDirty() const { return _m_bits & BIT_ISDIRTY; }
  bool		isComputing() const { return _m_bits & BIT_ISCOMPUTING; }
  bool		isSaved() const { return _m_bits & BIT_ISSAVED; }

  // Gets ****************************************************************** //
  virtual bool				getAsBool() { return false; }
  virtual int				getAsInt() { return 0; }
  virtual double			getAsDouble() { return 0.0; }

  // Reads ***************************************************************** //
  virtual void				readAsSignal();
  virtual bool				readAsBool();
  virtual int				readAsInt();
  virtual double			readAsDouble();
  virtual const String			&readAsString();
  virtual const Pnt2d			&readAsPnt2d();
  virtual const Mat2i			&readAsMat2i();
  virtual const Pnt3d			&readAsPnt3d();
  virtual const Vec3d			&readAsVec3d();
  virtual const Rgb3d			&readAsRgb3d();
  virtual const Mat3d			&readAsMat3d();
  virtual const TwistLog		&readAsTwistLog();
  virtual const Mat4d			&readAsMat4d();
  virtual const std::vector<int>	&readAsIntArray();
  virtual const std::vector<double>	&readAsDoubleArray();
  virtual const std::vector<Pnt2d>	&readAsPnt2dArray();
  virtual const Pnt2d			&readAsPnt2dArray(const int i);
  virtual const std::vector<Pnt3d>	&readAsPnt3dArray();
  virtual const Pnt3d			&readAsPnt3dArray(const int i);
  virtual const std::vector<Vec3d>	&readAsVec3dArray();
  virtual const Vec3d			&readAsVec3dArray(const int i);
  virtual const std::vector<TwistLog>	&readAsTwistLogArray();
  virtual const TwistLog		&readAsTwistLogArray(const int i);
  virtual const std::list<Pnt2d>	&readAsPnt2dList();
  virtual void				*readAsCustom();

  // Writes **************************************************************** //
  virtual void		writeAsSignal();
  virtual void		writeAsBool(const bool);
  virtual void		writeAsInt(const int);
  virtual void		writeAsDouble(const double);
  virtual void		writeAsString(const String &s);
  virtual void		writeAsPnt2d(const Pnt2d &);
  virtual void		writeAsMat2i(const Mat2i &);
  virtual void		writeAsPnt3d(const Pnt3d &);
  virtual void		writeAsVec3d(const Vec3d &);
  virtual void		writeAsRgb3d(const Rgb3d &);
  virtual void		writeAsMat3d(const Mat3d &);
  virtual void		writeAsTwistLog(const TwistLog &);
  virtual void		writeAsMat4d(const Mat4d &);
  virtual void		writeAsIntArray(const std::vector<int> &);
  virtual void		writeAsDoubleArray(const std::vector<double> &);
  virtual void		writeAsPnt2dArray(const std::vector<Pnt2d> &);
  virtual void		writeAsPnt2dArray(const int i, const Pnt2d &);
  virtual void		writeAsPnt3dArray(const std::vector<Pnt3d> &);
  virtual void		writeAsPnt3dArray(const int i, const Pnt3d &);
  virtual void		writeAsVec3dArray(const std::vector<Vec3d> &);
  virtual void		writeAsVec3dArray(const int i, const Vec3d &);
  virtual void		writeAsTwistLogArray(const std::vector<TwistLog> &);
  virtual void		writeAsTwistLogArray(const int i, const TwistLog &);
  virtual void		writeAsPnt2dList(const std::list<Pnt2d> &);
  virtual void		writeAsCustom(const void *value);

  // *********************************************************************** //
  DagNode	&getNode();
  DagNode	*getInputNode();
  void		getOutputNodes(std::vector<DagNode*> &onodes);
  const String	&longName() const { return _m_longNamei; }
  const String	&shortName() const { return _m_shortNamei; } 
  const String	&name() const { return _m_shortNamei; } 

  virtual bool	hasTopParentInput() const;

  // *********************************************************************** //
  virtual bool		isArray() const { return false; }
  virtual unsigned	size() const { return 1; }
  virtual void		resize(const unsigned, const void * = 0);
  int			getId() const { return _m_id; }
  virtual void		getPlugByType(const int, std::list<Plug *> &hnd);

  // *********************************************************************** //
  static std::map<String, Plug *(*)(DataBlock &)>	&creators();
  static void	registerPlug(const String &typeStr,
			     Plug *(*creator)(DataBlock &));

  // *********************************************************************** //
  virtual int		getTypeNbr() const = 0;
  virtual const String	&getTypeStr() const = 0;
  virtual void		freeData() = 0;
  virtual void		newData() = 0;
  virtual void		**rawdata() = 0;
  virtual void		linkdata(Plug &d);
  virtual void		copy(const void *) = 0;

  // Save/Load ************************************************************* //
  virtual void	saveInfo(std::ostream &os);
  virtual void	saveData(std::ostream &os) = 0;
  virtual void	saveConnections(std::ostream &os);
  virtual void	loadInfo(std::istream &is, std::map<int, void *> &v2real,
			 std::map<int, std::list<void **> > &ref2v);
  virtual void	loadData(std::istream &is) = 0;
  virtual void	loadConnections(std::istream &is,
				std::map<int, void *> &v2real,
				std::map<int, std::list<void **> > &ref2v);
  virtual void	saveAll(std::ostream &os);
  virtual Plug	*loadAll(std::istream &is, std::map<int, void *> &v2real,
			 std::map<int, std::list<void **> > &ref2v);
  void		loadIO(std::istream &is,
		       std::map<int, std::list<void **> > &ref2v);

  // Push/Pop ************************************************************** //
  virtual void	pushName() const;
  virtual void	popName() const;
};

// ************************************************************************* //
/// DESCRIPTION: Signal data.
// ************************************************************************* //
class PlugSignal: public Plug
{
public:
  static String	typeStr;
public:
  PlugSignal(DataBlock &db): Plug(db) { newData(); }
  PlugSignal(DataBlock &db, const String &shortName, const bool isVisible,
	     const bool isSaved, TPlug *parent = 0):
    Plug(db, shortName, shortName, isVisible, isSaved, parent) { newData(); }
  virtual ~PlugSignal();
  static Plug	*creator(DataBlock &db) { return new PlugSignal(db); }
  virtual int	getTypeNbr() const { return ID_SIGNAL; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual void	freeData() {}
  virtual void	newData() {}
  virtual void	saveData(std::ostream &) {}
  virtual void	loadData(std::istream &) {}
  virtual void	copy(const void *) {}
  virtual unsigned	size() const { return 0; }
  virtual void	**rawdata() { static void *p(0); return &p; }
  virtual void	readAsSignal();
  virtual void	writeAsSignal();
};

// ************************************************************************* //
/// DESCRIPTION: Bool data.
// ************************************************************************* //
class PlugBool: public Plug
{
protected:
  bool	*_m_data;
public:
  static String	typeStr;
public:
  PlugBool(DataBlock &db): Plug(db), _m_data(0) { newData(); }
  PlugBool(DataBlock &db, const String &shortName, const bool isVisible,
	   const bool isSaved, TPlug *parent = 0, bool *data = 0):
    Plug(db, shortName, shortName, isVisible, isSaved, parent), _m_data(data)
  { newData(); }
  virtual ~PlugBool();
  static Plug	*creator(DataBlock &db) { return new PlugBool(db); }
  virtual int	getTypeNbr() const { return ID_BOOL; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual void	freeData()
  { if (!_m_parent && _m_data) delete _m_data; _m_data = 0; }
  virtual void	newData() { if (!_m_parent && !_m_data) _m_data = new bool; }
  virtual void	saveData(std::ostream &os) { os << *_m_data; }
  virtual void	loadData(std::istream &is) { is >> *_m_data; }
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const bool*>(v); }
  virtual void	**rawdata() { return (void**)(&_m_data); }
  virtual bool	getAsBool();
  virtual bool	readAsBool();
  virtual void	writeAsBool(const bool value);
};

// ************************************************************************* //
/// DESCRIPTION: Int data.
// ************************************************************************* //
class PlugInt: public Plug
{
protected:
  int	_m_minValue, _m_maxValue;
  int	*_m_data;
public:
  static String	typeStr;
public:
  PlugInt(DataBlock &db): Plug(db), _m_minValue(MININT), _m_maxValue(MAXINT),
			  _m_data(0) { newData(); }
  PlugInt(DataBlock &db, const String &shortName, const bool isVisible,
	  const bool isSaved, TPlug *parent = 0, int *data = 0):
    Plug(db, shortName, shortName, isVisible, isSaved, parent),
    _m_minValue(MININT), _m_maxValue(MAXINT), _m_data(data)
  { newData(); }
  virtual ~PlugInt();
  static Plug	*creator(DataBlock &db) { return new PlugInt(db); }
  virtual int	getTypeNbr() const { return ID_INT; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual void	freeData()
  { if (!_m_parent && _m_data) delete _m_data; _m_data = 0; }
  virtual void	newData() { if (!_m_parent && !_m_data) _m_data = new int; }
  virtual void	saveData(std::ostream &os)
  { os << *_m_data << ' ' << _m_minValue << ' ' << _m_maxValue; }
  virtual void	loadData(std::istream &is)
  { is >> *_m_data >> _m_minValue >> _m_maxValue; }
  virtual void	copy(const void *v) { *_m_data = *static_cast<const int*>(v); }
  virtual void	**rawdata() { return (void**)(&_m_data); }
  virtual int	getAsInt();
  virtual int	readAsInt();
  virtual void	writeAsInt(const int value);
  void	setMinValue(const int minValue) { _m_minValue = minValue; }
  void	setMaxValue(const int maxValue) { _m_maxValue = maxValue; }
  int	minValue() const { return _m_minValue; }
  int	maxValue() const { return _m_maxValue; }
};

// ************************************************************************* //
/// DESCRIPTION: Int data.
// ************************************************************************* //
class PlugDouble: public Plug
{
protected:
  double	_m_minValue, _m_maxValue;
  double	*_m_data;
public:
  static String	typeStr;
public:
  PlugDouble(DataBlock &db): Plug(db), _m_minValue(MINFLOAT),
			     _m_maxValue(MAXFLOAT), _m_data(0) { newData(); }
  PlugDouble(DataBlock &db, const String &shortName, const bool isVisible,
	     const bool isSaved, TPlug *parent = 0, double *data = 0):
    Plug(db, shortName, shortName, isVisible, isSaved, parent),
    _m_minValue(MINFLOAT), _m_maxValue(MAXFLOAT), _m_data(data) { newData();}
  virtual ~PlugDouble();
  static Plug	*creator(DataBlock &db) { return new PlugDouble(db); }
  virtual int	getTypeNbr() const { return ID_DOUBLE; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual void	freeData()
  { if (!_m_parent && _m_data) delete _m_data; _m_data = 0; }
  virtual void	newData() { if (!_m_parent && !_m_data) _m_data = new double; }
  virtual void	saveData(std::ostream &os)
  { os << *_m_data << ' ' << _m_minValue << ' ' << _m_maxValue; }
  virtual void	loadData(std::istream &is)
  { is >> *_m_data >> _m_minValue >> _m_maxValue; }
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const double*>(v); }
  virtual void	**rawdata() { return (void**)(&_m_data); }
  virtual double	getAsDouble();
  virtual double	readAsDouble();
  virtual void		writeAsDouble(const double value);
  void	setMinValue(const double minValue) { _m_minValue = minValue; }
  void	setMaxValue(const double maxValue) { _m_maxValue = maxValue; }
  double	minValue() const { return _m_minValue; }
  double	maxValue() const { return _m_maxValue; }
};

// ************************************************************************* //
/// DESCRIPTION: Int data.
// ************************************************************************* //
class PlugString: public Plug
{
protected:
  String	*_m_data;
public:
  static String	typeStr;
public:
  PlugString(DataBlock &db): Plug(db), _m_data(0) { newData(); }
  PlugString(DataBlock &db, const String &shortName, const bool isVisible,
	     const bool isSaved, TPlug *parent = 0, String *data = 0):
    Plug(db, shortName, shortName, isVisible, isSaved, parent), _m_data(data)
  { newData(); }
  virtual ~PlugString();
  static Plug	*creator(DataBlock &db) { return new PlugString(db); }
  virtual int	getTypeNbr() const { return ID_STRING; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual void	freeData()
  { if (!_m_parent && _m_data) delete _m_data; _m_data = 0; }
  virtual void	newData() { if (!_m_parent && !_m_data) _m_data = new String; }
  virtual void	saveData(std::ostream &os) { _m_data->saveAscii(os); }
  virtual void	loadData(std::istream &is) { _m_data->loadAscii(is); }
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const String*>(v); }
  virtual void	**rawdata() { return (void**)(&_m_data); }
  virtual const String	&readAsString();
  virtual void		writeAsString(const String &value);
};

// ************************************************************************* //
unsigned	item2index(const String &item);

// ************************************************************************* //
/// DESCRIPTION: Compound data.
// ************************************************************************* //
class TPlug: public Plug
{
  friend class DataBlock;
protected:
  std::map<unsigned, Plug*>	_m_children;
public:
  // *********************************************************************** //
  // DESCRIPTION: iterator on data.
  // *********************************************************************** //
  class iterator
  {
    friend class TPlug;
    std::map<unsigned, Plug*>::iterator	_m_cit;
    std::list<Pnt2d>::iterator	_m_pit;
    unsigned	_m_index;
  public:
    iterator() {}
    iterator(const std::map<unsigned, Plug*>::iterator &cit,
	     const std::list<Pnt2d>::iterator &pit,
	     const int i): _m_cit(cit), _m_pit(pit), _m_index(i) {}
    Plug	*&operator*() { return _m_cit->second; }
    Plug	*&operator->() { return _m_cit->second; }
    bool	operator!=(const iterator &that) const
    { return _m_cit != that._m_cit; }
    bool	operator==(const iterator &that) const
    { return _m_cit == that._m_cit; }
    iterator	&operator++() { ++_m_cit; ++_m_pit; ++_m_index; return *this;}
    iterator	&operator--() { --_m_cit; ++_m_pit; --_m_index; return *this; }
    iterator	operator++(int) { iterator ret(*this); return ++ret; }
    iterator	operator--(int) { iterator ret(*this); return ++ret; }
  };
  // *********************************************************************** //
public:
  TPlug(DataBlock &db): Plug(db) {}
  TPlug(DataBlock &db, const String &name, const bool isVisible = true,
	const bool isSaved = true, TPlug *parent = 0);

  // *********************************************************************** //
  virtual void	disconnectAll();

  // Add methods *********************************************************** //
  void		addSignal(const int index);
  void		addBool(const int index);
  void		addInt(const int index);
  void		addDouble(const int index);
  void		addString(const int index);
  void		addPnt2d(const int index);
  void		addPnt3d(const int index);
  void		addVec3d(const int index);
  void		addRgb3d(const int index);
  void		addTwistLog(const int index);
  void		add(const int index);

  // Connection ************************************************************ //
  virtual void	newData();
  virtual void	freeData();
  virtual void	linkdata(Plug &d);

  // Selects *************************************************************** //
  virtual void	deselect();
  virtual bool	isSelected(const int i);

  // Extended ************************************************************** //
  virtual const String	&getChildTypeStr() const = 0;
  virtual bool	empty() const { return size() == 0; }
  virtual void	resize(const unsigned, const void * = 0) {}
  virtual void	*subrawdata(unsigned i) = 0;
  virtual void	getPlugByType(const int type, std::list<Plug *> &hnd);

  // Iterate *************************************************************** //
  iterator	begin();
  iterator	end();
  void		push_back(const void *d);
  void		push_front(const void *d);
  void		insert(const iterator &cit, const void *d);
  void		erase(const iterator &cit);
  Plug		*&front();
  Plug		*&back();

  // Save/Load ************************************************************* //
  virtual void	loadConnections(std::istream &is,
				std::map<int, void *> &v2real,
				std::map<int, std::list<void **> > &ref2v);
  virtual void	saveAll(std::ostream &os);
  virtual Plug	*loadAll(std::istream &is, std::map<int, void *> &v2real,
			 std::map<int, std::list<void **> > &ref2v);
};

// ************************************************************************* //
/// DESCRIPTION: Int data.
// ************************************************************************* //
class PlugPnt2d: public TPlug
{
protected:
  Pnt2d	*_m_data;
public:
  static String	typeStr;
public:
  PlugPnt2d(DataBlock &db): TPlug(db), _m_data(0)
  { newData(); }
  PlugPnt2d(DataBlock &db, const String &shortName, const bool isVisible,
	    const bool isSaved, TPlug *parent = 0, Pnt2d *data = 0):
    TPlug(db, shortName, isVisible, isSaved, parent), _m_data(data)
  { newData(); }
  virtual ~PlugPnt2d();
  static Plug	*creator(DataBlock &db) { return new PlugPnt2d(db); }
  virtual int	getTypeNbr() const { return ID_PNT2D; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual const String	&getChildTypeStr() const { return PlugDouble::typeStr;}
  virtual void	freeData()
  { if (!_m_parent && _m_data) delete _m_data; _m_data = 0; }
  virtual void	newData() { if (!_m_parent && !_m_data) _m_data = new Pnt2d; }
  virtual void	saveData(std::ostream &os) { os << *_m_data; }
  virtual void	loadData(std::istream &is) { is >> *_m_data; }
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const Pnt2d*>(v); }
  virtual const Pnt2d	&readAsPnt2d();
  virtual void		writeAsPnt2d(const Pnt2d &value);
  virtual unsigned	size() const { return 2; }
  virtual void		**rawdata() { return (void**)(&_m_data); }
  virtual void		*subrawdata(unsigned i) { return &(*_m_data)(i); } 
};

// ************************************************************************* //
/// DESCRIPTION: Int data.
// ************************************************************************* //
class PlugMat2i: public TPlug
{
protected:
  Mat2i	*_m_data;
public:
  static String	typeStr;
public:
  PlugMat2i(DataBlock &db): TPlug(db), _m_data(0)
  { newData(); }
  PlugMat2i(DataBlock &db, const String &shortName, const bool isVisible,
	    const bool isSaved, TPlug *parent = 0, Mat2i *data = 0):
    TPlug(db, shortName, isVisible, isSaved, parent), _m_data(data)
  { newData(); }
  virtual ~PlugMat2i();
  static Plug	*creator(DataBlock &db) { return new PlugMat2i(db); }
  virtual int	getTypeNbr() const { return ID_MAT2I; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual const String	&getChildTypeStr() const { return PlugDouble::typeStr;}
  virtual void	freeData()
  { if (!_m_parent && _m_data) delete _m_data; _m_data = 0; }
  virtual void	newData() { if (!_m_parent && !_m_data) _m_data = new Mat2i; }
  virtual void	saveData(std::ostream &os) { os << *_m_data; }
  virtual void	loadData(std::istream &is) { is >> *_m_data; }
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const Mat2i*>(v); }
  virtual const Mat2i	&readAsMat2i();
  virtual void		writeAsMat2i(const Mat2i &value);
  virtual unsigned	size() const { return 9; }
  virtual void		**rawdata() { return (void**)(&_m_data); }
  virtual void		*subrawdata(unsigned i) { return &(*_m_data)(i); } 
};

// ************************************************************************* //
/// DESCRIPTION: Int data.
// ************************************************************************* //
class PlugPnt3d: public TPlug
{
protected:
  double	_m_minValue, _m_maxValue;
  Pnt3d		*_m_data;
public:
  static String	typeStr;
public:
  PlugPnt3d(DataBlock &db): TPlug(db), _m_minValue(MINDOUBLE),
			    _m_maxValue(MAXDOUBLE), _m_data(0) { newData(); }
  PlugPnt3d(DataBlock &db, const String &shortName, const bool isVisible,
	    const bool isSaved, TPlug *parent = 0, Pnt3d *data = 0):
    TPlug(db, shortName, isVisible, isSaved, parent),
    _m_minValue(MINDOUBLE), _m_maxValue(MAXDOUBLE), _m_data(data)
  { newData(); }
  virtual ~PlugPnt3d();
  static Plug	*creator(DataBlock &db) { return new PlugPnt3d(db); }
  virtual int	getTypeNbr() const { return ID_PNT3D; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual const String	&getChildTypeStr() const { return PlugDouble::typeStr;}
  virtual void	freeData()
  { if (!_m_parent && _m_data) delete _m_data; _m_data = 0; }
  virtual void	newData() { if (!_m_parent && !_m_data) _m_data = new Pnt3d; }
  virtual void	saveData(std::ostream &os)
  { os << *_m_data << ' ' << _m_minValue << ' ' << _m_maxValue; }
  virtual void	loadData(std::istream &is)
  { is >> *_m_data >> _m_minValue >> _m_maxValue; }
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const Pnt3d*>(v); }
  virtual const Pnt3d	&readAsPnt3d();
  virtual void		writeAsPnt3d(const Pnt3d &value);
  virtual unsigned	size() const { return 3; }
  virtual void		**rawdata() { return (void**)(&_m_data); }
  virtual void		*subrawdata(unsigned i) { return &(*_m_data)(i); } 
  void		setMinValue(const double minValue) { _m_minValue = minValue; }
  void		setMaxValue(const double maxValue) { _m_maxValue = maxValue; }
  double	minValue() const { return _m_minValue; }
  double	maxValue() const { return _m_maxValue; }
};

// ************************************************************************* //
/// DESCRIPTION: Vec3d data, in [minValue,maxValue]^3
// ************************************************************************* //
class PlugVec3d: public TPlug
{
  friend class PlugVec3dArray;
protected:
  double	_m_minValue, _m_maxValue;
  Vec3d		*_m_data;
public:
  static String	typeStr;
public:
  PlugVec3d(DataBlock &db):
    TPlug(db), _m_minValue(MINDOUBLE), _m_maxValue(MAXDOUBLE), _m_data(0)
  { newData(); }
  PlugVec3d(DataBlock &db, const String &shortName, const bool isVisible,
	    const bool isSaved, TPlug *parent = 0, Vec3d *data = 0):
    TPlug(db, shortName, isVisible, isSaved, parent),
    _m_minValue(MINDOUBLE), _m_maxValue(MAXDOUBLE), _m_data(data)
  { newData(); }
  virtual ~PlugVec3d();
  static Plug	*creator(DataBlock &db) { return new PlugVec3d(db); }
  virtual int	getTypeNbr() const { return ID_VEC3D; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual const String	&getChildTypeStr() const { return PlugDouble::typeStr;}
  virtual void	freeData()
  { if (!_m_parent && _m_data) delete _m_data; _m_data = 0; }
  virtual void	newData() { if (!_m_parent && !_m_data) _m_data = new Vec3d; }
  virtual void	saveData(std::ostream &os)
  { os << *_m_data << ' ' << _m_minValue << ' ' << _m_maxValue; }
  virtual void	loadData(std::istream &is)
  { is >> *_m_data >> _m_minValue >> _m_maxValue; }
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const Vec3d*>(v); }
  virtual const Vec3d	&readAsVec3d();
  virtual void		writeAsVec3d(const Vec3d &value);
  virtual unsigned	size() const { return 3; }
  virtual void		**rawdata() { return (void**)(&_m_data); }
  virtual void		*subrawdata(unsigned i) { return &(*_m_data)(i); } 
  void		setMinValue(const double minValue) { _m_minValue = minValue; }
  void		setMaxValue(const double maxValue) { _m_maxValue = maxValue; }
  double	minValue() const { return _m_minValue; }
  double	maxValue() const { return _m_maxValue; }
};

// ************************************************************************* //
/// DESCRIPTION: Rgb3d data, in [minValue,maxValue]^3
// ************************************************************************* //
class PlugRgb3d: public TPlug
{
protected:
  double	_m_minValue, _m_maxValue;
  Rgb3d		*_m_data;
public:
  static String	typeStr;
public:
  PlugRgb3d(DataBlock &db):
    TPlug(db), _m_minValue(MINDOUBLE), _m_maxValue(MAXDOUBLE), _m_data(0)
  { newData(); }
  PlugRgb3d(DataBlock &db, const String &shortName, const bool isVisible,
	    const bool isSaved, TPlug *parent = 0, Rgb3d *data = 0):
    TPlug(db, shortName, isVisible, isSaved, parent),
    _m_minValue(MINDOUBLE), _m_maxValue(MAXDOUBLE), _m_data(data)
  { newData(); }
  virtual ~PlugRgb3d();
  static Plug	*creator(DataBlock &db) { return new PlugRgb3d(db); }
  virtual int	getTypeNbr() const { return ID_RGB3D; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual const String	&getChildTypeStr() const { return PlugDouble::typeStr;}
  virtual void	freeData()
  { if (!_m_parent && _m_data) delete _m_data; _m_data = 0; }
  virtual void	newData() { if (!_m_parent && !_m_data) _m_data = new Rgb3d; }
  virtual void	saveData(std::ostream &os)
  { os << *_m_data << ' ' << _m_minValue << ' ' << _m_maxValue; }
  virtual void	loadData(std::istream &is)
  { is >> *_m_data >> _m_minValue >> _m_maxValue; }
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const Rgb3d*>(v); }
  virtual const Rgb3d	&readAsRgb3d();
  virtual void		writeAsRgb3d(const Rgb3d &value);
  virtual unsigned	size() const { return 3; }
  virtual void		**rawdata() { return (void**)(&_m_data); }
  virtual void		*subrawdata(unsigned i) { return &(*_m_data)(i); } 
  void		setMinValue(const double minValue) { _m_minValue = minValue; }
  void		setMaxValue(const double maxValue) { _m_maxValue = maxValue; }
  double	minValue() const { return _m_minValue; }
  double	maxValue() const { return _m_maxValue; }
};

// ************************************************************************* //
/// DESCRIPTION: Int data.
// ************************************************************************* //
class PlugMat3d: public TPlug
{
protected:
  Mat3d	*_m_data;
public:
  static String	typeStr;
public:
  PlugMat3d(DataBlock &db): TPlug(db), _m_data(0)
  { newData(); }
  PlugMat3d(DataBlock &db, const String &shortName, const bool isVisible,
	    const bool isSaved, TPlug *parent = 0, Mat3d *data = 0):
    TPlug(db, shortName, isVisible, isSaved, parent), _m_data(data)
  { newData(); }
  virtual ~PlugMat3d();
  static Plug	*creator(DataBlock &db) { return new PlugMat3d(db); }
  virtual int	getTypeNbr() const { return ID_MAT3D; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual const String	&getChildTypeStr() const { return PlugDouble::typeStr;}
  virtual void	freeData()
  { if (!_m_parent && _m_data) delete _m_data; _m_data = 0; }
  virtual void	newData() { if (!_m_parent && !_m_data) _m_data = new Mat3d; }
  virtual void	saveData(std::ostream &os) { os << *_m_data; }
  virtual void	loadData(std::istream &is) { is >> *_m_data; }
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const Mat3d*>(v); }
  virtual const Mat3d	&readAsMat3d();
  virtual void		writeAsMat3d(const Mat3d &value);
  virtual unsigned	size() const { return 9; }
  virtual void		**rawdata() { return (void**)(&_m_data); }
  virtual void		*subrawdata(unsigned i) { return &(*_m_data)(i); } 
};

// ************************************************************************* //
/// DESCRIPTION: Int data.
// ************************************************************************* //
class PlugTwistLog: public TPlug
{
protected:
  TwistLog	*_m_data;
public:
  static String	typeStr;
public:
  PlugTwistLog(DataBlock &db): TPlug(db), _m_data(0)
  { newData(); }
  PlugTwistLog(DataBlock &db, const String &shortName, const bool isVisible,
	       const bool isSaved, TPlug *parent = 0, TwistLog *data = 0):
    TPlug(db, shortName, isVisible, isSaved, parent), _m_data(data)
  { newData(); }
  virtual ~PlugTwistLog();
  static Plug	*creator(DataBlock &db) { return new PlugTwistLog(db); }
  virtual int	getTypeNbr() const { return ID_TWISTLOG; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual const String	&getChildTypeStr() const { return PlugDouble::typeStr;}
  virtual void	freeData()
  { if (!_m_parent && _m_data) delete _m_data; _m_data = 0; }
  virtual void	newData()
  { if (!_m_parent && !_m_data) _m_data = new TwistLog; }
  virtual void	saveData(std::ostream &os) { os << *_m_data; }
  virtual void	loadData(std::istream &is) { is >> *_m_data; }
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const TwistLog*>(v); }
  virtual const TwistLog	&readAsTwistLog();
  virtual void		writeAsTwistLog(const TwistLog &value);
  virtual unsigned	size() const { return 6; }
  virtual void		**rawdata() { return (void**)(&_m_data); }
  virtual void		*subrawdata(unsigned i) { return &(*_m_data)(i); } 
};

// ************************************************************************* //
/// DESCRIPTION: Int data.
// ************************************************************************* //
class PlugMat4d: public TPlug
{
protected:
  Mat4d	*_m_data;
public:
  static String	typeStr;
public:
  PlugMat4d(DataBlock &db): TPlug(db), _m_data(0)
  { newData(); }
  PlugMat4d(DataBlock &db, const String &shortName, const bool isVisible,
	    const bool isSaved, TPlug *parent = 0, Mat4d *data = 0):
    TPlug(db, shortName, isVisible, isSaved, parent), _m_data(data)
  { newData(); }
  virtual ~PlugMat4d();
  static Plug	*creator(DataBlock &db) { return new PlugMat4d(db); }
  virtual int	getTypeNbr() const { return ID_MAT4D; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual const String	&getChildTypeStr() const { return PlugDouble::typeStr;}
  virtual void	freeData()
  { if (!_m_parent && _m_data) delete _m_data; _m_data = 0; }
  virtual void	newData() { if (!_m_parent && !_m_data) _m_data = new Mat4d; }
  virtual void	saveData(std::ostream &os) { os << *_m_data; }
  virtual void	loadData(std::istream &is) { is >> *_m_data; }
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const Mat4d*>(v); }
  virtual const Mat4d	&readAsMat4d();
  virtual void		writeAsMat4d(const Mat4d &value);
  virtual unsigned	size() const { return 16; }
  virtual void		**rawdata() { return (void**)(&_m_data); }
  virtual void		*subrawdata(unsigned i) { return &(*_m_data)(i); } 
};

// ************************************************************************* //
/// DESCRIPTION: Int data.
// ************************************************************************* //
class PlugIntArray: public TPlug
{
  friend class TPlug;
  friend class DataBlock;
protected:
  std::vector<int>	*_m_data;
  unsigned	_m_childrensize;
public:
  static String	typeStr;
public:
  PlugIntArray(DataBlock &db): TPlug(db), _m_data(0)
  { newData(); _m_childrensize = 0; }
  PlugIntArray(DataBlock &db, const String &shortName, const bool isVisible,
	       const bool isSaved, TPlug *parent = 0,
	       std::vector<int> *data = 0):
    TPlug(db, shortName, isVisible, isSaved, parent), _m_data(data)
  { newData(); _m_childrensize = 0; }
  virtual ~PlugIntArray();
  static Plug	*creator(DataBlock &db) { return new PlugIntArray(db); }
  virtual int	getTypeNbr() const { return ID_INTARRAY; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual const String	&getChildTypeStr() const { return PlugInt::typeStr;}
  virtual void	freeData();
  virtual void	newData();
  virtual void	saveData(std::ostream &os);
  virtual void	loadData(std::istream &is);
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const std::vector<int>*>(v); }
  virtual const std::vector<int>	&readAsIntArray();
  virtual void		writeAsIntArray(const std::vector<int> &value);
  virtual bool		isArray() const { return true; }
  virtual unsigned	size() const { return _m_childrensize; }
  virtual void		resize(const unsigned n, const void *defaultvalue = 0);
  virtual void		**rawdata() { return (void**)(&_m_data); }
  virtual void		*subrawdata(unsigned i) { return &(*_m_data)[i]; } 
};

// ************************************************************************* //
/// DESCRIPTION: Int data.
// ************************************************************************* //
class PlugDoubleArray: public TPlug
{
  friend class TPlug;
  friend class DataBlock;
protected:
  std::vector<double>	*_m_data;
  unsigned	_m_childrensize;
public:
  static String	typeStr;
public:
  PlugDoubleArray(DataBlock &db): TPlug(db), _m_data(0)
  { newData(); _m_childrensize = 0; }
  PlugDoubleArray(DataBlock &db, const String &shortName, const bool isVisible,
		  const bool isSaved, TPlug *parent = 0,
		  std::vector<double> *data = 0):
    TPlug(db, shortName, isVisible, isSaved, parent), _m_data(data)
  { newData(); _m_childrensize = 0; }
  virtual ~PlugDoubleArray();
  static Plug	*creator(DataBlock &db) { return new PlugDoubleArray(db); }
  virtual int	getTypeNbr() const { return ID_DOUBLEARRAY; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual const String	&getChildTypeStr() const { return PlugDouble::typeStr;}
  virtual void	freeData();
  virtual void	newData();
  virtual void	saveData(std::ostream &os);
  virtual void	loadData(std::istream &is);
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const std::vector<double>*>(v); }
  virtual const std::vector<double>	&readAsDoubleArray();
  virtual void		writeAsDoubleArray(const std::vector<double> &value);
  virtual bool		isArray() const { return true; }
  virtual unsigned	size() const { return _m_childrensize; }
  virtual void		resize(const unsigned n, const void *defaultvalue = 0);
  virtual void		**rawdata() { return (void**)(&_m_data); }
  virtual void		*subrawdata(unsigned i) { return &(*_m_data)[i]; } 
};

// ************************************************************************* //
/// DESCRIPTION: Int data.
// ************************************************************************* //
class PlugPnt2dArray: public TPlug
{
  friend class TPlug;
  friend class DataBlock;
protected:
  std::vector<Pnt2d>	*_m_data;
  unsigned	_m_childrensize;
public:
  static String	typeStr;
public:
  PlugPnt2dArray(DataBlock &db): TPlug(db), _m_data(0)
  { newData(); _m_childrensize = 0; }
  PlugPnt2dArray(DataBlock &db, const String &shortName, const bool isVisible,
		 const bool isSaved, TPlug *parent = 0,
		 std::vector<Pnt2d> *data = 0):
    TPlug(db, shortName, isVisible, isSaved, parent), _m_data(data)
  { newData(); _m_childrensize = 0; }
  virtual ~PlugPnt2dArray();
  static Plug	*creator(DataBlock &db) { return new PlugPnt2dArray(db); }
  virtual int	getTypeNbr() const { return ID_PNT2DARRAY; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual const String	&getChildTypeStr() const { return PlugPnt2d::typeStr;}
  virtual void	freeData();
  virtual void	newData();
  virtual void	saveData(std::ostream &os);
  virtual void	loadData(std::istream &is);
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const std::vector<Pnt2d>*>(v); }
  virtual const std::vector<Pnt2d>	&readAsPnt2dArray();
  virtual const Pnt2d			&readAsPnt2dArray(const int i);
  virtual void		writeAsPnt2dArray(const std::vector<Pnt2d> &value);
  virtual void		writeAsPnt2dArray(const int i, const Pnt2d &value);
  virtual bool		isArray() const { return true; }
  virtual unsigned	size() const { return _m_childrensize; }
  virtual void		resize(const unsigned n, const void *defaultvalue = 0);
  virtual void		**rawdata() { return (void**)(&_m_data); }
  virtual void		*subrawdata(unsigned i) { return &(*_m_data)[i]; } 
};

// ************************************************************************* //
/// DESCRIPTION: Int data.
// ************************************************************************* //
class PlugPnt3dArray: public TPlug
{
  friend class TPlug;
  friend class DataBlock;
protected:
  std::vector<Pnt3d>	*_m_data;
  unsigned	_m_childrensize;
public:
  static String	typeStr;
public:
  PlugPnt3dArray(DataBlock &db): TPlug(db), _m_data(0)
  { newData(); _m_childrensize = 0; }
  PlugPnt3dArray(DataBlock &db, const String &shortName, const bool isVisible,
		 const bool isSaved, TPlug *parent = 0,
		 std::vector<Pnt3d> *data = 0):
    TPlug(db, shortName, isVisible, isSaved, parent), _m_data(data)
  { newData(); _m_childrensize = 0; }
  virtual ~PlugPnt3dArray();
  static Plug	*creator(DataBlock &db) { return new PlugPnt3dArray(db); }
  virtual int	getTypeNbr() const { return ID_PNT3DARRAY; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual const String	&getChildTypeStr() const { return PlugPnt3d::typeStr;}
  virtual void	freeData();
  virtual void	newData();
  virtual void	saveData(std::ostream &os);
  virtual void	loadData(std::istream &is);
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const std::vector<Pnt3d>*>(v); }
  virtual const std::vector<Pnt3d>	&readAsPnt3dArray();
  virtual const Pnt3d			&readAsPnt3dArray(const int i);
  virtual void		writeAsPnt3dArray(const std::vector<Pnt3d> &value);
  virtual void		writeAsPnt3dArray(const int i, const Pnt3d &value);
  virtual bool		isArray() const { return true; }
  virtual unsigned	size() const { return _m_childrensize; }
  virtual void		resize(const unsigned n, const void *defaultvalue = 0);
  virtual void		**rawdata() { return (void**)(&_m_data); }
  virtual void		*subrawdata(unsigned i) { return &(*_m_data)[i]; } 
};

// ************************************************************************* //
/// DESCRIPTION: Vec3dArray data.
// ************************************************************************* //
class PlugVec3dArray: public TPlug
{
  friend class TPlug;
  friend class DataBlock;
protected:
  std::vector<Vec3d>	*_m_data;
  unsigned	_m_childrensize;
public:
  static String	typeStr;
public:
  PlugVec3dArray(DataBlock &db): TPlug(db), _m_data(0)
  { newData(); _m_childrensize = 0; }
  PlugVec3dArray(DataBlock &db, const String &shortName, const bool isVisible,
		 const bool isSaved, TPlug *parent = 0,
		 std::vector<Vec3d> *data = 0):
    TPlug(db, shortName, isVisible, isSaved, parent), _m_data(data)
  { newData(); _m_childrensize = 0; }
  virtual ~PlugVec3dArray();
  static Plug	*creator(DataBlock &db) { return new PlugVec3dArray(db); }
  virtual int	getTypeNbr() const { return ID_VEC3DARRAY; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual const String	&getChildTypeStr() const { return PlugVec3d::typeStr;}
  virtual void	freeData();
  virtual void	newData();
  virtual void	saveData(std::ostream &os);
  virtual void	loadData(std::istream &is);
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const std::vector<Vec3d>*>(v); }
  virtual const std::vector<Vec3d>	&readAsVec3dArray();
  virtual const Vec3d			&readAsVec3dArray(const int i);
  virtual void		writeAsVec3dArray(const std::vector<Vec3d> &value);
  virtual void		writeAsVec3dArray(const int i, const Vec3d &value);
  virtual bool		isArray() const { return true; }
  virtual unsigned	size() const { return _m_childrensize; }
  virtual void		resize(const unsigned n, const void *defaultvalue = 0);
  virtual void		**rawdata() { return (void**)(&_m_data); }
  virtual void		*subrawdata(unsigned i) { return &(*_m_data)[i]; } 
};

// ************************************************************************* //
/// DESCRIPTION: Twistlogarray data.
// ************************************************************************* //
class PlugTwistLogArray: public TPlug
{
  friend class TPlug;
  friend class DataBlock;
protected:
  std::vector<TwistLog>	*_m_data;
  unsigned	_m_childrensize;
public:
  static String	typeStr;
public:
  PlugTwistLogArray(DataBlock &db): TPlug(db), _m_data(0)
  { newData(); _m_childrensize = 0; }
  PlugTwistLogArray(DataBlock &db, const String &shortName,
		    const bool isVisible, const bool isSaved,
		    TPlug *parent = 0, std::vector<TwistLog> *data = 0):
    TPlug(db, shortName, isVisible, isSaved, parent), _m_data(data)
  { newData(); _m_childrensize = 0; }
  virtual ~PlugTwistLogArray();
  static Plug	*creator(DataBlock &db) { return new PlugTwistLogArray(db); }
  virtual int	getTypeNbr() const { return ID_TWISTLOGARRAY; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual const String	&getChildTypeStr() const { return PlugVec3d::typeStr;}
  virtual void	freeData();
  virtual void	newData();
  virtual void	saveData(std::ostream &os);
  virtual void	loadData(std::istream &is);
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const std::vector<TwistLog>*>(v); }
  virtual const std::vector<TwistLog>	&readAsTwistLogArray();
  virtual const TwistLog		&readAsTwistLogArray(const int i);
  virtual void	writeAsTwistLogArray(const std::vector<TwistLog> &value);
  virtual void	writeAsTwistLogArray(const int i, const TwistLog &value);
  virtual bool		isArray() const { return true; }
  virtual unsigned	size() const { return _m_childrensize; }
  virtual void		resize(const unsigned n, const void *defaultvalue = 0);
  virtual void		**rawdata() { return (void**)(&_m_data); }
  virtual void		*subrawdata(unsigned i) { return &(*_m_data)[i]; } 
};

// ************************************************************************* //
/// DESCRIPTION: Int data.
// ************************************************************************* //
class PlugPnt2dList: public TPlug
{
  friend class TPlug;
  friend class DataBlock;
protected:
  std::list<Pnt2d>	*_m_data;
  unsigned	_m_childrensize;
public:
  static String	typeStr;
public:
  PlugPnt2dList(DataBlock &db): TPlug(db), _m_data(0)
  { newData(); _m_childrensize = 0; }
  PlugPnt2dList(DataBlock &db, const String &shortName, const bool isVisible,
		const bool isSaved, TPlug *parent = 0,
		std::list<Pnt2d> *data = 0):
    TPlug(db, shortName, isVisible, isSaved, parent), _m_data(data)
  { newData(); _m_childrensize = 0; }
  virtual ~PlugPnt2dList();
  static Plug	*creator(DataBlock &db) { return new PlugPnt2dList(db); }
  virtual int	getTypeNbr() const { return ID_PNT2DLIST; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual const String	&getChildTypeStr() const { return PlugPnt2d::typeStr;}
  virtual void	freeData();
  virtual void	newData();
  virtual void	saveData(std::ostream &os);
  virtual void	loadData(std::istream &is);
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const std::list<Pnt2d>*>(v); }
  virtual const std::list<Pnt2d>	&readAsPnt2dList();
  virtual void		writeAsPnt2dList(const std::list<Pnt2d> &value);
  virtual unsigned	size() const { return _m_childrensize; }
  virtual void		resize(const unsigned n, const void *defaultvalue = 0);
  virtual void		**rawdata() { return (void**)(&_m_data); }
  virtual void		*subrawdata(unsigned i);
};

// ************************************************************************* //
#endif//__PLUG_HH__
