// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:30 $
/// $Revision: 1.1 $
// ************************************************************************* //
#ifndef	__PLUGMESH_HH__
#define	__PLUGMESH_HH__

// ************************************************************************* //
#include <qimage.h>
#include <GL/gl.h>
#include "plug.h"
#ifdef _WIN32
#include "glext.h"
#endif//_WIN32

// ************************************************************************* //
class Mesh;

// ************************************************************************* //
double	dist2Triangle(const Pnt3d &p0, const Pnt3d &p1, const Pnt3d &p2,
		      const Pnt3d &p);

// ************************************************************************* //
std::ostream	&operator<<(std::ostream &os, const Mesh &m);
std::istream	&operator>>(std::istream &is, Mesh &m);

// ************************************************************************* //
/// DESCRIPTION:
///  vertices	- array of vertices
///  normals	- array of normals per vertex.
///  triangles	- triangle, defined by index in array "vertices" and "normals"
///  edges	- edges, defined by index in array "vertices" and "normals"
///  tedges	- triangles of the edge (at most two).
// ************************************************************************* //
class Mesh
{
  typedef std::pair<unsigned, unsigned>	Edge;
  typedef std::pair<unsigned, unsigned>	TEdge;
  std::vector<Pnt3d>	_m_vertices;//saved
  std::vector<Vec3d>	_m_normals;//saved
  std::vector<unsigned>	_m_triangles;//saved
  std::vector<Edge>	_m_edges;//stored index of edge's vertices
  std::vector<TEdge>	_m_tedges;//stored index of edge's triangle
  BBox3d		_m_bbox;
  GLuint		_m_texName; //
  std::vector<Pnt2d>	_m_uvs; // per triangle!
  std::vector<unsigned>	_m_texcoords;//saved
public:
  Mesh() {}
  const std::vector<Pnt3d>	&vertices() const { return _m_vertices; }
  const std::vector<Edge>	&edges() const { return _m_edges; }
  friend std::ostream	&operator<<(std::ostream &os, const Mesh &m);
  friend std::istream	&operator>>(std::istream &is, Mesh &m);
  void	readAsOBS(std::istream &is);
  void	readAsWRL(std::istream &is);
  void	glRender(const bool isFlatShaded = false) const;
  void	computeSamples(std::vector<Pnt3d> &samples,
		       const unsigned resolution) const;
  void	computeNormals(std::vector<Vec3d> &normals,
		       const unsigned resolution) const;
  void	computeCenters(std::vector<Pnt3d> &centers, const double eps,
		       const unsigned resolution) const;
  void	computeTangents(std::vector<Vec3d> &tangents) const;
  void	computeCNormals(std::vector<Vec3d> &cnormals,
			const unsigned resolution) const;
  bool		contains(Pnt3d p) const;
  double	dist2(const Pnt3d &p) const;
  double	field(const Pnt3d &p, const double safeZone,
		      const double smoothZone) const;
  
  // *********************************************************************** //
  void	setTexture(const QImage &qimage);
  void	cwRotateTexture();
  void	ccwRotateTexture();
  void	mirroru();
  void	mirrorv();
  void	createEdges();
  void	createNormals();
};

// ************************************************************************* //
inline std::ostream	&operator<<(std::ostream	&os,
				    const Mesh		&m)
{
  os << m._m_vertices.size() << ' ';
  for (std::vector<Pnt3d>::const_iterator vit = m._m_vertices.begin();
       vit != m._m_vertices.end(); ++vit)
    os << *vit << ' ';
  os << m._m_normals.size() << ' ';
  for (std::vector<Vec3d>::const_iterator nit = m._m_normals.begin();
       nit != m._m_normals.end(); ++nit)
    os << *nit << ' ';
  os << m._m_triangles.size() << ' ';
  for (std::vector<unsigned>::const_iterator iit = m._m_triangles.begin();
       iit != m._m_triangles.end(); ++iit)
    os << *iit << ' ';
  return os;
}

// ************************************************************************* //
/// DESCRIPTION: Loads the image and assigns texture number to mesh.
// ************************************************************************* //
inline void	Mesh::setTexture(const QImage &qimage)
{
  if (_m_texName)
    glDeleteTextures(1, &_m_texName);

  if (qimage.isNull())
    std::cerr << "Texture image cannot be loaded." << std::endl;
  else if (qimage.depth() != 32)
    std::cerr << "Cannot convert image to 32 bits encoding." << std::endl;
  else
    {
      glGenTextures(1, &_m_texName);
      glBindTexture(GL_TEXTURE_2D, _m_texName);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 
		   qimage.width(), qimage.height(), 
		   0, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, qimage.bits());
	}
}

// ************************************************************************* //
inline void	Mesh::cwRotateTexture()
{
  for (std::vector<Pnt2d>::iterator nit = _m_uvs.begin();
       nit != _m_uvs.end(); ++nit)
    *nit = Pnt2d(-nit->y, nit->x);
}

// ************************************************************************* //
inline void	Mesh::ccwRotateTexture()
{
  for (std::vector<Pnt2d>::iterator nit = _m_uvs.begin();
       nit != _m_uvs.end(); ++nit)
    *nit = Pnt2d(nit->y, -nit->x);
}

// ************************************************************************* //
inline void	Mesh::mirroru()
{
  for (std::vector<Pnt2d>::iterator nit = _m_uvs.begin();
       nit != _m_uvs.end(); ++nit)
    nit->x = 1.0 - nit->x;
}

// ************************************************************************* //
inline void	Mesh::mirrorv()
{
  for (std::vector<Pnt2d>::iterator nit = _m_uvs.begin();
       nit != _m_uvs.end(); ++nit)
    nit->y = 1.0 - nit->y;
}

// ************************************************************************* //
inline void	Mesh::computeTangents(std::vector<Vec3d> &tangents) const
{
  for (unsigned	i(0), n(_m_triangles.size()); i < n; i += 3)
    {
      const Pnt3d	&p0(_m_vertices[_m_triangles[i + 0]]);
      const Pnt3d	&p1(_m_vertices[_m_triangles[i + 1]]);
      const Pnt3d	&p2(_m_vertices[_m_triangles[i + 2]]);
      const Vec3d	z(((p2 - p1) ^ (p0 - p1)).unit());
      Vec3d		x, y;

      if (fabs(z.x) < fabs(z.y) && fabs(z.x) < fabs(z.z))
	y = Vec3d(1.0, 0.0, 0.0) - z * z.x;
      else if (fabs(z.y) < fabs(z.z))
	y = Vec3d(0.0, 1.0, 0.0) - z * z.y;
      else
	y = Vec3d(0.0, 0.0, 1.0) - z * z.z;
      y.eqUnit();
      x = y ^ z;
      for (unsigned j(0), m(4); j < m; ++j)
	for (unsigned k(0); k < m; ++k)
	  if (k + j < m)
	    {
	      const Pnt2d	p(j / double(m), k / double(m));
	      const Pnt2d	v1(p.x + 0.66 / m, p.y + 0.66 / m);

	      tangents.push_back(x), tangents.push_back(y);
	      if (v1.x + v1.y < 1.0)
		tangents.push_back(x), tangents.push_back(y);
	    }
    }
}

// ************************************************************************* //
inline void	Mesh::computeCenters(std::vector<Pnt3d> &samples,
				     const double	epsilon,
				     const unsigned	resolution) const
{
  const unsigned	m(resolution);

  for (unsigned	i(0), n(_m_triangles.size()); i < n; i += 3)
    {
      const Pnt3d	&p0(_m_vertices[_m_triangles[i + 0]]);
      const Pnt3d	&p1(_m_vertices[_m_triangles[i + 1]]);
      const Pnt3d	&p2(_m_vertices[_m_triangles[i + 2]]);
      const Vec3d	&n0(_m_normals[_m_triangles[i + 0]]);
      const Vec3d	&n1(_m_normals[_m_triangles[i + 1]]);
      const Vec3d	&n2(_m_normals[_m_triangles[i + 2]]);

      for (unsigned j(1); j < m; ++j)
	for (unsigned k(1); k < m; ++k)
	  if (k + j < m)
	    {
	      const Pnt2d	p(j / double(m), k / double(m));
	      const Vec3d	n((n0 * (1.0 - p.x - p.y) +
				   n1 * p.x + n2 * p.y).unit());

	      samples.push_back(p0 * (1.0 - p.x - p.y) + p1 * p.x + p2 * p.y -
				n * epsilon);
	    }
    }
  for (unsigned	i(0), n(_m_edges.size()); i < n; ++i)
    {
      const Pnt3d	&p0(_m_vertices[_m_edges[i].first]);
      const Pnt3d	&p1(_m_vertices[_m_edges[i].second]);
      const Vec3d	&n0(_m_normals[_m_edges[i].first]);
      const Vec3d	&n1(_m_normals[_m_edges[i].second]);

      for (unsigned j(1); j < m; ++j)
	{
	  const double	p(j / double(m));
	  const Vec3d	n(n0 * (1.0 - p) + n1 * p);

	  samples.push_back(p0 * (1.0 - p) + p1 * p - n * epsilon);
	}
    }
  for (unsigned	i(0), n(_m_vertices.size()); i < n; ++i)
    samples.push_back(_m_vertices[i] - _m_normals[i] * epsilon);
}

// ************************************************************************* //
inline void	Mesh::computeSamples(std::vector<Pnt3d> &samples,
				     const unsigned	resolution) const
{
  const unsigned	m(resolution);

  for (unsigned	i(0), n(_m_triangles.size()); i < n; i += 3)
    {
      const Pnt3d	&p0(_m_vertices[_m_triangles[i + 0]]);
      const Pnt3d	&p1(_m_vertices[_m_triangles[i + 1]]);
      const Pnt3d	&p2(_m_vertices[_m_triangles[i + 2]]);

      for (unsigned j(1); j < m; ++j)
	for (unsigned k(1); k < m; ++k)
	  if (k + j < m)
	    {
	      const Pnt2d	p(j / double(m), k / double(m));

	      samples.push_back(p0 * (1.0 - p.x - p.y) + p1 * p.x + p2 * p.y);
	    }
    }
  for (unsigned	i(0), n(_m_edges.size()); i < n; ++i)
    {
      const Pnt3d	&p0(_m_vertices[_m_edges[i].first]);
      const Pnt3d	&p1(_m_vertices[_m_edges[i].second]);

      for (unsigned j(1); j < m; ++j)
	{
	  const double	p(j / double(m));
	  
	  samples.push_back(p0 * (1.0 - p) + p1 * p);
	}
    }
  for (unsigned	i(0), n(_m_vertices.size()); i < n; ++i)
    samples.push_back(_m_vertices[i]);
}

// ************************************************************************* //
inline void	Mesh::computeCNormals(std::vector<Vec3d>	&normals,
				      const unsigned	resolution) const
{
  const unsigned	m(resolution);

  for (unsigned	i(0), n(_m_triangles.size()); i < n; i += 3)
    {
      const Pnt3d	&p0(_m_vertices[_m_triangles[i + 0]]);
      const Pnt3d	&p1(_m_vertices[_m_triangles[i + 1]]);
      const Pnt3d	&p2(_m_vertices[_m_triangles[i + 2]]);
      const Vec3d	normal(((p2 - p1) ^ (p0 - p1)).unit());

      for (unsigned j(1); j < m; ++j)
	for (unsigned k(1); k < m; ++k)
	  if (k + j < m)
	    normals.push_back(normal);
    }
  for (unsigned	i(0), n(_m_tedges.size()); i < n; ++i)
    {
      const unsigned	t0(_m_tedges[i].first);
      const unsigned	t1(_m_tedges[i].second);
      const Pnt3d	&pi0(_m_vertices[_m_triangles[t0 + 0]]);
      const Pnt3d	&pi1(_m_vertices[_m_triangles[t0 + 1]]);
      const Pnt3d	&pi2(_m_vertices[_m_triangles[t0 + 2]]);
      const Pnt3d	&pj0(_m_vertices[_m_triangles[t1 + 0]]);
      const Pnt3d	&pj1(_m_vertices[_m_triangles[t1 + 1]]);
      const Pnt3d	&pj2(_m_vertices[_m_triangles[t1 + 2]]);
      const Vec3d	ni(((pi2 - pi1) ^ (pi0 - pi1)).unit());
      const Vec3d	nj(((pj2 - pj1) ^ (pj0 - pj1)).unit());
      const Vec3d	normal((ni + nj).unit());

      for (unsigned j(1); j < m; ++j)
	normals.push_back(normal);
    }
  for (unsigned	i(0), n(_m_normals.size()); i < n; ++i)
    normals.push_back(_m_normals[i]);
}

// ************************************************************************* //
inline void	Mesh::computeNormals(std::vector<Vec3d>	&normals,
				     const unsigned	resolution) const
{
  const unsigned	m(resolution);

  for (unsigned	i(0), n(_m_triangles.size()); i < n; i += 3)
    {
      const Pnt3d	&p0(_m_vertices[_m_triangles[i + 0]]);
      const Pnt3d	&p1(_m_vertices[_m_triangles[i + 1]]);
      const Pnt3d	&p2(_m_vertices[_m_triangles[i + 2]]);
      const Vec3d	normal(((p2 - p1) ^ (p0 - p1)).unit());

      for (unsigned j(1); j < m; ++j)
	for (unsigned k(1); k < m; ++k)
	  if (k + j < m)
	    normals.push_back(normal);
    }
  for (unsigned	i(0), n(_m_tedges.size()); i < n; ++i)
    {
      const unsigned	t0(_m_tedges[i].first);
      const unsigned	t1(_m_tedges[i].second);
      const Pnt3d	&pi0(_m_vertices[_m_triangles[t0 + 0]]);
      const Pnt3d	&pi1(_m_vertices[_m_triangles[t0 + 1]]);
      const Pnt3d	&pi2(_m_vertices[_m_triangles[t0 + 2]]);
      const Pnt3d	&pj0(_m_vertices[_m_triangles[t1 + 0]]);
      const Pnt3d	&pj1(_m_vertices[_m_triangles[t1 + 1]]);
      const Pnt3d	&pj2(_m_vertices[_m_triangles[t1 + 2]]);
      const Vec3d	ni(((pi2 - pi1) ^ (pi0 - pi1)).unit());
      const Vec3d	nj(((pj2 - pj1) ^ (pj0 - pj1)).unit());
      const Vec3d	normal((ni + nj).unit());

      for (unsigned j(1); j < m; ++j)
	normals.push_back(normal);
    }
  for (unsigned	i(0), n(_m_normals.size()); i < n; ++i)
    normals.push_back(_m_normals[i]);
}

// ************************************************************************* //
/// DESCRIPTION: Extends a data node. This is stored in the scene graph.
// ************************************************************************* //
class PlugMesh: public Plug
{
protected:
  Mesh	*_m_data;
public:
  static int	typeNbr;
  static String	typeStr;
public:
  PlugMesh(DataBlock &db): Plug(db), _m_data(0) { newData(); }
  virtual ~PlugMesh() { freeData(); }
  static Plug	*creator(DataBlock &db) { return new PlugMesh(db); }
  virtual int	getTypeNbr() const { return typeNbr; }
  virtual const String	&getTypeStr() const { return typeStr; }
  virtual void	freeData() { if (_m_data) delete _m_data; _m_data = 0; }
  virtual void	newData() { if (!_m_data) _m_data = new Mesh; }
  virtual void	saveData(std::ostream &os) { os << *_m_data; }
  virtual void	loadData(std::istream &is) { is >> *_m_data; }
  virtual void	copy(const void *v)
  { *_m_data = *static_cast<const Mesh*>(v); }
  virtual void	**rawdata() { return (void**)(&_m_data); }
};

// ************************************************************************* //
#endif//__PLUGMESH_HH__

