// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:30 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <QFontMetrics>
#include <QPixmap>
#include <QPainter>
#include <QLayout>
#include <QPushButton>
#include <QMenuBar>
#include <QLineEdit> 
#include <QColorDialog>

// ************************************************************************* //
#include "qtAttributeEditor.h"
#include "qtCurveEditor.h"
#include "plug.h"
#include "datablock.h"
#include "dagnode.h"
#include "command.h"
#include "mystring.h"
#include "input.xpm"
#include "output.xpm"

// ************************************************************************* //
/// DESCRIPTION: Creates a floating attribute editor on node.
// ************************************************************************* //
AttributeEditor::AttributeEditor(DagNode	&node,
				 QWidget	*parent): 
  QMainWindow(parent, Qt::Window),
  _m_node(&node)
{
  DataBlock	&db(node.getDataBlock());
  DataBlock::iterator	dit;
  int		size(0);

  db.addValidBit(this);

  for (dit = db.begin(); dit != db.end(); ++dit)
    if ((*dit)->isVisible())
      ++size;

  setWindowTitle("Attribute Editor");

  // Menu ****************************************************************** //
  QMenu	*file = menuBar()->addMenu(tr("&File"));

  file->addAction("&Close", this, SLOT(close()), Qt::CTRL + Qt::Key_Q);

  QMenu	*edit = menuBar()->addMenu(tr("&Edit"));

  edit->addAction("&Parent", this, SLOT(editParent()), Qt::CTRL + Qt::Key_P);
  edit->addAction("&Curves", this, SLOT(editCurves()), Qt::CTRL + Qt::Key_C);

  // Attributes ************************************************************ //
  QWidget	*cw = new AttributeCentral(node, size, this);//centralwidget

  setCentralWidget(cw);

  setFocusPolicy(Qt::ClickFocus);
}

// ************************************************************************* //
AttributeEditor::~AttributeEditor()
{
  _m_node->getDataBlock().remValidBit(this);
}

// ************************************************************************* //
void AttributeEditor::keyPressEvent(QKeyEvent *e)
{
  if (e->key() == Qt::Key_Escape)
    close();
  QMainWindow::keyPressEvent(e);
}

// ************************************************************************* //
void	AttributeEditor::paintEvent(QPaintEvent *e)
{
  QMainWindow::paintEvent(e);
}

// ************************************************************************* //
void	AttributeEditor::editCurves()
{
  CurveEditor *newf = new CurveEditor(*_m_node);
  newf->show();
}

// ************************************************************************* //
void	AttributeEditor::editParent()
{
  if (_m_node->getNodeParent())
    {
      AttributeEditor *newf = new AttributeEditor(*_m_node->getNodeParent());
      newf->show();
    }
}

// ************************************************************************* //
AttributeCentral::AttributeCentral(DagNode	&node,
				   const int	size,
				   QWidget	*parent): QWidget(parent)
{
  DataBlock	&db(node.getDataBlock());
  DataBlock::iterator	dit;
  QGridLayout	*grid = new QGridLayout(this);//, size + 2, 3, 2, 2
  QWidget	*nodeName = new AttributeName(this, node.getInstStr().data());
  int	i;

  grid->setMargin(0);
  grid->setSpacing(0);
  grid->addWidget(nodeName, 0, 0, 1, 4, Qt::AlignCenter | Qt::AlignVCenter);
  for (i = 1, dit = db.begin(); dit != db.end(); ++dit)
    if ((*dit)->isVisible())
      {
	//grid->setRowMinimumHeight(i, 2);
	QWidget *wid1 = new AttributeName(this, (*dit)->longName().data());
	QWidget *wid2(0);

	grid->addWidget(wid1, i, 1, Qt::AlignCenter | Qt::AlignVCenter);

	if ((*dit)->getTypeNbr() == ID_BOOL)
	  wid2 = new AttrBool(**dit, this);
	else if ((*dit)->getTypeNbr() == ID_INT)
	  wid2 = new AttrInt(static_cast<PlugInt&>(**dit), this);
	else if ((*dit)->getTypeNbr() == ID_DOUBLE)
	  wid2 = new AttrDouble(static_cast<PlugDouble&>(**dit), this);
	else if ((*dit)->getTypeNbr() == ID_STRING)
	  wid2 = new AttrString(**dit, this);
	else if ((*dit)->getTypeNbr() == ID_PNT3D)
	  wid2 = new AttrPnt3d(static_cast<PlugPnt3d&>(**dit), this);
	else if ((*dit)->getTypeNbr() == ID_VEC3D)
	  wid2 = new AttrVec3d(static_cast<PlugVec3d&>(**dit), this);
	else if ((*dit)->getTypeNbr() == ID_RGB3D)
	  wid2 = new AttrRgb3d(static_cast<PlugRgb3d&>(**dit), this);
	if (wid2)
	  grid->addWidget(wid2, i, 2, Qt::AlignRight | Qt::AlignVCenter);
	if ((*dit)->getTypeNbr() == ID_DOUBLE ||
	    (*dit)->getTypeNbr() == ID_RGB3D ||
	    ((*dit)->getTypeNbr() == ID_PNT3D && (*dit)->getInputNode()))
	  {
	    QPushButton	*wid0 = new QPushButton(this);
	    QPushButton	*wid3 = new QPushButton(this);
	    QPixmap	inputIcon((const char **)input_xpm);
	    QPixmap	outputIcon((const char **)output_xpm);
	    InputEditorLauncher	*cel;

	    wid0->setIcon(inputIcon);
	    wid3->setIcon(outputIcon);
	    wid0->setFixedSize(inputIcon.width() + 8, inputIcon.height() + 8);
	    wid3->setFixedSize(inputIcon.width() + 8, inputIcon.height() + 8);

	    cel = new InputEditorLauncher(node, **dit, wid2, this);
	    // TODO: pass the wid2, the AttrRgb3d? ... for refresh?
	    QWidget::connect(wid0, SIGNAL(clicked()), cel, SLOT(editInput()));
	    QWidget::connect(wid3, SIGNAL(clicked()), cel, SLOT(editOutput()));
	    grid->addWidget(wid0, i, 0, Qt::AlignCenter | Qt::AlignVCenter);
	    grid->addWidget(wid3, i, 3, Qt::AlignCenter | Qt::AlignVCenter);
	  }
	++i;
      }
  // Buttons *************************************************************** //
  QPushButton *close = new QPushButton("&Close", this);

  QWidget::connect(close, SIGNAL(clicked()), parentWidget(), SLOT(close()));

  grid->addWidget(close, size + 1, 0, 1, 4,
		  Qt::AlignCenter | Qt::AlignVCenter);
  grid->setRowStretch(size + 1, 10);
}

// ************************************************************************* //
InputEditorLauncher::InputEditorLauncher(DagNode	&node,
					 Plug		&data,
					 QWidget	*edited,
					 QWidget	*parent):
  QObject(parent), _m_node(node), _m_data(data), _m_edited(edited)
{
}

// ************************************************************************* //
void	InputEditorLauncher::editInput()
{
  DagNode	*isInputNode = _m_data.getInputNode();

  if (!isInputNode)
    if (_m_data.getTypeNbr() == ID_DOUBLE)
      {
	DagNode	*clock = DagNode::getNodeByPath(":Clock0");
	  
	isInputNode = Command::run(String("new -type Curve -path ") +
				   _m_node.path());
	DagNode::connect(*clock, "fd", *isInputNode, "x");
	DagNode::connect(*isInputNode->getPlug("y"), _m_data);
      }
    else if (_m_data.getTypeNbr() == ID_RGB3D)
      {
	const Rgb3d	c = static_cast<AttrRgb3d*>(_m_edited)->getValue();
	QColor		color = QColorDialog::getColor(QColor(int(c.r * 255),
							      int(c.g * 255),
							      int(c.b * 255)));

        if (color.isValid())
	  {
	    QObject::connect(this, SIGNAL(valueChanged(Rgb3d)),
			     _m_edited, SLOT(setValue(Rgb3d)));
	    emit valueChanged(Rgb3d(color.red() / 255.0,
				    color.green() / 255.0,
				    color.blue() / 255.0));
	  }
      }
  if (isInputNode)
    {
      QWidget *newf;

      if (isInputNode->getTypeStr() == "Curve")
	newf = new CurveEditor(*isInputNode);
      else 
	newf = new AttributeEditor(*isInputNode);
      newf->show();
    }
}

// ************************************************************************* //
void	InputEditorLauncher::editOutput()
{
}

// ************************************************************************* //
AttributeName::AttributeName(QWidget	*parent,
			     const char	*name): 
  QWidget(parent), _m_name(name)
{
  QFontMetrics	fm(font());
  QRect		qs(fm.boundingRect(_m_name.data()));

  setBaseSize(qs.width() + 8, qs.height() + 8);
  setMinimumSize(qs.width() + 8, qs.height() + 8);
}

// ************************************************************************* //
void	AttributeName::paintEvent(QPaintEvent *)
{
  QFontMetrics	fm(font());
  QPainter	p;

  p.begin(this);
  p.setPen(Qt::NoPen);
  p.setBrush(QColor(230, 230, 230));
  //p.drawRect(0, 0, width(), height());//REMOVE
  QRectF	qs(0, 0, width(), height());

  p.drawText(qs, _m_name.data(), QTextOption(Qt::AlignCenter));
  p.end();
}

// ************************************************************************* //
AttrBool::AttrBool(Plug		&data,
		   QWidget	*parent):
  QWidget(parent), _m_data(&data)
{
  QCheckBox	*cb = new QCheckBox(this);
  QLayout	*lyt = layout()? layout(): new QHBoxLayout(this);//, 1, 1, 0
  
  cb->setChecked(_m_data->readAsBool());

  lyt->addWidget(cb);
  lyt->setMargin(0);
  lyt->setSpacing(0);
  connect(cb, SIGNAL(toggled(bool)), this, SLOT(toggle(bool)));
}

// ************************************************************************* //
void	AttrBool::toggle(bool on)
{
  _m_data->writeAsBool(on);
}

// ************************************************************************* //
AttrInt::AttrInt(PlugInt	&data,
		 QWidget	*parent):
  QWidget(parent), _m_data(&data)
{
  QFontMetrics	fm(parent->font());
  QRect		qs(fm.boundingRect("10000.0000"));
  QSpinBox	*sp = new QSpinBox(this);
  QLayout	*lyt = layout()? layout(): new QHBoxLayout(this);//, 1, 1, 0

  sp->setRange(data.minValue(), data.maxValue());
  sp->setBaseSize(qs.width() + 8, qs.height() + 8);
  sp->setMinimumSize(qs.width() + 8, qs.height() + 8);
  sp->setMaximumSize(qs.width() + 8, qs.height() + 8);
  sp->setFrame(false);
  sp->setValue(_m_data->readAsInt());

  lyt->addWidget(sp);
  lyt->setMargin(1);
  lyt->setSpacing(0);
  connect(sp, SIGNAL(valueChanged(int)), this, SLOT(setValue(int)));
}

// ************************************************************************* //
void	AttrInt::setValue(int v)
{
  _m_data->writeAsInt(v);
}

// ************************************************************************* //
AttrDouble::AttrDouble(PlugDouble	&data,
		       QWidget		*parent):
  QWidget(parent), _m_data(&data)
{
  QFontMetrics	fm(parent->font());
  QRect		qs(fm.boundingRect("10000.0000"));
  QLayout	*lyt = layout()? layout(): new QHBoxLayout(this);//, 1, 1, 0
  QDoubleSpinBox	*sp = new QDoubleSpinBox(this);

  sp->setRange(data.minValue(), data.maxValue());
  sp->setDecimals(6);
  sp->setBaseSize(qs.width() + 8, qs.height() + 8);
  sp->setMinimumSize(qs.width() + 8, qs.height() + 8);
  sp->setMaximumSize(qs.width() + 8, qs.height() + 8);
  sp->setFrame(false);
  sp->setValue(_m_data->readAsDouble());

  lyt->addWidget(sp);
  lyt->setMargin(1);
  lyt->setSpacing(0);
  connect(sp, SIGNAL(valueChanged(double)), this, SLOT(setValue(double)));
}

// ************************************************************************* //
void	AttrDouble::setValue(double v)
{
  _m_data->writeAsDouble(v);
}

// ************************************************************************* //
AttrString::AttrString(Plug		&data,
		       QWidget		*parent):
  QWidget(parent), _m_data(&data)
{
  QFontMetrics	fm(parent->font());
  //QSize		qs(fm.size(Qt::TextSingleLine, "10000.0000"));
  QLayout	*lyt = layout()? layout(): new QHBoxLayout(this);//, 1, 1, 0
  QLineEdit	*sp = new QLineEdit(this);
  
  //sp->setBaseSize(qs.width() + 8, qs.height() + 8);
  //sp->setMinimumSize(qs.width() + 8, qs.height() + 8);
  //sp->setMaximumSize(qs.width() + 8, qs.height() + 8);
  sp->setText(_m_data->readAsString().data());
  lyt->addWidget(sp);
  lyt->setMargin(1);
  lyt->setSpacing(0);
  connect(sp, SIGNAL(textChanged(const QString &)),
	  this, SLOT(setText(const QString &)));
}

// ************************************************************************* //
void	AttrString::setText(const QString &v)
{
  _m_data->writeAsString(v.toAscii().data());
}

// ************************************************************************* //
AttrPnt3d::AttrPnt3d(PlugPnt3d	&data,
		     QWidget	*parent): QWidget(parent), _m_data(&data)
{
  QFontMetrics	fm(parent->font());
  QLayout	*lyt = layout()? layout(): new QHBoxLayout(this);//, 1, 1, 0
  SpinBoxPnt3d	*sp = new SpinBoxPnt3d(MINDOUBLE, MAXDOUBLE, 1.0, this);

  sp->setRange(data.minValue(), data.maxValue());
  sp->setValue(_m_data->readAsPnt3d());

  lyt->addWidget(sp);
  lyt->setMargin(1);
  lyt->setSpacing(0);
  connect(sp, SIGNAL(valueChanged(Pnt3d)), this, SLOT(setValue(Pnt3d)));
}

// ************************************************************************* //
void	AttrPnt3d::setValue(Pnt3d v)
{
  _m_data->writeAsPnt3d(v);
}

// ************************************************************************* //
AttrVec3d::AttrVec3d(PlugVec3d	&data,
		     QWidget	*parent): QWidget(parent), _m_data(&data)
{
  QFontMetrics	fm(parent->font());
  QLayout	*lyt = layout()? layout(): new QHBoxLayout(this);//, 1, 1, 0
  SpinBoxVec3d	*sp = new SpinBoxVec3d(MINDOUBLE, MAXDOUBLE, 1.0, this);

  sp->setRange(data.minValue(), data.maxValue());
  sp->setValue(_m_data->readAsVec3d());
  //

  lyt->addWidget(sp);
  lyt->setMargin(1);
  lyt->setSpacing(0);
  connect(sp, SIGNAL(valueChanged(Vec3d)), this, SLOT(setValue(Vec3d)));
}

// ************************************************************************* //
void	AttrVec3d::setValue(Vec3d v)
{
  _m_data->writeAsVec3d(v);
}

// ************************************************************************* //
AttrRgb3d::AttrRgb3d(PlugRgb3d	&data,
		     QWidget	*parent): QWidget(parent), _m_data(&data)
{
  QFontMetrics	fm(parent->font());
  QLayout	*lyt = layout()? layout(): new QHBoxLayout(this);//, 1, 1, 0

  _m_sp = new SpinBoxRgb3d(MINDOUBLE, MAXDOUBLE, 1.0, this);

  _m_sp->setRange(data.minValue(), data.maxValue());
  _m_sp->setValue(_m_data->readAsRgb3d());

  lyt->addWidget(_m_sp);
  lyt->setMargin(1);
  lyt->setSpacing(0);
  connect(_m_sp, SIGNAL(valueChanged(Rgb3d)), this, SLOT(setValue(Rgb3d)));
}

// ************************************************************************* //
Rgb3d	AttrRgb3d::getValue() const
{
  return _m_data->readAsRgb3d();
}

// ************************************************************************* //
/// DESCRIPTION: Assign value to the plug. Spinbox needs update when this
///  method is called by another editor.
// ************************************************************************* //
void	AttrRgb3d::setValue(Rgb3d v)
{
  _m_data->writeAsRgb3d(v);
  _m_sp->setValue(v);
}
