// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:30 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <qlayout.h>
#include <qpixmap.h>
#include <qpainter.h>
#include <qmenubar.h>
#include "qtCurveEditor.h"
#include "plug.h"
#include "dagnode.h"
#include "datablock.h"
#include "nodetransform.h"
#include "nodecurve.h"

// ************************************************************************* //
/// DESCRIPTION: Creates a floating attribute editor on node.
/// Structure of Widgets:
/// (CurveEditor)
///  `-(CurveCentralWidget)
///     `-(CECurveView)
///     `-(CEDataView)
// ************************************************************************* //
CurveEditor::CurveEditor(DagNode	&node,
			 QWidget	*parent):
  QMainWindow(parent, Qt::Window), _m_node(&node)
{
  setWindowTitle("Curve Editor");

  node.getDataBlock().addValidBit(this);

  CurveCentralWidget	*centralwidget = new CurveCentralWidget(node, this);

  setCentralWidget(centralwidget);

  QMenu	*file = menuBar()->addMenu(tr("&File"));
  //QPopupMenu	*file = new QPopupMenu(this);

  file->addAction("&Close", this, SLOT(close()), Qt::CTRL + Qt::Key_Q);
  //menuBar()->insertItem("&File", file, 0, 0);

  QMenu	*edit = menuBar()->addMenu(tr("&Edit"));
  //QPopupMenu	*edit = new QPopupMenu(this);

  edit->addAction("&Frame", &centralwidget->curveview(),
		  SLOT(frameView()), Qt::Key_Space);
  //menuBar()->insertItem("&Edit", edit, 1, 1);
  setFocusPolicy(Qt::ClickFocus);
}

// ************************************************************************* //
CurveEditor::~CurveEditor()
{
  _m_node->getDataBlock().remValidBit(this);
}

// ************************************************************************* //
void CurveEditor::keyPressEvent(QKeyEvent *e)
{
  if (e->key() == Qt::Key_Escape)
    close();
  QMainWindow::keyPressEvent(e);
}

// ************************************************************************* //
void	CurveEditor::paintEvent(QPaintEvent *e)
{
  QMainWindow::paintEvent(e);
}

// ************************************************************************* //
CurveCentralWidget::CurveCentralWidget(DagNode		&node,
				       QWidget		*parent):
  QWidget(parent)
{
  QBoxLayout	*box = new QHBoxLayout(this);

  _m_curveview = new CECurveView(node, this);
  _m_dataview = new CEDataView(node, this);
  box->addWidget(_m_dataview, 0);
  box->addWidget(_m_curveview, 1);
  QWidget::connect(_m_dataview, SIGNAL(curveChanged(CEDataItem*)),
		   _m_curveview, SLOT(setCurve(CEDataItem*)));
}

// ************************************************************************* //
CECurveView::CECurveView(DagNode	&node,
			 QWidget	*parent):
  QWidget(parent), _m_node(&node),
  _m_min(-10.0, -10.0), _m_max(10.0, 10.0),
  _m_e0(QEvent::MouseButtonRelease, QPoint(0, 0), Qt::NoButton, Qt::NoButton,
	Qt::NoModifier),
  _m_dataItem(0)
{
  setBackgroundRole(QPalette::Background);
  //setBackgroundMode(Qt::NoBackground);
  //setFocusPolicy(Qt::StrongFocus);
}

// ************************************************************************* //
/// DESCRIPTION: Converts a point from the domain to window coordinates.
// ************************************************************************* //
Pnt2i	CECurveView::domainToWindow(const Pnt2d &p) const
{
  return Pnt2i(domainWidthToWindow(p.x), domainHeightToWindow(p.y));
}

// ************************************************************************* //
int	CECurveView::domainWidthToWindow(const double px) const
{
  return int(width() * (px - _m_min.x) / (_m_max.x - _m_min.x));
}

// ************************************************************************* //
int	CECurveView::domainHeightToWindow(const double py) const
{
  return int(height() * (_m_max.y - py) / (_m_max.y - _m_min.y));
}

// ************************************************************************* //
/// DESCRIPTION: Converts a point from window coordinates to the domain.
// ************************************************************************* //
Pnt2d	CECurveView::windowToDomain(const Pnt2i &q) const
{
  return Pnt2d(windowWidthToDomain(q.x), windowHeightToDomain(q.y));
}

// ************************************************************************* //
double	CECurveView::windowWidthToDomain(const int qx) const
{
  return _m_min.x + qx * (_m_max.x - _m_min.x) / double(width());
}

// ************************************************************************* //
double 	CECurveView::windowHeightToDomain(const int qy) const
{
  return _m_max.y - qy * (_m_max.y - _m_min.y) / double(height());
}

// ************************************************************************* //
/// DESCRIPTION: Inserts or delete a key.
/// TODO: Do not shortcut the graph if possible?
// ************************************************************************* //
void	CECurveView::mousePressEvent(QMouseEvent *e)
{
  if (_m_dataItem)
    {
      bool	isKeyDeleted(false);
      const std::list<Pnt2d> &ci(_m_dataItem->node().readAsPnt2dList("ci"));
      std::list<Pnt2d>::const_iterator kit;
      Pnt2d	torem;

      if (e->button() & Qt::LeftButton &&
	  !(e->modifiers() & Qt::ControlModifier))
	{
	  for (kit = ci.begin(); kit != ci.end(); ++kit)
	    {
	      const Pnt2i	p(domainToWindow(*kit));

	      if (abs(p.x - e->x()) < 4 && abs(p.y - e->y()) < 4)
		{
		  torem = *kit;
		  isKeyDeleted = true;
		  _m_dataItem->node().remkey(torem.x);
		  break;
		}
	    }
	  if (!isKeyDeleted)
	    _m_dataItem->node().setkey(windowToDomain(Pnt2i(e->x(), e->y())));
	  update();
	}
    }
  _m_e0 = *e;
}

// ************************************************************************* //
/// TODO: Select and drag
// ************************************************************************* //
void	CECurveView::mouseMoveEvent(QMouseEvent *e)
{
  if (e->modifiers() & Qt::ControlModifier)
    if (e->buttons() & Qt::LeftButton)
      {
	const double	s(pow(1.01, _m_e0.y() - e->y()));
	const Pnt2d	c((_m_min + _m_max) * 0.5);

	_m_min = c + (_m_min - c) * s;
	_m_max = c + (_m_max - c) * s;
	update();
      }
    else if (e->buttons() & Qt::MidButton)
      {
	const Pnt2d	p0(windowToDomain(Pnt2i(_m_e0.x(), _m_e0.y())));
	const Pnt2d	p1(windowToDomain(Pnt2i(e->x(), e->y())));
	const Vec2d	v(p1 - p0);

	_m_min -= v;
	_m_max -= v;
	update();
      }
    else if (e->buttons() & Qt::RightButton)
      {
	const double	sx(pow(1.01, _m_e0.x() - e->x()));
	const double	cx((_m_min.x + _m_max.x) * 0.5);

	_m_min.x = cx + (_m_min.x - cx) * sx;
	_m_max.x = cx + (_m_max.x - cx) * sx;

	const double	sy(pow(1.01, e->y() - _m_e0.y()));
	const double	cy((_m_min.y + _m_max.y) * 0.5);

	_m_min.y = cy + (_m_min.y - cy) * sy;
	_m_max.y = cy + (_m_max.y - cy) * sy;

	update();
      }
  _m_e0 = *e;
}

// ************************************************************************* //
/// DESCRIPTION: Draw grid.
// ************************************************************************* //
void	CECurveView::drawGrid(QPainter &p)
{
  const double	inci(fabs(windowHeightToDomain(40) - windowHeightToDomain(0)));
  const double	incj(fabs(windowWidthToDomain(40) - windowWidthToDomain(0)));
  const int	hline(height());

  double iFps = 1.0;
  for (int i(1), b(1); b < inci; ++i, b = !(i % 3)? 2 * b + b / 2: 2 * b)
    iFps = b;
  double jFps = 1.0;
  for (int i(1), b(1); b < incj; ++i, b = !(i % 3)? 2 * b + b / 2: 2 * b)
    jFps = b;

  for (double i(iFps); i < windowHeightToDomain(0); i += iFps)
    {
      p.drawText(0, domainHeightToWindow(i), QString::number(i));
      p.drawRect(0, domainHeightToWindow(i), width(), 1);
    }
  for (double i(iFps); windowHeightToDomain(height()) < i; i -= iFps)
    {
      p.drawText(0, domainHeightToWindow(i), QString::number(i));
      p.drawRect(0, domainHeightToWindow(i), width(), 1);
    }
  for (double j(jFps); j < windowWidthToDomain(width()); j += jFps)
    {
      p.drawText(domainWidthToWindow(j), hline, QString::number(j));
      p.drawRect(domainWidthToWindow(j), 0, 1, height());
    }
  for (double j(jFps); windowWidthToDomain(0) < j; j -= jFps)
    {
      p.drawText(domainWidthToWindow(j), hline, QString::number(j));
      p.drawRect(domainWidthToWindow(j), 0, 1, height());
    }
}

// ************************************************************************* //
/// DESCRIPTION: Draws curve.
// ************************************************************************* //
void	CECurveView::drawCurve(QPainter &p)
{
  if (!_m_dataItem)
    return;
      
  const std::list<Pnt2d> &ci(_m_dataItem->node().readAsPnt2dList("ci"));
  std::list<Pnt2d>::const_iterator kit;
  std::vector<const Pnt2d*>	dpoints(ci.size());
  std::vector<Pnt2i>	wpoints(ci.size());

  QPen	pen(Qt::SolidLine);

  if (ci.empty())
    return;
  pen.setWidth(2);
  pen.setColor(Qt::red);
  p.setPen(pen);
  kit = ci.begin();
  for (int i(0); kit != ci.end(); ++kit, ++i)
    {
      dpoints[i] = &(*kit);
      wpoints[i] = domainToWindow(*dpoints[i]);
      p.drawRect(wpoints[i].x - 2, wpoints[i].y - 2, 4, 4);
    }
  pen.setColor(Qt::black);
  pen.setWidth(2);
  p.setPen(pen);
  for (int i(1), n(ci.size()); i < n; ++i)
    {
      std::vector<Pnt2i>	ptab;
      const Pnt2i	&wl(wpoints[i - 1]), &wr(wpoints[i]);
      const Pnt2d	&pl(*dpoints[i - 1]), &pr(*dpoints[i]);
	  
      if ((wl.x < 0 && wr.x < 0) || (width() < wl.x && width() < wr.x) ||
	  (wl.y < 0 && wr.y < 0) || (height() < wl.y && height() < wr.y))
	continue;
	  
      const int	m(max(1, max(abs(wl.x - wr.x), abs(wl.y - wr.y)) / 2));
	  
      ptab = std::vector<Pnt2i>(m + 1);
      for (int j(0); j <= m; ++j)
	{
	  const double	x(pl.x + (pr.x - pl.x) * (j / double(m)));

	  _m_dataItem->node().writeAsDouble("x", x);
	  ptab[j] =
	    domainToWindow(Pnt2d(x, _m_dataItem->node().readAsDouble("y")));
	}
      for (int j(1); j <= m; ++j)
	p.drawLine(ptab[j - 1].x, ptab[j - 1].y, ptab[j].x, ptab[j].y);
    }
}

// ************************************************************************* //
/// DESCRIPTION: Deselects the previous one, updates display and assigns the 
///  new one.
// ************************************************************************* //
void CECurveView::setCurve(CEDataItem *c)
{
  // TODO: deselect previous one!!
  if (_m_dataItem)
    {
      _m_dataItem->select(false);
      _m_dataItem->update();
    }
  _m_dataItem = c;
  update();
}

// ************************************************************************* //
void	CECurveView::frameView()
{
  if (!_m_dataItem)
    return;

  const std::list<Pnt2d>	&ci(_m_dataItem->node().
				    readAsPnt2dList("ci"));

  if (ci.size() <= 1)
    return;

  std::list<Pnt2d>::const_iterator kit;
  Pnt2d	min(ci.front()), max(min);
  
  for (kit = ci.begin(); kit != ci.end(); ++kit)
    {
      if (kit->x < min.x)
	min.x = kit->x;
      else if (max.x < kit->x)
	max.x = kit->x;
      if (kit->y < min.y)
	min.y = kit->y;
      else if (max.y < kit->y)
	max.y = kit->y;
    }
  if (fabs(min.x - max.x) < 1.0e-16 || fabs(min.y - max.y) < 1.0e-16)
    return;
  _m_min = min;
  _m_max = max;

  const Pnt2d	c((_m_min + _m_max) * 0.5);
  const double	s(pow(1.01, 4.0));

  _m_min = c + (_m_min - c) * s;
  _m_max = c + (_m_max - c) * s;

  update();
}

// ************************************************************************* //
void	CECurveView::keyPressEvent(QKeyEvent *e)
{
  if (e->key() == Qt::Key_Space && _m_dataItem)
    frameView();
}

// ************************************************************************* //
void	CECurveView::paintEvent(QPaintEvent *)//e
{
  //QPixmap	pm(size());
  QPainter	p;
  DagNode	*clock = DagNode::getNodeByPath(":Clock0");

  p.begin(this);

  const Pnt2i	ps(domainToWindow(Pnt2d(clock->readAsInt("sf"), 0.0)));
  const Pnt2i	pe(domainToWindow(Pnt2d(clock->readAsInt("ef"), 0.0)));
  const int	pcx(domainWidthToWindow(double(clock->readAsInt("f"))));

  // Draw background
  p.setPen(Qt::NoPen);
  p.setBrush(QColor(189, 189, 189));
  p.drawRect(ps.x, 0, pe.x - ps.x, height());
  p.setPen(Qt::NoPen);
  p.setBrush(QColor(230, 230, 230));
  p.drawRect(0, 0, ps.x, height());
  p.drawRect(pe.x, 0, width() - ps.x, height());
  p.setBrush(QColor(128, 128, 128));
  p.drawRect(0, pe.y, width(), 2);

  // Draw grid
  drawGrid(p);
  
  // Draw current line
  p.setBrush(Qt::black);
  p.drawRect(pcx, 0, 1, height());

  // FIND the selected curve!!! PLOT!!!
  drawCurve(p);

  p.end();
  //bitBlt(this, -1, -1, &pm, -1, -1);
}

// ************************************************************************* //
/// DESCRIPTION:
///   Finds the input connection to all direct children nodes.
/// TODO: pass "exist" bits to the datablock and nodes (to prevent segfaults).
// ************************************************************************* //
CEDataView::CEDataView(DagNode		&node,
		       QWidget		*parent): 
  QWidget(parent), _m_node(&node)
{
  DataBlock::iterator	dit;
  QBoxLayout	*box = new QVBoxLayout(this);
  DagNode	*d1;
  std::list<DagNode *>	nodes;
  std::list<DagNode *>::iterator	nit;
  std::list<NCurve *>	curves;

  std::list<NCurve *>::iterator	cit;
  std::list<String>	curveNames;
  std::list<String>::iterator	cnit;

  int	count(0);

  node.getNodesByDepth(1, nodes);
  for (nit = nodes.begin(); nit != nodes.end(); ++nit)
    if ((*nit)->getNodeParent() && 
	(*nit)->getNodeParent()->getTypeStr() == "Transform")
      nodes.push_front((*nit)->getNodeParent());

  for (nit = nodes.begin(); nit != nodes.end(); ++nit)
    {
      DataBlock	&db((*nit)->getDataBlock());

      for (dit = db.begin(); dit != db.end(); ++dit)
	if ((d1 = (*dit)->getInputNode()) &&
	    d1->getTypeNbr() == NCurve::typeNbr)
	  curves.push_back(static_cast<NCurve*>(d1)),
	    curveNames.push_back((*dit)->shortName());
    }
  if (node.getTypeNbr() == NCurve::typeNbr)
    {
      curves.push_back(static_cast<NCurve*>(&node));
      curveNames.push_back("y(x)");
    }

  for (cit = curves.begin(), cnit = curveNames.begin(); cit != curves.end();
       ++cit, ++cnit)
    {
      ++count;
      _m_dataItems.push_back
	(new CEDataItem(**cit, this, (*cnit).data()));
      box->addWidget(_m_dataItems.back(), 0);
      QWidget::connect
	(_m_dataItems.back(), SIGNAL(curveChanged(CEDataItem*)),
	 this, SIGNAL(curveChanged(CEDataItem*)));
    }
  //box->addWidget(new QWidget(this), 10);
  box->insertStretch(count, 10);
}

// ************************************************************************* //
CEDataItem::CEDataItem(NCurve		&curve, 
		       QWidget		*parent,
		       const char	*name): 
  QWidget(parent), _m_curve(&curve), _m_isSelected(false), _m_name(name)
{
  QFontMetrics	fm(font());
  QSize		qs(fm.size(Qt::TextSingleLine, name));

  setBaseSize(qs.width() + 4, qs.height() + 4);
  setMinimumSize(qs.width() + 4, qs.height() + 4);
}

// ************************************************************************* //
void	CEDataItem::select(bool on)
{
  if ((_m_isSelected = on))
    _m_curve->writeAsInt("sel", true);
}

// ************************************************************************* //
NCurve	&CEDataItem::node()
{
  return *_m_curve;
}

// ************************************************************************* //
/// DESCRIPTION: Selects the curve
// ************************************************************************* //
void	CEDataItem::mousePressEvent(QMouseEvent *)//e
{
  if ((_m_isSelected = !_m_isSelected))
    {
      _m_curve->writeAsInt("sel", true);
      emit curveChanged(this);
    }
  // TODO: send signal to CEDataView with ref to that curve??
  // TODO!!! ADD OR REMOVE CURVES FROM PARENT!!!
  update();
}

// ************************************************************************* //
void	CEDataItem::paintEvent(QPaintEvent *)//e
{
  //QPixmap	pm(size());
  QFontMetrics	fm(font());
  QPainter	p;

  p.begin(this);
  if (_m_isSelected)
    p.setPen(Qt::white), p.setBrush(QColor(0, 0, 0));
  else
    p.setPen(Qt::black), p.setBrush(QColor(255, 255, 255));
  p.drawRect(0, 0, width(), height());
  p.drawText(2, height() - fm.descent() - 2, _m_name.data());
  p.end();
  //bitBlt(this, -1, -1, &pm, -1, -1);
}
