// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:30 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <list>
#include <vector>
#include <iostream>
#include <fstream>
#include <string>

// ************************************************************************* //
#include <math.h>
#include <GL/glu.h>

// ************************************************************************* //
#include <qfiledialog.h>
#include <qmessagebox.h>
#include <qinputdialog.h> 
#include <qtimer.h>
#include <qgl.h>
#include <qcolordialog.h> 
#include <qpainter.h>
#include <QBoxLayout>
  
// ************************************************************************* //
#include "gfxtypes.h"
#include "qtMainWidget.h"
#include "qtGlWidget.h"
#include "qtClockEditor.h"
#include "qtAttributeEditor.h"
#include "qtCurveEditor.h"
#include "nodetransform.h"
#include "datablock.h"
#include "plug.h"
#include "gbb.h"

// ************************************************************************* //
CentralWidget::CentralWidget(QWidget	*parent): QWidget(parent)
{
  ClockEditor	 *clockeditor = new ClockEditor(this);
  SceneGLWidget *glwidget = new SceneGLWidget(this);
  QVBoxLayout	 *grid = new QVBoxLayout(this);
  CommandEditor *commandeditor = new CommandEditor(this);

  QWidget::connect(clockeditor, SIGNAL(repaintglwidget()),
		   glwidget, SLOT(updateGL()));
  QWidget::connect(commandeditor, SIGNAL(repaintglwidget()),
		   glwidget, SLOT(updateGL()));
  QWidget::connect(glwidget, SIGNAL(repaintframeslider()),
		   &clockeditor->frameSlider(), SLOT(update()));

  _m_glwidget = glwidget;
  _m_clockEditor = clockeditor;
  grid->addWidget(glwidget, 10);
  grid->addWidget(clockeditor, 0, Qt::AlignBottom);
  grid->addWidget(commandeditor, 0, Qt::AlignBottom);
}

// ************************************************************************* //
void CommandEditor::keyPressEvent(QKeyEvent *e)
{
  if (e->key() == Qt::Key_Return)
    {
      if (!text().isEmpty())
	{
	  _m_history.push_back(text());
	  _m_hit = _m_history.end();
	  Command::run(text().toAscii().data());
	  emit repaintglwidget();
	  setText("");
	}
    }
  else if (e->key() == Qt::Key_Up)
    {
      if (!_m_history.empty())
	if (_m_hit != _m_history.begin())
	  setText(*--_m_hit);
	else
	  setText(*_m_hit);
    }
  else if (e->key() == Qt::Key_Down)
    {
      if (_m_hit != _m_history.end())
	if (++_m_hit != _m_history.end())
	  setText(*_m_hit);
	else
	  setText("");
    }
  else
    QLineEdit::keyPressEvent(e);
}

// ************************************************************************* //
SceneGLWidget::SceneGLWidget(QWidget	*parent):
  QGLWidget(parent),
  _m_trackball(_m_view),
  _m_isInViewMode(false),
  _m_e0(QEvent::None, QPoint(0, 0), Qt::NoButton, Qt::NoButton,
	Qt::NoModifier),
  _m_mouseHasDragged(false)
{
  _m_context = new QGLContext(format(), this);
  setFocusPolicy(Qt::StrongFocus);
}

// ************************************************************************* //
void	drawGrid()
{
  const int	n(8);
  const double	size(5.0);

  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);
  glPushAttrib(GL_LINE_BIT);
  glLineWidth(2.0);
  glBegin(GL_LINES);
  glColor3d(0.0, 0.0, 0.0);
  glVertex3d(-size, 0.0, 0.0);
  glVertex3d(size, 0.0, 0.0);
  glVertex3d(0.0, -size, 0.0);
  glVertex3d(0.0, size, 0.0);
  glEnd();//GL_LINES
  glPopAttrib();//GL_LINE_BIT
  glBegin(GL_LINES);
  glColor3d(0.5, 0.5, 0.5);
  for (int i(-n); i <= n; ++i)
    {
      const double	pi = i * size / double(n);

      glVertex3d(pi, -size, 0.0);
      glVertex3d(pi, size, 0.0);
      glVertex3d(-size, pi, 0.0);
      glVertex3d(size, pi, 0.0);
    }
  glEnd();//GL_LINES
  glPopAttrib();//GL_LIGHTING_BIT
}

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
void	SceneGLWidget::paintGL()
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  _m_trackball.glRenderPlace();

  if (_m_isInViewMode)
    _m_trackball.glRender();

  std::list<DagNode*>	cameras;

  DagNode::root().DagNode::getNodesByType("Camera", cameras);

  DagNode	*camera = cameras.front();
  DagNode	*imageLayer = DagNode::getNodeByPath(":ImageLayer0");
  DagNode	*clock = DagNode::getNodeByPath(":Clock0");
  DagNode	*sky = DagNode::getNodeByPath(":Sky0");
  Vec3d		dir, up;
  Frustum	fs;

  dir.glGetViewdir();
  up.glGetUp();
  fs.getWorldFrustum();

  camera->writeAsVec3d("up", up);
  camera->writeAsVec3d("dir", dir);
  camera->writeAsCustom("frustum", &fs);

  camera->writeAsDouble("zNear", _m_trackball.getzNear());
  camera->writeAsDouble("zFar", _m_trackball.getzFar());
  
  camera->writeAsMat4d("model", _m_view._m_model);
  camera->writeAsMat4d("proj", _m_view._m_proj);
  camera->writeAsMat2i("view", _m_view._m_view);

  DagNode::disconnect(*sky->getPlug("camera")->getInputNode(), "signal",
		      *sky, "camera");
  DagNode::connect(*camera, "signal", *sky, "camera");

  DagNode::disconnect(*imageLayer->getPlug("camera")->getInputNode(),
		      "signal", *imageLayer, "camera");
  DagNode::connect(*camera, "signal", *imageLayer, "camera");

  drawGrid();

  camera->writeAsBool("isVisible", false);
  DagNode::root().writeAsInt("renderIn", Gbb::Modeling);
  DagNode::root().writeAsSignal("renderIn");
  DagNode::root().readAsSignal("renderOut");//???
  camera->writeAsBool("isVisible", true);

  QFontMetrics	fm(font());
  QString	str(QString::number(int(clock->readAsDouble("efps"))) + "fps");
  
  glColor3d(0.0, 0.0, 0.0);
  renderText(5, 5 + fm.height(), str);

  glFlush();
}

// ************************************************************************* //
/// TODO: Migrate everything not openGL in MainWidget.
// ************************************************************************* //
void	SceneGLWidget::keyPressEvent(QKeyEvent *e)
{
  std::list<DagNode*> nodes;
  std::list<DagNode*>::iterator	nit;

  if (e->key() == Qt::Key_Space)
    _m_isInViewMode = !_m_isInViewMode, updateGL();
  else if (e->key() == Qt::Key_A && (e->modifiers() & Qt::ControlModifier))
    {
      DagNode::getNodesSelected(nodes);
      for (nit = nodes.begin(); nit != nodes.end(); ++nit)
	{
	  AttributeEditor *newf = new AttributeEditor(**nit);
	  newf->show();
	}
    }
  else if (e->key() == Qt::Key_C && (e->modifiers() & Qt::ControlModifier))
    {
      DagNode::getNodesSelected(nodes);
      for (nit = nodes.begin(); nit != nodes.end(); ++nit)
	{
	  CurveEditor *newf = new CurveEditor(**nit);
	  newf->show();
	}
    }
  else if (e->key() == Qt::Key_Delete)
    {
      DagNode::getNodesSelected(nodes);
      for (nit = nodes.begin(); nit != nodes.end(); ++nit)
	Command::run(String("delete -path ") + (*nit)->path());
      updateGL();
    }
  else if (e->key() == Qt::Key_Escape)
    static_cast<CentralWidget*>(parentWidget())->clockEditor()->cmdPause();
}

/* ************************************************************************* */
void	View::glOpen(const int	x,
		     const int	y) const
{
  //Mat2i	view(GL_VIEWPORT);
  glSelectBuffer(BUFSIZE, _m_buffer);

  glRenderMode(GL_SELECT);
  glClear(GL_DEPTH_BUFFER_BIT);
  glPushAttrib(GL_TRANSFORM_BIT);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPickMatrix(x, y, 5.0, 5.0, _m_view._m_m);
  glPopAttrib();// GL_TRANSFORM_BIT
  glInitNames();
}

// ************************************************************************* //
/// DESCRIPTION: returns array [nb, near depth, far depth, name0, ... , 0]
// ************************************************************************* //
GLuint	*View::glClose() const
{
  glFlush();
  if ((_m_hits = glRenderMode(GL_RENDER)) < 1)
    return 0;

  const GLuint	*ptr(_m_buffer);
  const GLuint	*retptr(0);
  GLuint	retz(MAXUINT);
  GLuint	retnames(0);

  for (int i(0); i < _m_hits; ++i, ptr += 3 + *ptr)
    if (ptr[1] <= retz)
      {
		  retnames = ptr[0];
		  retz = ptr[1];
		  retptr = ptr;
      }
  for (GLuint i(0); i < retnames + 3; ++i)
    _m_ret[i] = retptr[i];
  return _m_ret;
}

// ************************************************************************* //
/// DESCRIPTION:
///   If gizmo is visible, tries to grab gizmo.
// ************************************************************************* //
void	SceneGLWidget::mouseReleaseEvent(QMouseEvent *e)
{
  const GLuint	*buf(0);

  if (!_m_isInViewMode && !_m_mouseHasDragged)
    {
      DagNode	*gizmo = DagNode::getNodeByPath(":Gizmo0");
      
      if (gizmo->readAsBool("visible"))
	{
	  _m_view.glOpen(e->x(), _m_view.view()(3) - e->y());
	  _m_trackball.glSelectPlace();
	  gizmo->writeAsInt("renderIn", Gbb::Selecting);
	  gizmo->writeAsSignal("renderIn");
	  gizmo->readAsSignal("renderOut");
	  buf = _m_view.glClose();
	}
      if (!buf)
	{
	  _m_view.glOpen(e->x(), _m_view.view()(3) - e->y());
	  _m_trackball.glSelectPlace();
	  DagNode::root().writeAsInt("renderIn", Gbb::Selecting);
	  DagNode::root().writeAsSignal("renderIn");
	  DagNode::root().readAsSignal("renderOut");
	  buf = _m_view.glClose();
	}
      if (buf && buf[0])
	{
	  const int	level((e->modifiers() & Qt::ControlModifier)? 1: 0);
	  String	cmd(String("select -level ") + level + " ");

	  if (e->modifiers() & Qt::ShiftModifier)
	    cmd += "-multiple ";
	  cmd += "-path ";

	  String	path;
	  bool		isNodeSelected(true);
	  DagNode	*node(&DagNode::root());

	  _m_z0 = buf[1];
	  for (unsigned int i(0); i < buf[0]; i += 3)
	    if (node->getTypeStr() == "Gizmo")
	      {
		node->writeAsInt("axis", buf[3 + i]);
		break;
	      }
	    else if (buf[3 + i] == 0)
	      {
		if ((node = node->getNode(buf[3 + i + 1], buf[3 + i + 2])))
		  path += String(":") + node->getInstStr();
		else
		  std::cerr << "Error: no such node id as " << buf[3 + i + 1]
			    << " " << buf[3 + i + 2] << std::endl;
	      }
	    else if (buf[3 + i] == 1 && level == 1)
	      {
		Plug	*hnd;

		if (node && (hnd = node->getPlug(buf[3 + i + 2])))
		  {
		    if (hnd->isArray())
		      path += (String(".") + hnd->shortName() + "[" + 
			       buf[3 + i++ + 3] + "]");
		    else
		      path += String(".") + hnd->shortName();
		  }
		isNodeSelected = false;
		break;
	      }
	    else
	      break;
	  if (node->getTypeStr() != "Gizmo")
	    Command::run(cmd + path);
	  emit repaintframeslider();
	  updateGL();
	}
      else
	{
	  Command::run("deselect");
	  emit repaintframeslider();
	  updateGL();
	}
    }
  updateGL();
  _m_mouseHasDragged = false;
  _m_e0 = *e;
}

// ************************************************************************* //
void	SceneGLWidget::mousePressEvent(QMouseEvent *e)
{
  _m_mouseHasDragged = false;
  if (_m_isInViewMode)
    _m_trackball.mousePressEvent(e), updateGL();
  _m_e0 = *e;
}

// ************************************************************************* //
void	SceneGLWidget::mouseMoveEvent(QMouseEvent *e)
{
  if (_m_isInViewMode)
    {
      _m_trackball.mouseMove(_m_e0, *e);
      updateGL();
    }
  else
    if (e->buttons() & Qt::LeftButton)
      {
	const GLuint	*buf(0);
	View	view = _m_view;
	DagNode	*clicked(0);
	DagNode	*gizmo = DagNode::getNodeByPath(":Gizmo0");

	if (gizmo->readAsBool("visible"))
	  {
	    _m_view.glOpen(_m_e0.x(), _m_view.view()(3) - _m_e0.y());
	    _m_trackball.glSelectPlace();
	    gizmo->writeAsInt("renderIn", Gbb::Selecting);
	    gizmo->writeAsSignal("renderIn");
	    gizmo->readAsSignal("renderOut");
	    buf = _m_view.glClose();
	  }
	if (!buf)
	  {
	    _m_view.glOpen(_m_e0.x(), _m_view.view()(3) - _m_e0.y());
	    _m_trackball.glSelectPlace();
	    DagNode::root().writeAsInt("renderIn", Gbb::Selecting);
	    DagNode::root().writeAsSignal("renderIn");
	    DagNode::root().readAsSignal("renderOut");
	    buf = _m_view.glClose();
	  }

	bool	isSelClicked(false);

	if (buf && buf[0])
	  {
	    DagNode	*node(&DagNode::root());
	    String	path;
	    
	    for (unsigned int i(0); i < buf[0]; i += 3)
	      if (node->getTypeStr() == "Gizmo")
		break;
	      else if (buf[3 + i] == 0)
		{
		  if ((node = node->getNode(buf[3 + i + 1], buf[3 + i + 2])))
		    path += String(":") + node->getInstStr();
		  else
		    std::cerr << "Error: no such node id as " << buf[3 + i + 1]
			      << " " << buf[3 + i + 2] << std::endl;
		}
	      else if (buf[3 + i] == 1)
		{
		  Plug	*hnd;
		  
		  if (node && (hnd = node->getPlug(buf[3 + i + 2])))
		    {
		      if (hnd->isArray())
			path += (String(".") + hnd->shortName() + "[" + 
				 buf[3 + i++ + 3] + "]");
		      else
			path += String(".") + hnd->shortName();
		    }
		  break;
		}
	    //clicked = DagNode::getNodeByPath(path.chopBack('.'));
	    //std::cout << clicked << " " << node << std::endl;
	    //if (node->)
	    clicked = node;
	    if (clicked->getPlug("sel") && clicked->readAsInt("sel"))
	      isSelClicked = true;
	  }
	if (clicked && !clicked->readAsInt("sel") && !_m_mouseHasDragged &&
	    !(_m_e0.modifiers() & Qt::ShiftModifier))
	  Command::run(String("select -level 0 -path ") + clicked->path());
	if (_m_mouseHasDragged || isSelClicked ||
	    (clicked && clicked->getTypeStr() == "Gizmo"))
	  {
	    if (!_m_mouseHasDragged)
	      _m_z0 = buf[1];

	    const Pnt3d	p0(_m_e0.x(), (view.view()(3) - _m_e0.y()),
			   _m_z0 / 4294967295.0);
	    const Pnt3d	p1(e->x(), (view.view()(3) - e->y()),
			   _m_z0 / 4294967295.0);
	    const Pnt3d	q1(view.unProject(p1));
	    const Pnt3d	q0(view.unProject(p0));

	    String cmd = 
	      String("translate -p0 ") + q0.x + " " + q0.y + " " + q0.z +
	      String(" -p1 ") + q1.x + " " + q1.y + " " + q1.z;
	    Command::run(cmd);
	    _m_mouseHasDragged = true;
	  }
      }
  _m_e0 = *e;
  updateGL();
}
 
// ************************************************************************* //
void	SceneGLWidget::resetView()
{
  _m_trackball.reset();
  updateGL();
}

// ************************************************************************* //
bool	SceneGLWidget::saveView()
{
  QString s = QFileDialog::getSaveFileName(this, "save file dialog", ".",
					   "View (*.vu)");

  if (s.isEmpty())
    return false;
  if (!s.contains(".vu"))
    s = s + ".vu";

  std::ofstream	ofs(s.toAscii().data(), std::ifstream::binary);

  if (!ofs)
    return false;
  ofs << _m_trackball;
  return true;
}

// ************************************************************************* //
bool	SceneGLWidget::loadView()
{
  QString s = QFileDialog::getOpenFileName(this, "open file dialog", ".",
					   "View (*.vu)");

  if (s.isEmpty())
    return false;

  std::ifstream	ifs(s.toAscii().data(), std::ifstream::binary);

  if (!ifs)
    return false;
  ifs >> _m_trackball;
  updateGL();
  return true;
}

// ************************************************************************* //
void	SceneGLWidget::setGouraud()
{
  glPolygonMode(GL_FRONT, GL_FILL);
  glEnable(GL_CULL_FACE);
  //glEnable(GL_BLEND);
  //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  //glAlphaFunc(GL_EQUAL, 1.0);
  updateGL();
}

// ************************************************************************* //
void	SceneGLWidget::setWireframe()
{
  glPolygonMode(GL_FRONT, GL_LINE);
  glEnable(GL_CULL_FACE);
  updateGL();
}

// ************************************************************************* //
CommandEditor::CommandEditor(QWidget	*parent): QLineEdit(parent)
{
  _m_hit = _m_history.begin();
}

// ************************************************************************* //
void	SceneGLWidget::initializeGL()
{
  const GLfloat	specular[] = {1.0, 1.0, 1.0, 1.0};
  const GLfloat	shininess[] = {50.0};

  glClearColor(0.647, 0.647, 0.647, 0.0);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_NORMALIZE);

  glShadeModel(GL_SMOOTH);
  glEnable(GL_COLOR_MATERIAL);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);
  glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);

  const GLfloat	lDiffuse0[] = {0.1f, 0.1f, 0.1f, 1.0f};
  const GLfloat	lDiffuse1[] = {0.6f, 1.0f, 0.6f, 1.0f};
  const GLfloat	lDiffuse2[] = {1.0f, 0.6f, 0.6f, 1.0f};
  const GLfloat	lDiffuse3[] = {1.0f, 1.0f, 1.0f, 1.0f};

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  glEnable(GL_LIGHT2);
  glEnable(GL_LIGHT3);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, lDiffuse0);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, lDiffuse1);
  glLightfv(GL_LIGHT2, GL_DIFFUSE, lDiffuse2);
  glLightfv(GL_LIGHT2, GL_DIFFUSE, lDiffuse3);
  glPointSize(4.0);
}
